#!/usr/bin/lua
local coaf = require "coaf.core"
local thread = require "coaf.thread"
local socket = require "cosocket.core"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
local pt = require "pl.pretty"

coaf.Init("coaf_call")
client.InitSocket(socket, adapter)

local function usage()
	print( "Usage:" )
	print( "  cocall.lua [--addr=<address:port>] [--key=<author-key>]" )
	print( "             <location> <service> <object> [:]<method> [arg1 arg2 arg3 ...]" )
	print( "" )
	print( "  If you add ':' before method name, that means call specified method by" )
	print( "  'obj:Call(...)', otherwise the call is 'obj.Call(...)' " )
	print( "" )
	print( "  Note that if 'argX' is (digitals) e.g. (123) (-1234) (0x12345) ..., it" )
	print( "  will convert it to number, otherwise the argument type is string." )
	print( "" )
end

---- Main.
local function Main(args)
	local i, n
	local adr, key
	for i=1, #args do
		if args[i]:sub(1,2) == "--" then
			local k,v = args[i]:match("(.-)=(.*)$")
			if k == "--addr" then
				adr = v
			elseif k == "--key" then
				key = v
			elseif k == "--help" then
				usage()
				coaf.Exit(0)
				return
			else
				print( "*** ERROR: Invalid option " .. key )
				usage()
				coaf.Exit(1)
				return
			end
		else
			n = i
			break
		end
	end

	if not n then
		usage()
		coaf.Exit(0)
		return
	end

	local loc = args[n]
	local svc = args[n+1]
	local obj = args[n+2]
	local mtd = args[n+3]

	local va_args = {}
	local arg_idx
	for arg_idx = n+4, #args do
		local x = args[ arg_idx ]
		x = x:match("^%((.+)%)$")
		if x then
			table.insert( va_args, tonumber(x) or args[arg_idx] )
		else
			table.insert( va_args, args[arg_idx] )
		end
	end

	if not loc or not svc or not obj or not mtd then
		usage()
		coaf.Exit(1)
		return
	end

	local cli = client.New( svc, loc, adr, nil, key )
	if not cli then
		print("*** ERROR: Connect to service: " .. loc .. "/" .. svc .. " failed!" )
	else
		local proxy = cli:CreateProxy( obj, true )
		if not proxy then
			print( "*** ERROR: Create proxy " .. obj .. " failed!" )
		else
			local objcall
			if mtd:sub(1,1) == ":" then
				objcall = true
				mtd = mtd:sub(2)
			end

			local r 
			if objcall then
				r = { proxy[mtd]( proxy, table.unpack(va_args) ) }
			else
				r = { proxy[mtd]( table.unpack(va_args) ) }
			end
			proxy:Destroy()

			local i, v
			print( ">>> Call returns:" )
			for i, v in ipairs(r) do
				if type(v) == "table" then
					print( "#" .. i .. ":" )
					pt.dump(v)
				else
					print( "#" .. i .. ":", v )
				end
			end
		end
		cli:Destroy()
	end
	coaf.Exit(0)
end

coaf.Start(Main)

