local coaf=require "coaf.core"
local _dbg = require "coaf.debug"
local periphery = require "periphery"
local GPIO = periphery.GPIO
local video_formats = require "video_formats"

local chip_sii9136 = require "chips.sii9136"
local chip_ite66021 = require "chips.ite66021"
local chip_aic3104 = require "chips.aic3104"


local _M = {
	work_mode = "encoding",
	i_hdmi_tx = nil,
	i_hdmi_rx = nil,
	i_aic3104 = nil,
	cur_video_ver = nil,
	cur_audio_ver = nil,

	output_video_set = nil,
	output_audio_set = nil,
	output_video_fmt = nil,
	output_video_err = nil,
	output_audio_fmt = nil,

	last_video_changed = nil,
	last_audio_changed = nil
}

--=====================================================
local deinterlace_map = {
	["1080i60"] = "1080p60",
	["1080i59"] = "1080p59",
	["1080i50"] = "1080p50"
}
--=====================================================

local function onRetrivalVideoFormat(sn, fourCC, xRes, yRes, frN, frD, interlaced, restrict )
	--TODO: Get saved resolutions
	local m, fmt, failsafe
	local fr
	if frN and frD and frD > 0 then
		fr = math.floor(frN*100/frD)/100
	else
		fr = 60
	end

	if restrict == "default" then
		m = "xy"
		fmt = video_formats.GetFailSafeFormat()
		failsafe = false --For initialized format, make it be valid.
	else
		if not xRes or xRes <= 0 then
			xRes = nil
		end
		if not yRes or yRes <= 0 then
			yRes = nil
		end

		local specFr
		if _M.output_video_set and _M.output_video_set.frame_rate and _M.output_video_set.frame_rate > 23 then
			specFr = _M.output_video_set.frame_rate
		end

		if _G.DISABLE_4K and xRes and xRes > 2160 and yRes and yRes > 1280 then
			m = "xy"
			failsafe = true
			_M.output_video_fmt = "match-error"
			fmt = video_formats.GetFailSafeFormat()
		elseif fourCC == "SHQ0" or fourCC == "shq0" then
			m, fmt = video_formats.GetFormatSetting( "UHD_420", specFr, xRes, yRes, fr, interlaced or false )
			if not fmt then
				m = "xy"
				failsafe = true
				_M.output_video_err = "match-error"
				fmt = video_formats.GetFailSafeFormat()
			else
				_M.output_video_err = nil
			end
		elseif fourCC == "SHQ2" or fourCC == "shq2" or fourCC == "SHQ7" or fourCC == "shq7" then
			m, fmt = video_formats.GetFormatSetting( { "UHD", "HD", "SD" }, specFr, xRes, yRes, fr, interlaced or false )
			if not fmt then
				failsafe = true
				m = "xy"
				_M.output_video_err = "match-error"
				fmt = video_formats.GetFailSafeFormat()
			else
				_M.output_video_err = nil
			end
		else
			failsafe = true
			m = "xy"
			_M.output_video_err = "unsupported-codec"
			fmt = video_formats.GetFailSafeFormat()
		end
	end

	_M.output_video_fmt = {}

	if _M.output_video_set.format == "deint" and fmt.interlaced then
		local deinterlace_id = deinterlace_map[fmt.id]
		local deinterlace_fmt
		if deinterlace_id then
			deinterlace_fmt = video_formats.GetFormatByID(deinterlace_id)
		end

		if not deinterlace_fmt then
			for k,v in pairs(fmt) do
				_M.output_video_fmt[k] = v
			end
			_M.output_video_fmt.deInterlace = false
		else
			for k,v in pairs(deinterlace_fmt) do
				_M.output_video_fmt[k] = v
			end
			_M.output_video_fmt.name = fmt.name --Use the original format name (i.e. 1920x1080i 60Hz)
			_M.output_video_fmt.deInterlace = true
		end
	else
		for k,v in pairs(fmt) do
			_M.output_video_fmt[k] = v
		end
		_M.output_video_fmt.deInterlace = false
	end

	if m == "surround" then
		_M.output_video_fmt.id = "custom"
	end
	_M.output_video_fmt.codec = fourCC and fourCC:upper() or "UNKNOWN"
	_M.output_video_fmt.failsafe = failsafe
	if failsafe then
		_M.output_video_fmt.original_xRes = xRes
		_M.output_video_fmt.original_yRes = yRes
		_M.output_video_fmt.original_frame_rate = fr
		_M.output_video_fmt.original_interlaced = interlaced
		_M.output_video_fmt.original_name = (xRes or 0) .. "x" .. (yRes or 0) .. (interlaced and "i " or "p ") .. (fr or "0") .. "Hz"
	end
	_M.last_video_changed = true
	g_event:Notify( "decoder", {SN=sn, action="RES", resInfo=_M.output_video_fmt} )
end

local function onRetrivalAudioFormat(sn, sample_rate, no_channels, no_samples, restrict )
	local set_sample_rate
	local set_no_channels
	local set_no_samples = 768 --TODO: Now fixed.
	if _M.output_audio_set and _M.output_audio_set.sample_rate and _M.output_audio_set.sample_rate > 0 then
		set_sample_rate = _M.output_audio_set.sample_rate
	else
		set_sample_rate = (sample_rate and sample_rate > 0) and sample_rate or 48000
	end

	if _M.output_audio_set and _M.output_audio_set.no_channels and _M.output_audio_set.no_channels > 0 then
		set_no_channels = _M.output_audio_set.no_channels
	else
		set_no_channels = (no_channels and no_channels > 0) and no_channels or 2
	end

	_M.output_audio_set = {
		sample_rate = set_sample_rate,
		no_channels = set_no_channels,
		no_samples = set_no_samples
	}

	_M.last_audio_changed = true

	g_event:Notify( "decoder", {SN=sn, action="AUDIO", audioSetting=_M.output_audio_set} )
end

local function onAVFormatRetrival( topic, msg, msgType )
	if msgType ~= "table" then
		_dbg.Warn( "AV-FORMAT", "Got invalid retrival event message of type:", msgType )
		return
	end
	_dbg.Info( "AV-FORMAT", "#", msg.SN, ": action=", msg.action )
	if msg.action == "RES" and type(msg.args) == "table" then
		onRetrivalVideoFormat(msg.SN, msg.args.fourCC, msg.args.xRes, msg.args.yRes, msg.args.frN, msg.args.frD, msg.args.interlaced, msg.args.restrict)
	elseif msg.action == "AUDIO" and type(msg.args) == "table" then
		onRetrivalAudioFormat(msg.SN, msg.args.sample_rate, msg.args.no_channels, msg.args.no_samples, msg.args.restrict)
	else
		_dbg.Warn( "AV-FORMAT", "Invalid action or invalid arguments!" )
	end
end

--=====================================================
local function control_sii9136_reset(val)
	local rst_9136 = GPIO( 983, (not val or val == 0) and "low" or "high" )
	rst_9136:write( val and val ~= 0 )
	rst_9136:close()
end

--=====================================================

function _M.Init(work_mode)
	local reset_pin = GPIO( 915, "high" )
	reset_pin:write(false)
	coaf.Sleep(0.5)
	reset_pin:write(true)
	reset_pin:close()
	coaf.Sleep(0.5)

	control_sii9136_reset(0)
	coaf.Sleep(0.5)
	control_sii9136_reset(1)

	coaf.Sleep(1) --TODO: here? Why default 4?

	_M.i_aic3104 = chip_aic3104()

	if work_mode ~= "decoding" then
		_M.i_hdmi_rx = chip_ite66021()
	else
		_M.work_mode = "decoding"
		require "ndi_decoding"
		g_event:Subscribe( "decoder", onAVFormatRetrival )
	end

	_M.i_hdmi_tx = chip_sii9136("/dev/i2c-1", control_sii9136_reset)
end

function _M.Check()
	if _M.i_hdmi_rx then
		_M.i_hdmi_rx:Step()
	end

	_M.i_hdmi_tx:Step()
	_M.i_aic3104:Step()

	if _M.work_mode == "decoding" then
		if _M.last_video_changed then
			_M.last_video_changed = nil
			if _M.output_video_fmt then
				
				require "ndi_decoding"

				local colorSpace = "yuv422"
				local video_mode = "hd"
				local format_type = "CEA"
				if _M.output_video_fmt.format_mode == "UHD_420" then
					colorSpace = "yuv420"
					video_mode = "4k60"
					format_type = "HDMI"
				elseif _M.output_video_fmt.format_mode == "UHD" then
					video_mode = "4k"
					format_type = "HDMI"
				elseif _M.output_video_fmt.format_mode == "SD" then
					video_mode = "sd"
				end
				if _M.output_video_fmt.vicCode == 0 then
					format_type = "PC"
				end
				if _M.output_video_fmt.deInterlace then
					NDI.DECODING.write_reg( 0x280, 0x0e, 0x80 )
				else
					NDI.DECODING.write_reg( 0x280, 0x0e, 0x00 )
				end
				_M.i_hdmi_tx:UpdateVideoFormat( _M.output_video_fmt.vicCode, _M.output_video_fmt.vicCode, colorSpace, video_mode ) --, format_type )
			end
		end

		if _M.last_audio_changed then
			_M.last_audio_changed = nil
			if _M.output_audio_fmt then
				_M.i_hdmi_tx:UpdateAudioFormat( _M.output_audio_fmt.sample_rate/1000 )
			end
		end
	else
		if _M.i_hdmi_rx.videoFormat.version ~= _M.cur_video_ver then
			--TODO: 检查输入芯片的 CEA-ID；确定色彩空间，尤其是4Kp60时候的情况。
			--      注意：_M.i_hdmi_rx.videoFormat.format_code 表示视频格式
			--            _M.i_hdmi_rx.videoFormat.video_mode 对应 "sd"/"hd"/"4k"/"4kp60" 可用于更进一步的判定
			--            _M.i_hdmi_rx.videoFormat.colorSpace 对应 "default"/"yuv422"/"yuv444"/"rgb444"表示输入的原始色彩空间
			--            _M.i_hdmi_rx.videoFormat.cea_id 对应 从输入芯片获取的 AVI InfoFrame 中的CEA-ID（视频格式ID），这个需要在 chips_n20/ite66021.lua 中完善。
			--
			if _M.i_hdmi_rx.videoFormat._valid then
				--参见：chips_n20/sii9136.lua 中的函数参数说明
				--print("_M.i_hdmi_rx.videoFormat.format_code")
				--print(_M.i_hdmi_rx.videoFormat.format_code)
				--print(_M.i_hdmi_rx.videoFormat.video_mode)
				--print(_M.i_hdmi_rx.videoFormat.colorSpace)
				--print(_M.i_hdmi_rx.videoFormat.cea_id)
				_M.i_hdmi_tx:UpdateVideoFormat( _M.i_hdmi_rx.videoFormat.format_code,_M.i_hdmi_rx.videoFormat.cea_id,
				_M.i_hdmi_rx.videoFormat.colorSpace,_M.i_hdmi_rx.videoFormat.video_mode)
			end
	
			_M.cur_video_ver = _M.i_hdmi_rx.videoFormat.version
			--pt.dump( _M.i_hdmi_rx.videoFormat )
		end
	
		if _M.i_hdmi_rx.audioFormat.version ~= _M.cur_audio_ver then
			if _M.i_hdmi_rx.audioFormat._valid then
				--TODO: 要完善 chips_n20/ite66021.lua 中对于 fs 检测。目前fs恒为256.
				_M.i_hdmi_tx:UpdateAudioFormat( _M.i_hdmi_rx.audioFormat.sampling/1000, _M.i_hdmi_rx.audioFormat.fs )
			end
			_M.cur_audio_ver = _M.i_hdmi_rx.audioFormat.version
		end
	end
end

function _M.GetVideoSignalStatus()
	if not _M.i_hdmi_rx then
		return nil, nil
	end

	local rate_n = _M.i_hdmi_rx.videoFormat._formatInfo and _M.i_hdmi_rx.videoFormat._formatInfo.frameRate and _M.i_hdmi_rx.videoFormat._formatInfo.frameRate*1000
	if rate_n and _M.i_hdmi_rx.videoFormat._formatInfo.interlaced then
		rate_n = math.floor(rate_n/2)
	end

	if rate_n and rate_n <= 1000 then
		rate_n = 30000
	end

	return _M.i_hdmi_rx.videoFormat.version, {
		mode = _M.i_hdmi_rx.videoFormat.video_mode,
		signal = _M.i_hdmi_rx.videoFormat.signal_locked,
		format_name = _M.i_hdmi_rx.videoFormat.format_name,
		width = _M.i_hdmi_rx.videoFormat._formatInfo and _M.i_hdmi_rx.videoFormat._formatInfo.hAct,
		height = _M.i_hdmi_rx.videoFormat._formatInfo and _M.i_hdmi_rx.videoFormat._formatInfo.vAct0,
		hTotal = _M.i_hdmi_rx.videoFormat._formatInfo and _M.i_hdmi_rx.videoFormat._formatInfo.hTotal,
		vTotal = _M.i_hdmi_rx.videoFormat._formatInfo and _M.i_hdmi_rx.videoFormat._formatInfo.vTotal0,
		interlaced = _M.i_hdmi_rx.videoFormat._formatInfo and _M.i_hdmi_rx.videoFormat._formatInfo.interlaced,
		frame_rate = _M.i_hdmi_rx.videoFormat._formatInfo and _M.i_hdmi_rx.videoFormat._formatInfo.frameRate or 0,
		rate_n = rate_n,
		rate_d = 1000 --TODO:
	}
end

function _M.GetAudioSignalStatus()

	if not _M.i_hdmi_rx then
		return nil, nil
	end

	return _M.i_hdmi_rx.audioFormat.version, {
		sampling = _M.i_hdmi_rx.audioFormat.sampling,
		signal = _M.i_hdmi_rx.audioFormat._valid,
		channels = 2 --TODO
	}
end

function _M.EnableLoopback(en)
	--TODO:
end

--Aka.
-- formatInfo is a table:
-- {
--     mode = "dvi"|"hdmi",
--     width = ?, --Now ignored.
--     height = ?, --Now ignored.
--     interlaced = true|false, --Now ignored.
--     frame_rate = 0 ~ N, --0 is auto.
-- }
function _M.SetVideoOutputFormat(formatInfo)
	_M.output_video_set = formatInfo
	NDI.DECODING.update_video()
end

--Return the actual output format info, Aka.
--{
--    mode = "dvi"|"hdmi",
--    width = ?,
--    height = ?,
--    frame_rate = N,
--    interlaced = true|false,
--    format_name = "1920x1080@30Hz", -- for example.
--    error_tag = "error-tag"
--}
function _M.GetVideoOutputFormat()
	if not _M.output_video_fmt then
		return {
			mode = _M.output_video_set and _M.output_video_set.mode or "hdmi",
			width = 0,
			height = 0,
			frame_rate = 0,
			interlaced = false,
			format_name = "Invalid",
			codec = "UNKNOWN",
			deInterlace = false,
			error_tag = "invalid"
		}
	else
		return {
			mode = _M.output_video_set and _M.output_video_set.mode or "hdmi",
			width = _M.output_video_fmt.failsafe and _M.output_video_fmt.original_xRes or _M.output_video_fmt.xRes,
			height = _M.output_video_fmt.failsafe and _M.output_video_fmt.original_yRes or _M.output_video_fmt.yRes,
			frame_rate = _M.output_video_fmt.failsafe and _M.output_video_fmt.original_frame_rate or _M.output_video_fmt.frame_rate,
			interlaced = _M.output_video_fmt.failsafe and _M.output_video_fmt.original_interlaced or _M.output_video_fmt.interlaced,
			format_name = _M.output_video_fmt.failsafe and _M.output_video_fmt.original_name or _M.output_video_fmt.name,
			codec = _M.output_video_fmt.codec,
			deInterlace = _M.output_video_fmt.deInterlace,
			error_tag = _M.output_video_err
		}
	end
end

--Aka.
-- formatInfo is a table:
-- {
--     sample_rate = 0 ~ N, --0 is auto
--     no_channels = 0 ~ N, --0 is auto
-- }
function _M.SetAudioOutputFormat(formatInfo)
	_M.output_audio_set = formatInfo
	NDI.DECODING.update_audio()
end

--Return the actual audio output format info, Aka.
--{
--     sample_rate = N,
--     no_channels = N
--}
function _M.GetAudioOutputFormat()
	if not _M.output_audio_fmt then
		return {
			sample_rate = 48000,
			no_channels = 2,
			no_samples = 768
		}
	else
		return _M.output_audio_fmt
	end
end

return _M

