-----------------------------------------------------------------------------------------------------
-- ITE68051 chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2019/3/18
-----------------------------------------------------------------------------------------------------

--package.path = '/boot/backup/?.lua;/usr/local/bin/codec/?.lua'
local coaf=require "coaf.core"
local _dbg = require "coaf.debug"

local periphery = require "periphery"
local I2C_REG = require "chips.i2c_reg"
--For debug only.
--local pretty = require "pl.pretty"

local MAIN_FREQ = 27000000
local OSC_TYPICAL_VALUE = 4300
local RCLKVALUE = 0
local hdtv_formats = require "chips.hdtv_formats"

hdtv_formats.SetMainFreq( MAIN_FREQ )

--Init both HDTV and VGA format standards
hdtv_formats.Init( "STDI" )

------------------------------------------------------
-- The factory function to create a ITE68051 handle object
------------------------------------------------------
local function CreateITE68051( i2c_dev, rxAddr )
	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local ITE68051 = {
		currentSource = "HDMI",
		currentSignal = false,
		freeRun = false,
		currentFormat = nil,
		forceReassignFormat = false,
		lastFormatRecord = {},
		videoinfo={pix_saved=0,pixclk_cnt=0,pixclk_last=0},
		videoFormat = { _valid = false },
		audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }, --type 0 aka. HDMI
		orgAudioFormat = { _valid = false },
		nostandardFormatInfo = {},
		i2c_reg = nil,
	}

	------------------------------------------------------
	i2c_dev = i2c_dev or "/dev/i2c-1"
	
	local BASE_MAP = rxAddr or 0x48
	local EDID_MAP = 0x54
	local MHL_ADDR = 0X64

	--_dbg.Info("ITE68051", "Creating ITE68051 control object at 0x" .. string.format( "%02x", BASE_MAP) .. "@" .. i2c_dev .. "..." )

	ITE68051.i2c_reg = I2C_REG.Create()
	if not ITE68051.i2c_reg:Open( i2c_dev ) then
		return nil, "Fail to open I2C device"
	end

	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function ITE68051:ReadReg( map, addr, itemArr )
		return self.i2c_reg:ReadReg(map, addr, itemArr)
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function ITE68051:WriteReg( map, addr, itemArr )
		return self.i2c_reg:WriteReg(map, addr, itemArr)
	end
	
	------------------------------------------------------
	-- Register write with mask
	------------------------------------------------------
	function ITE68051:hdmirxset( map, addr,mask,itemArr)
		local temp1 = self.i2c_reg:ReadReg(map, addr, {0x00})
		local temp2 = {(temp1[1]&(~mask))| itemArr[1]}
		return self.i2c_reg:WriteReg(map, addr, temp2)
	end	

	------------------------------------------------------
	
	function ITE68051:ForceUpdate()
		self.lastFormatRecord = {}		--Clear the recorded format information.
	end


	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------
	function ITE68051:Reset()
	
		--print("Reset_ite68051")
		self:_InitDevice()		
		self:_Cal_oclk()
		self:_InitInterrupts()
		self:_initcscmatrix()
		self:_InitEDID()	--Init EDID
		self:_ConfigureHDMIFormat()	--Force default blue
		
		self.currentSignal = false
		self.freeRun = false
		self.currentFormat = nil
		self.forceReassignFormat = false

		self.lastFormatRecord = {}		--Clear recorded format detection informations.
		self.audioFormat = { _valid = false, audio_src = "HDMI", audio_signal_type = 0 }
		self.orgAudioFormat = { _valid = false }
		self.videoFormat = {_valid = false}
	end
	------------------------------------------------------
	-- ITE68051 Calibration Ring Oscillator
	------------------------------------------------------	
	function ITE68051:_Cal_oclk()
		local rddata,rddata2
		local OSCCLK
		local CPOCLK
		local RCLKFreqSel = 0
		local Read_Blcok
		self:WriteReg( MHL_ADDR, 0x01, {0x41} )
		periphery.sleep_ms(99) --Wait 5ms for Calc CRC
		self:WriteReg( MHL_ADDR, 0x01, {0x40} )
		local rslt = self:ReadReg( MHL_ADDR, 0x12, {0x00,0x00,0x00} )
		rddata	= (rslt[3]<< 16) + (rslt[2] << 8) +rslt[1]
		OSCCLK = rddata / 100			
		
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0xF8,{0xC3} )
		self:WriteReg( BASE_MAP,0xF8,{0xA5} )
		self:WriteReg( BASE_MAP,0x34,{0x00} )
		self:WriteReg( BASE_MAP,0x0F,{0x01} )
		self:WriteReg( BASE_MAP,0x5F,{0x04} )
		self:WriteReg( BASE_MAP,0x5F,{0x05} )
		self:WriteReg( BASE_MAP,0x58,{0x12} )
		self:WriteReg( BASE_MAP,0x58,{0x02} )
		self:WriteReg( BASE_MAP,0x57,{0x01} )
		
		self:WriteReg( BASE_MAP,0x50,{0x00} )
		self:WriteReg( BASE_MAP,0x51,{0x00} )
		self:WriteReg( BASE_MAP,0x54,{0x04} )		
		local temp1 = self:ReadReg( BASE_MAP, 0x61, {0x00,0x00} )
		
		self:WriteReg( BASE_MAP,0x50,{0x00} )
		self:WriteReg( BASE_MAP,0x51,{0x01} )
		self:WriteReg( BASE_MAP,0x54,{0x04} )
		local temp2 = self:ReadReg( BASE_MAP, 0x61, {0x00,0x00} )
		
		if temp1[1]==0xFF and temp1[2]==0xFF and temp2[1]==0x00 and temp2[2]==0x00 then
			Read_Blcok = 0x04
		else
			Read_Blcok = 0x00
		end

		self:WriteReg( BASE_MAP,0x50,{Read_Blcok} )
		self:WriteReg( BASE_MAP,0x51,{0xb0} )
		self:WriteReg( BASE_MAP,0x54,{0x04} )		
		temp1 = self:ReadReg( BASE_MAP, 0x61, {0x00,0x00} )	

		self:WriteReg( BASE_MAP,0x50,{Read_Blcok} )
		self:WriteReg( BASE_MAP,0x51,{0xb1} )
		self:WriteReg( BASE_MAP,0x54,{0x04} )
		temp2 = self:ReadReg( BASE_MAP, 0x61, {0x00,0x00} )	
		CPOCLK = temp1[1] + (temp1[2] << 8) + (temp2[1] << 16)
		if (temp2[2]&0xc0) == 0xc0 then
			CPOCLK = CPOCLK/100
		end
		if CPOCLK > 70000 or CPOCLK < 25000 then
			CPOCLK	= OSC_TYPICAL_VALUE
		end
		self:WriteReg( BASE_MAP,0x5F,{0x00} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0xF8,{0x00} )
		rddata2 = (48000 * 1000/CPOCLK) * 64;

		self:WriteReg( BASE_MAP,0x0F,{0x03} )
		--#ifdef EVB_BOARD_FOR_RING_FREQ
		--CPOCLK = OSCCLK;
		--#endif
		local AA = 0x0C
		if CPOCLK <  31984 then
			AA = 0x1F
		elseif CPOCLK >= 31984 and CPOCLK <  32586 then 
			AA = 0x1E
		elseif CPOCLK >= 32586 and CPOCLK <  33155 then
			AA = 0x1D
		elseif CPOCLK >= 33155 and CPOCLK <  33293 then
			AA = 0x1C
		elseif CPOCLK >= 33293 and CPOCLK <  34274 then
			AA = 0x1B
		elseif CPOCLK >= 34274 and CPOCLK <  35006 then
			AA = 0x1A
		elseif CPOCLK >= 35006 and CPOCLK <  35709 then
			AA = 0x19
		elseif CPOCLK >= 35709 and CPOCLK <  36119 then
			AA = 0x18
		elseif CPOCLK >= 36119 and CPOCLK <  36890 then
			AA = 0x17
		elseif CPOCLK >= 36890 and CPOCLK <  37798 then
			AA = 0x16
		elseif CPOCLK >= 37798 and CPOCLK <  38682 then
			AA = 0x15
		elseif CPOCLK >= 38682 and CPOCLK <  39437 then
			AA = 0x14
		elseif CPOCLK >= 39437 and CPOCLK <  40434 then
			AA = 0x13
		elseif CPOCLK >= 40434 and CPOCLK <  41618 then
			AA = 0x12
		elseif CPOCLK >= 41618 and CPOCLK <  42778 then
			AA = 0x11
		elseif CPOCLK >= 42778 and CPOCLK <  43545 then
			AA = 0x10
		elseif CPOCLK >= 43545 and CPOCLK <  44858 then
			AA = 0x0F
		elseif CPOCLK >= 44858 and CPOCLK <  46443 then
			AA = 0x0E
		elseif CPOCLK >= 46443 and CPOCLK <  48024 then
			AA = 0x0D
		elseif CPOCLK >= 48024 and CPOCLK <  49408 then
			AA = 0x0C
		elseif CPOCLK >= 49408 and CPOCLK <  51270 then
			AA = 0x0B
		elseif CPOCLK >= 51270 and CPOCLK <  53574 then
			AA = 0x0A
		elseif CPOCLK >= 53574 and CPOCLK <  55916 then
			AA = 0x09
		elseif CPOCLK >= 55916 and CPOCLK <  57284 then
			AA = 0x08
		elseif CPOCLK >= 57284 and CPOCLK <  60101 then
			AA = 0x07
		elseif CPOCLK >= 60101 and CPOCLK <  63649 then
			AA = 0x06
		elseif CPOCLK >= 63649 and CPOCLK <  67405 then
			AA = 0x05
		elseif CPOCLK >= 67405 and CPOCLK <  70839 then
			AA = 0x04
		elseif CPOCLK >= 70839 and CPOCLK <  72851 then
			AA = 0x03
		elseif CPOCLK >= 72851 and CPOCLK <  82503 then
			AA = 0x02
		elseif CPOCLK >= 82503 and CPOCLK <  90053 then
			AA = 0x01
		else
			AA = 0x00
		end 
		AA = AA + 1
		
		self:hdmirxset(BASE_MAP,0xAA,0x1f,{AA})
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		
		self:WriteReg( MHL_ADDR, 0x01, {0x41} )
		periphery.sleep_ms(99) --Wait 5ms for Calc CRC
		self:WriteReg( MHL_ADDR, 0x01, {0x40} )
		rslt = self:ReadReg( MHL_ADDR, 0x12, {0x00,0x00,0x00} )
		rddata	= (rslt[3]<<16) + (rslt[2] <<8) + rslt[1]
		OSCCLK = rddata / 100
		local oscdiv
		if OSCCLK > 49000 then
			oscdiv = 5
		else
			oscdiv = 4
		end
		OCLK = OSCCLK / oscdiv
		self:hdmirxset(MHL_ADDR,0x01,0x70,{(oscdiv<<4)})

		if RCLKFreqSel == 0 then
			RCLKVALUE = OSCCLK / 2		-- 20 MHz
		elseif RCLKFreqSel == 1 then
			RCLKVALUE = OSCCLK / 4   	--10 MHz
		elseif RCLKFreqSel == 2 then
			RCLKVALUE = OSCCLK / 8
		else
			RCLKVALUE = OSCCLK / 16
		end
		RCLKVALUE=RCLKVALUE*1.1    -- 20170327 MHL CTS
		local t1usint = math.floor(RCLKVALUE / 1000)
		local t1usflt = math.floor((RCLKVALUE % 1000) * 256 / 1000)	
		self:hdmirxset(BASE_MAP,0x91,0x3F,{t1usint&0x3F})
		self:WriteReg( BASE_MAP, 0x92, {t1usflt} )
		RCLKVALUE = RCLKVALUE/1.1    -- 20170327 MHL CTS add for CTS (RCLK*=1.1)	
		
	end	
	
	------------------------------------------------------
	-- Init the ITE68051 Device modify ok
	------------------------------------------------------
	
	function ITE68051:_InitDevice()
	
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x22,{0x08} )
		self:WriteReg( BASE_MAP,0x22,{0x17} )
		self:WriteReg( BASE_MAP,0x23,{0x1F} )
		self:WriteReg( BASE_MAP,0x2B,{0x1F} )
		self:WriteReg( BASE_MAP,0x24,{0xF8} )
		self:WriteReg( BASE_MAP,0x22,{0x30} )
		self:WriteReg( BASE_MAP,0x23,{0xA0} )
		self:WriteReg( BASE_MAP,0x2B,{0xA0} )
		self:WriteReg( BASE_MAP,0x24,{0x00} )
		self:WriteReg( BASE_MAP,0x34,{0x00} )
		self:WriteReg( BASE_MAP,0x0F,{0x03} )
		self:WriteReg( BASE_MAP,0xAA,{0xEC} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x0F,{0x03} )
		self:WriteReg( BASE_MAP,0xAC,{0x40} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x49,{0xC8|0x01})  -- config MHL_ADDR 0X64 (0x64=0xC8>>1)
		self:WriteReg( BASE_MAP,0x4B,{0xA8|0x01})  -- config EDID_MAP 0X54 (0x54=0xA8>>1)
		self:WriteReg( BASE_MAP,0x43,{0x01} )
		self:WriteReg( BASE_MAP,0x0F,{0x04} )
		self:WriteReg( BASE_MAP,0x3A,{0x89} )
		self:WriteReg( BASE_MAP,0x0F,{0x03} )
		self:WriteReg( BASE_MAP,0xA8,{0x0B} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x4F,{0x84} ) 	  --config RCLKFreqSel=00 and audio fs = 256
		self:WriteReg( BASE_MAP,0x44,{0x19} )
		self:WriteReg( BASE_MAP,0x46,{0x15} )
		self:WriteReg( BASE_MAP,0x47,{0x88} )
		self:WriteReg( BASE_MAP,0xD9,{0x02} )
		self:WriteReg( BASE_MAP,0xF0,{0x78} )
		self:WriteReg( BASE_MAP,0xF1,{0x10} )
		self:WriteReg( BASE_MAP,0x0F,{0x03} )
		self:WriteReg( BASE_MAP,0x3A,{0x02} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x28,{0x08} )
		self:WriteReg( BASE_MAP,0x6E,{0x80} )
		self:WriteReg( BASE_MAP,0x77,{0x83} )
		self:WriteReg( BASE_MAP,0x7B,{0x00} )
		self:WriteReg( BASE_MAP,0x86,{0x00} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x36,{0x06} )
		self:WriteReg( BASE_MAP,0x8F,{0x41} )
		self:WriteReg( BASE_MAP,0x0F,{0x01} )
		self:WriteReg( BASE_MAP,0xC0,{0x42} )
		self:WriteReg( BASE_MAP,0xC4,{0X20} ) --Clock Phase Adjusting InVert:7 Delay 6:4 0x00
		self:WriteReg( BASE_MAP,0xC5,{0x00} )
		self:WriteReg( BASE_MAP,0xC6,{0x00} )
		self:WriteReg( BASE_MAP,0xC7,{0x00} )
		self:WriteReg( BASE_MAP,0xC8,{0x00} )
		self:WriteReg( BASE_MAP,0xC9,{0x11} )
		self:WriteReg( BASE_MAP,0xCA,{0x99} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x86,{0x08} )
		self:WriteReg( BASE_MAP,0x81,{0x80} )
		self:WriteReg( BASE_MAP,0x0F,{0x01} )
		self:WriteReg( BASE_MAP,0x10,{0x00} )
		self:WriteReg( BASE_MAP,0x11,{0x00} )
		self:WriteReg( BASE_MAP,0x12,{0x00} )
		self:WriteReg( BASE_MAP,0x13,{0x00} )
		self:WriteReg( BASE_MAP,0x28,{0x00} )
		self:WriteReg( BASE_MAP,0x29,{0x00} )
		self:WriteReg( BASE_MAP,0x2A,{0x00} )
		self:WriteReg( BASE_MAP,0x2B,{0x00} )
		self:WriteReg( BASE_MAP,0x2C,{0x00} )
		self:WriteReg( BASE_MAP,0xC0,{0X03} ) --BIT0: 0: Single pixel mode, 1: Dual pixel mode  BIT1: 1: Enable TTL output module clock
		self:WriteReg( BASE_MAP,0x0F,{0x03} )
		self:WriteReg( BASE_MAP,0xE3,{0x05} )
		self:WriteReg( BASE_MAP,0x27,{0x9F} ) --  EQ_INIT_VALUE
		self:WriteReg( BASE_MAP,0x28,{0x9F} ) --  EQ_INIT_VALUE
		self:WriteReg( BASE_MAP,0x29,{0x9F} ) --  EQ_INIT_VALUE
		local temp = self:ReadReg( BASE_MAP, 0xA7, {0x00} )
		self:WriteReg( BASE_MAP,0xA7,{0x40|temp[1]} )
		--self:WriteReg( BASE_MAP,0xA7 0x40 0x40
		self:WriteReg( BASE_MAP,0x0F,{0x07} )  
		self:WriteReg( BASE_MAP,0xE3,{0x05} )  
		self:WriteReg( BASE_MAP,0x27,{0x9F} ) --  EQ_INIT_VALUE
		self:WriteReg( BASE_MAP,0x28,{0x9F} ) --  EQ_INIT_VALUE
		self:WriteReg( BASE_MAP,0x29,{0x9F} ) --  EQ_INIT_VALUE
		--self:WriteReg( BASE_MAP,0xA7 0x40 0x40
		temp = self:ReadReg( BASE_MAP, 0xA7, {0x00} )
		self:WriteReg( BASE_MAP,0xA7,{0x40|temp[1]} )		
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		--self:WriteReg( BASE_MAP,0x21 0x04 0x00
		local temp = self:ReadReg( BASE_MAP, 0x21, {0x00} )
		self:WriteReg( BASE_MAP,0xA7,{0xbf&temp[1]} )		
		self:WriteReg( BASE_MAP,0x4A,{0xC8} ) --#define ADDR_CEC 0xC8  ???
		self:WriteReg( BASE_MAP,0xF8,{0xC3} )
		self:WriteReg( BASE_MAP,0xF8,{0xA5} )
		self:WriteReg( BASE_MAP,0x0F,{0x01} )
		self:WriteReg( BASE_MAP,0x5F,{0x04} )
		self:WriteReg( BASE_MAP,0x58,{0x12} )
		self:WriteReg( BASE_MAP,0x58,{0x02} )
		self:WriteReg( BASE_MAP,0x5F,{0x00} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0xF8,{0xFF} )	
		self:WriteReg( BASE_MAP,0x6B,{0x44} )  -- out color-space YUV422 BIT WIDTH :8 bit		
		--self:WriteReg( BASE_MAP,0xC5,{0x02} )  -- out color-space YUV422 BIT WIDTH :8 bit		
	end	
	------------------------------------------------------
	-- Init the ITE68051 interrupts 
	------------------------------------------------------
	
	function ITE68051:_InitInterrupts()
		--TODO: Nothing yet.
	end

	function ITE68051:_initcscmatrix()

		--for RGB TO yuv422 limit(16-235)
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
		self:WriteReg( BASE_MAP,0x6c,0x03,{0x02} )
		self:WriteReg( BASE_MAP,0x0F,{0x01} )
		self:WriteReg( BASE_MAP,0x86,0x0f,{0x00} )
		self:WriteReg( BASE_MAP,0x85,{0x00} )
		self:WriteReg( BASE_MAP,0x70,{0x10} )
		self:WriteReg( BASE_MAP,0x71,{0x80} )
		self:WriteReg( BASE_MAP,0x72,{0x10} )
		self:WriteReg( BASE_MAP,0x73,{0xe4} )
		self:WriteReg( BASE_MAP,0x74,{0x04} )
		self:WriteReg( BASE_MAP,0x75,{0x77} )
		self:WriteReg( BASE_MAP,0x76,{0x01} )
		self:WriteReg( BASE_MAP,0x77,{0x7f} )
		self:WriteReg( BASE_MAP,0x78,{0x00} )
		self:WriteReg( BASE_MAP,0x79,{0xd0} )
		self:WriteReg( BASE_MAP,0x7a,{0x3c} )
		self:WriteReg( BASE_MAP,0x7b,{0x83} )
		self:WriteReg( BASE_MAP,0x7c,{0x03} )
		self:WriteReg( BASE_MAP,0x7d,{0xad} )
		self:WriteReg( BASE_MAP,0x7e,{0x3f} )
		self:WriteReg( BASE_MAP,0x7f,{0x4b} )
		self:WriteReg( BASE_MAP,0x80,{0x3d} )
		self:WriteReg( BASE_MAP,0x81,{0x32} )
		self:WriteReg( BASE_MAP,0x82,{0x3f} )
		self:WriteReg( BASE_MAP,0x83,{0x84} )
		self:WriteReg( BASE_MAP,0x84,{0x03} )
		self:WriteReg( BASE_MAP,0x0F,{0x00} )
	--
--[[
	-- for RGB TO yuv422 full(0-255)
	self:WriteReg( BASE_MAP,0x0F,{0x00} )
	self:WriteReg( BASE_MAP,0x6c,0x03,{0x02} )
	self:WriteReg( BASE_MAP,0x6e,0xf0,{0x60} )
	self:WriteReg( BASE_MAP,0x0F,{0x01} )
	self:WriteReg( BASE_MAP,0x86,0x0f,{0x00} )
	self:WriteReg( BASE_MAP,0x85,{0x00} )
	self:WriteReg( BASE_MAP,0x70,{0x00} )
	self:WriteReg( BASE_MAP,0x71,{0x80} )
	self:WriteReg( BASE_MAP,0x72,{0x10} )
	self:WriteReg( BASE_MAP,0x73,{0xb8} )
	self:WriteReg( BASE_MAP,0x74,{0x05} )
	self:WriteReg( BASE_MAP,0x75,{0xb4} )
	self:WriteReg( BASE_MAP,0x76,{0x01} )
	self:WriteReg( BASE_MAP,0x77,{0x94} )
	self:WriteReg( BASE_MAP,0x78,{0x00} )
	self:WriteReg( BASE_MAP,0x79,{0x4a} )
	self:WriteReg( BASE_MAP,0x7a,{0x3c} )
	self:WriteReg( BASE_MAP,0x7b,{0x17} )
	self:WriteReg( BASE_MAP,0x7c,{0x04} )
	self:WriteReg( BASE_MAP,0x7d,{0x9f} )
	self:WriteReg( BASE_MAP,0x7e,{0x3f} )
	self:WriteReg( BASE_MAP,0x7f,{0xd9} )
	self:WriteReg( BASE_MAP,0x80,{0x3c} )
	self:WriteReg( BASE_MAP,0x81,{0x10} )
	self:WriteReg( BASE_MAP,0x82,{0x3f} )
	self:WriteReg( BASE_MAP,0x83,{0x17} )
	self:WriteReg( BASE_MAP,0x84,{0x04} )
	self:WriteReg( BASE_MAP,0x0F,{0x00} )

--]]		
	end
	
	------------------------------------------------------
	--EDID configuration 
	------------------------------------------------------
	function ITE68051:_InitEDID()
		local edid 
		if _G.DISABLE_4Kp60 then
			edid = require "chips.edid"	--EDID
		else
			edid = require "chips.edid2_uhd60"
		end
		self:WriteReg( BASE_MAP, 0x0F, {0x03} )
		self:WriteReg( BASE_MAP, 0xAB, {0x4A} ) 	--hpd out low 
		self:WriteReg( BASE_MAP, 0x0F, {0x00} )	

		local i
		for i=1, #edid, 8 do
			self:WriteReg( EDID_MAP, i-1, { edid[i], edid[i+1], edid[i+2], edid[i+3], edid[i+4], edid[i+5], edid[i+6], edid[i+7]} )
		end
		self:WriteReg( BASE_MAP, 0xc6, {0x9e} )		--VSDB address of EDID
		self:WriteReg( BASE_MAP, 0xc7, {0x10} )		--Physical address 1 of Port 0
		self:WriteReg( BASE_MAP, 0xC8, {0x00} )		--Physical address 2 of Port 0
		self:WriteReg( BASE_MAP, 0xC9, {edid[128]} )	--Block 0 Check sum of Port 0
		self:WriteReg( BASE_MAP, 0xCA, {edid[256]} )	--Block 1 Check sum of Port 0
		self:hdmirxset( BASE_MAP, 0xC5, 0x01, {0x00} )	--enable PORT0 internal EDID	
		--self:WriteReg( BASE_MAP, 0xC5, {0x02} )  		--enable PORT0 internal EDID
		--come from 68051 for reset EDID start
		self:hdmirxset( BASE_MAP, 0xC5, 0x10, {0x10} )	--reset PORT0
		periphery.sleep_ms(1) --Wait 5ms for Calc CRC
		self:hdmirxset( BASE_MAP, 0xC5, 0x10, {0x00} )
		-- control the hpd out high
		self:WriteReg( BASE_MAP, 0x0F, {0x03} )
		self:WriteReg( BASE_MAP, 0xAB, {0xCA} ) 	-- hpd out high
		self:WriteReg( BASE_MAP, 0x0F, {0x00} )
		periphery.sleep_ms(5) --Wait 5ms for Calc CRC
	end
	------------------------------------------------------
	--check the input signal is valid
	------------------------------------------------------	
	function ITE68051:_CheckSignal(src)
		local freeRun = false
		local rslt = self:ReadReg( BASE_MAP, 0x13, {0x00} )
		if rslt then
			if (rslt[1] & 0xad) == 0xad then			-- input stable
				rslt = self:ReadReg( BASE_MAP, 0x98, {0x00} )
				if rslt and (rslt[1]&0x08) ~= 0 then
					return true, freeRun
				else
					return false, freeRun
				end
			else
				return false, freeRun
			end
		else
			return false, freeRun
		end
	end

	------------------------------------------------------
	-- A step handler for ITE68051 processing.
	-- To fit the LUA working mode, you need call the 
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	------------------------------------------------------
	
	function ITE68051:Step()
		self.currentSignal, self.freeRun = self:_CheckSignal()	--Check current signal

		--TODO: consider free-run but now not support.
		if self.currentSignal ~= self.videoFormat.signal_locked then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end
		end

		if not self.audioFormat.version then
			self.audioFormat.version = 0
		end

		if self.currentSignal then
			self.videoFormat.signal_locked = true	--Assume it be, but when detecting, it may changed
			self.videoFormat.free_run = false
			self:_DetectCurrentFormat()
			self:_ConfigureCurrentFormat()
		elseif self.freeRun then
			self.videoFormat.signal_locked = true
			self.videoFormat.free_run = true
		else
			self.videoFormat.signal_locked = false
			self.videoFormat.free_run = false
		end
	end


	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------
	
	function ITE68051:_DetectCurrentFormat(src)
		self:_DetectHDMIAudio(src)
		self:_UpdateAudioFormatInfo()
		return self:_DetectHDMIFormat(src)
	end
	
	function ITE68051:_UpdateAudioFormatInfo()
		if self.audioFormat._valid ~= self.orgAudioFormat.withAudio or
			self.audioFormat.sampling ~= self.orgAudioFormat.sample then

			self.audioFormat.version = self.audioFormat.version + 1

			self.audioFormat._valid = self.orgAudioFormat.withAudio
			self.audioFormat.fs = self.orgAudioFormat.fs
			self.audioFormat.sampling = self.orgAudioFormat.sample
			self.audioFormat.channels = 2
			self.audioFormat.bit_width = 16
		end
	end

	------------------------------------------------------
	-- Configure signal format
	------------------------------------------------------
	
	function ITE68051:_ConfigureCurrentFormat(src)
		if not self.videoFormat._lastFormatChanged and not self.forceReassignFormat then
			return true
		end
		
		return self:_ConfigureHDMIFormat(src)
	end
	
	------------------------------------------------------
	-- According detected format information, record it to Amba based formation context.
	------------------------------------------------------
	
	function ITE68051:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end
		--pretty.dump( formatInfo )
		
		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			if not self.videoFormat.version then --Keep track the updated version
				self.videoFormat.version = 0
			else
				self.videoFormat.version = self.videoFormat.version + 1
			end

			self.videoFormat._lastFormatChanged = true
			self.videoFormat._formatInfo = formatInfo
			
			if src then
				self.videoFormat.video_src = src
			end
			
			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (formatInfo.hAct .. "x" .. formatInfo.vAct0)
			if formatInfo.rateMode == hdtv_formats.VIRATE_4K then
				self.videoFormat.video_mode = "4k"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_4K60 then
				self.videoFormat.video_mode = "4k60"
			elseif formatInfo.rateMode == hdtv_formats.VIRATE_SD then
				self.videoFormat.video_mode = "sd"
			else
				self.videoFormat.video_mode = "hd"
			end

			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	-- Detect HDMI Audio
	------------------------------------------------------
	function ITE68051:_DetectHDMIAudio(src)
		self.orgAudioFormat.withAudio = false
		local rslt

		rslt = self:ReadReg( BASE_MAP, 0x10, {0x00} )  
		if (rslt[1] & 0xf9)  ~= 0x00 then				-- if have error flag ,Reset the audio  [7] Audio FIFO error Interrup [6]Audio Auto Mute Interrupt
			self:WriteReg( BASE_MAP, 0x22, {0x33} )
			self:WriteReg( BASE_MAP, 0x10, {rslt[1] & 0xff} )		-- clear the error flag 
			periphery.sleep_ms(2)
			self:WriteReg( BASE_MAP, 0x22, {0x30} )
			--self:WriteReg( BASE_MAP, 0x10, {rslt[1] & 0xc0} )		-- clear the error flag 
		end
		local rslt = self:ReadReg( BASE_MAP, 0x13, {0x00} )
		if rslt then
			if (rslt[1] & 0x2c) == 0x2c then
				self.orgAudioFormat.pll_locked = true
			else
				self.orgAudioFormat.pll_locked = false
				return false
			end
		else
			self.orgAudioFormat.pll_locked = false
			return false
		end

		if (rslt[1] & 0x02) == 0 then --DVI mode
			return false
		end

		rslt = self:ReadReg( BASE_MAP, 0xb1, {0x00} )
		if not rslt or (rslt[1]&0x80) == 0 then --Audio ON flag is not set.
			return false
		end

		rslt = self:ReadReg( BASE_MAP, 0xb5, {0x00,0x00} )
		if not rslt then
			return false
		end

		local sampleFlag = rslt[1] & 0x0f
		if sampleFlag == 0 then
			self.orgAudioFormat.sample = 44100
		elseif sampleFlag == 2 then
			self.orgAudioFormat.sample = 48000
		elseif sampleFlag == 3 then
			self.orgAudioFormat.sample = 32000
		elseif sampleFlag == 8 then
			self.orgAudioFormat.sample = 88200
		elseif sampleFlag == 9 then
			self.orgAudioFormat.sample = 768000
		elseif sampleFlag == 10 then
			self.orgAudioFormat.sample = 96000
		elseif sampleFlag == 12 then
			self.orgAudioFormat.sample = 176400
		elseif sampleFlag == 14 then
			self.orgAudioFormat.sample = 192000
		else
			self.orgAudioFormat.sample = 48000
		end	

		self.orgAudioFormat.withAudio = true
		--print(self.orgAudioFormat.sample)
		--TODO: Check the fs type from register.
		self.orgAudioFormat.fs = 256 --(rslt[1]+1)*128
		return true
	end

	------------------------------------------------------
	-- Detect HDMI AVI info frame
	------------------------------------------------------
	
	function ITE68051:_DetectHDMIInfoFrame(src)
		--TODO: 从68051的寄存器中读取 AVI InfoFrame 中视频格式的CEA_ID
		--self.videoFormat.cea_id = 0 --TODO:
		--Check the color-space:
		--self:WriteReg( BASE_MAP, 0x0f, {0x02} ) --Read 2xx regs.
		--self.videoFormat.avi_infor = self:ReadReg( BASE_MAP, 0x12, {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0} ) --reg 0x215
		--self:WriteReg( BASE_MAP, 0x0f, {0x00} ) --Reset to 0xx regs.
		--for i=1, 18, 1 do
		--	print(self.videoFormat.avi_infor[i])
		--end			
		--return true
		local cea_id = 255
		local Colorimetry=0 
		local RGBQuantizationRange=0
		local YUVQuantizationRange=3
		--Check the cea_id:
		self:WriteReg( BASE_MAP, 0x0f, {0x02} ) --Read 2xx regs.
		local rslt = self:ReadReg( BASE_MAP, 0x16, {0x00,0x00,0x00,0x00} ) --reg 0x218
		self:WriteReg( BASE_MAP, 0x0f, {0x00} ) --Reset to 0xx regs.	
		--print(rslt[1],rslt[2],rslt[3],rslt[4])
		if rslt and rslt[1] then
			Colorimetry	= ((rslt[1]&0xC0)>>6)			--00:no data 01:ITU601 10:ITU709 11:Extended
		end

		if rslt and rslt[2] then
			RGBQuantizationRange = ((rslt[2]&0x0C)>>2)  --00:Default(depends on video format) 01:Limited Range(16 to 235) 10:Full Range(0 to 255) 11:Reserved
		end		
		
		if rslt and rslt[3] then	--reg 0x218
			cea_id = rslt[3] & 0x7f
		end
		if rslt and rslt[4] then
			YUVQuantizationRange = ((rslt[4]&0xC0)>>6)	--00:Limited Range(16 to 235) 01:Full Range(0 to 255) 11:Reserved (CEA-861-F) others:Reserved
		end							
		if cea_id ~= self.videoFormat.cea_id then
			self.videoFormat.cea_id = cea_id
			self.videoFormat._lastFormatChanged = true
		end
		--print("Colorimetry=",Colorimetry)
		--print("RGBQuantizationRange=",RGBQuantizationRange)
		--print("cea_id=",cea_id)
		--print("YUVQuantizationRange=",YUVQuantizationRange)
		return true	
	end
	
	------------------------------------------------------
	-- Get format
	------------------------------------------------------
	
	function ITE68051:_GetVideoSyncDetail()
		local interlaced = false	
		local diff_pixclk
		local pxclk = 0
		local sump = 0
		local rddata = 0
		local rslt = self:ReadReg( BASE_MAP, 0x98, {0x00} )
		if rslt and (rslt[1]&0x02) ~= 0 then
			interlaced = true
		end

		rslt = self:ReadReg( BASE_MAP, 0x99, {0x00,0x00} )		--PIXCLKSpeed[10:0]
		local pxclkspeed	= ((rslt[2] & 0x07) << 8) + rslt[1]
		--if not rslt or rslt[1] == 0 then
		if not rslt or pxclkspeed == 0 then
			return nil
		end

		for i=1, 5, 1 do
			periphery.sleep_ms(3)
			self:hdmirxset(BASE_MAP,0x9A,0x80,{0x00})
			rslt = self:ReadReg( BASE_MAP, 0x99, {0x00,0x00} )
			rddata = ((rslt[2]&0x07)<<8) + rslt[1]
			self:hdmirxset(BASE_MAP,0x9A,0x80,{0x80})
			sump = sump + rddata
		end	
		
		local pixclk_temp = math.floor(RCLKVALUE * 512 * 5 / sump * 1008)			-- // 512=2*256 because of 1T 2 pixel	)
		rslt = self:ReadReg( BASE_MAP, 0x9B, {0,0,0,0,0,0,0,0,0,0,0,0,0,0} )
		if not rslt then
			return nil
		end

		local HTotal = ((rslt[2] & 0x3F) << 8) + rslt[1]
		local HActive = ((rslt[4] & 0x3F) << 8) + rslt[3]
		--local HSyncWidth = ((rslt[6] & 0x01) << 8) + rslt[5]
		--local HFrontPorch = ((rslt[6] & 0xF0) << 4) + rslt[7]
		--local HBackPorch = HTotal - HActive - HFrontPorch - HSyncWidth
		local VTotal = ((rslt[9] & 0x3F) << 8) + rslt[8]
		local VActive = ((rslt[11] & 0x3F) << 8) + rslt[10]
		--local VSyncWidth = rslt[11] & 0x1F
		--local VFrontPorch = rslt[12] & 0x3F
		--local VBackPorch = VTotal - VActive - VFrontPorch - VSyncWidth
		--local VPolarity = (rslt[13] & 0x08) >> 3
		--local HPolarity = (rslt[13] & 0x04) >> 2

		if self.videoinfo.pixclk_last == 0 and self.videoinfo.pixclk_saved == 0 then 
			self.videoinfo.pixclk_last = pixclk_temp
			self.videoinfo.pixclk_saved = pixclk_temp
		end

		if self.videoinfo.HTotal ~= HTotal or self.videoinfo.HActive ~= HActive or self.videoinfo.VTotal ~= VTotal
			or self.videoinfo.VActive ~= VActive or self.videoinfo.interlaced ~= interlaced then 
		 	self.videoinfo.pixclk_last = pixclk_temp
		 	self.videoinfo.pixclk_saved = pixclk_temp
	 	end		

		if self.videoinfo.pixclk_last >= pixclk_temp then
			diff_pixclk = self.videoinfo.pixclk_last - pixclk_temp
		else
			diff_pixclk = pixclk_temp - self.videoinfo.pixclk_last
		end
		if diff_pixclk >= 1000000 then 
			self.videoinfo.pixclk_cnt = 0
		elseif diff_pixclk < 1000000  and self.videoinfo.pixclk_cnt < 10 then   -- < 1mhz
			self.videoinfo.pixclk_cnt = self.videoinfo.pixclk_cnt + 1
		else
			self.videoinfo.pixclk_saved = pixclk_temp		
		end

		self.videoinfo.pixclk_last = pixclk_temp	
		pxclk = self.videoinfo.pixclk_saved			
		self.videoinfo.HTotal = HTotal
		self.videoinfo.HActive = HActive
		self.videoinfo.VTotal = VTotal
		self.videoinfo.VActive = VActive
		self.videoinfo.interlaced = interlaced
		--print("_GetVideoSyncDetail")
		--print(pxclk, HTotal, HActive, VTotal, VActive, interlaced)
		--print(self.videoFormat.colorSpace)
		--print(self.videoFormat.video_mode)
		return pxclk, HTotal, HActive, VTotal, VActive, interlaced
	end
	
	------------------------------------------------------
	-- Assume HDMI customized format
	------------------------------------------------------
	function ITE68051:_AssumeHDMICustomFormat(src, pxclk, HTotal, actW, VTotal, actH, interlaced)
		local frame_pxs = HTotal * VTotal
		if frame_pxs <= 0 then
			return false
		end
		if pxclk > 550000000 then
			pxclk = 594000000
		elseif pxclk > 250000000 then
			pxclk  = 297000000
		elseif pxclk > 140000000 then
			pxclk  = 148500000
		elseif pxclk > 70500000 then
			pxclk = 74500000
		end			
		

		local refreshFreq = math.floor(pxclk/frame_pxs)
		if refreshFreq <= 24 then
			refreshFreq = 24
		elseif refreshFreq <= 27 then
			refreshFreq = 25
		elseif refreshFreq <= 35 then
			refreshFreq = 30
		elseif refreshFreq <=52 then
			refreshFreq = 50
		elseif refreshFreq <=65 then
			refreshFreq = 60
		elseif refreshFreq <=72 then
			refreshFreq = 70
		elseif refreshFreq <= 77 then
			refreshFreq = 75
		elseif refreshFreq <= 88 then
			refreshFreq = 85
		end	--Else ignore restrict.
		
		if refreshFreq <= 0 then
			return false
		end
		
		if actW < 160 or actW > 4096 or actH < 90 or actH > 2200 or refreshFreq < 23 then
			return false
		end

		local vMode = hdtv_formats.VIRATE_HD

		if pxclk > 550000000 then
			vMode = hdtv_formats.VIRATE_4K60
		elseif pxclk > 250000000 then
			vMode = hdtv_formats.VIRATE_4K
		elseif pxclk < 30000000 then
			vMode = hdtv_formats.VIRATE_SD
		end	

		--if pxclk > 250000000 then
		--	if refreshFreq >= 45 then --Assume it's 4K50 or 4K60
		--		HTotal = HTotal*2
		--		actW = actW*2
		--		vMode = hdtv_formats.VIRATE_4K60
		--	else
		--		vMode = hdtv_formats.VIRATE_4K
		--	end
		--elseif pxclk < 30000000 then
		--	vMode = hdtv_formats.VIRATE_SD
		--end

		actH2 = actH
		totalH2 = VTotal
		
		local formatName
		if interlaced then
			formatName = actW .. "x" .. (actH + actH2) .. "@" .. refreshFreq .. "i"
		else
			formatName = actW .. "x" .. actH .. "@" .. refreshFreq .. " Hz"
		end
		print(formatName)
		local v = self.nostandardFormatInfo[formatName]
		if not v then	--Create this format information
			v = {
				formatId = formatName,
				formatName = formatName,
				frameRate = refreshFreq,
				hTotal = HTotal,
				hAct = actW,
				vTotal0 = VTotal,
				vAct0 = actH,
				vTotal1 = totalH2,
				vAct1 = actH2,
				interlaced = interlaced,
				rateMode = vMode,
				tmdsFreq = pxclk,
				pixelRepeat = false,--TODO:
				aspect_ratio = 0, --That is ASPECT_RATIO_AUTO
				video_system = 0, --That is AMBA_VIDEO_SYSTEM_AUTO
			}
			self.nostandardFormatInfo[formatName] = v
		end
		
		return self:_RecordVideoFormatInfo( src, v )
	end

	------------------------------------------------------
	-- Detect HDMI input format
	------------------------------------------------------
	
	function ITE68051:_DetectHDMIFormat(src)
		local hdcp = false --TODO:
		self.videoFormat.hdcp_content = 0	--It may contains an error status, but, no matter.

		local rslt = self:ReadReg( BASE_MAP, 0x13, {0x00} )
		if rslt and (rslt[1]&0x02) ~= 0 then
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_HDMI or 1
		else
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_DVI or 0
		end

		self.videoFormat._valid = false
		self.videoFormat.signal_locked = true
		
		local colorSpace = "default"

		--Check the color-space:
		self:WriteReg( BASE_MAP, 0x0f, {0x02} ) --Read 2xx regs.
		local rslt = self:ReadReg( BASE_MAP, 0x15, {0x00} ) --reg 0x215
		if rslt and rslt[1] then
			local flg = rslt[1] & 0x60
			if flg == 0x20 then
				colorSpace = "yuv422"
			elseif flg == 0x40 then
				colorSpace = "yuv444"
			elseif flg == 0x00 then
				colorSpace = "rgb444"
			else
				colorSpace = "yuv420"
			end
		end
		self:WriteReg( BASE_MAP, 0x0f, {0x00} ) --Reset to 0xx regs.

		if colorSpace ~= self.videoFormat.colorSpace then
			self.videoFormat.colorSpace = colorSpace
			self.videoFormat._lastFormatChanged = true
		end

		self:_DetectHDMIInfoFrame(src)

		--So sorry, I need check special formats
		local pxclk, HTotal, HActive, VTotal, VActive, interlaced = self:_GetVideoSyncDetail()
		if pxclk then
			if not self.lastFormatRecord.pxClk or math.abs(pxclk - self.lastFormatRecord.pxClk) >= 20000000 
				or not self.lastFormatRecord.HTotal or math.abs(HTotal - self.lastFormatRecord.HTotal) > 4 
				or not self.lastFormatRecord.HActive or math.abs(HActive - self.lastFormatRecord.HActive) > 4
				or not self.lastFormatRecord.VTotal or math.abs(VTotal - self.lastFormatRecord.VTotal) > 2
				or not self.lastFormatRecord.VActive or math.abs(VActive - self.lastFormatRecord.VActive) > 2
				or interlaced ~= self.lastFormatRecord.interlaced then

				--[[ TODO: We needn't this step.
				rslt = hdtv_formats.Format_Match( pxclk/1000000, HTotal, HActive, VTotal, VActive, interlaced )
				print( "***********", rslt, rslt and rslt.formatName or "***" )
				if rslt then
					if self:_RecordVideoFormatInfo( src, rslt ) then
						self.videoFormat._valid = true
						self.lastFormatRecord.pxClk = pxclk
						self.lastFormatRecord.HTotal = HTotal
						self.lastFormatRecord.HActive = HActive
						self.lastFormatRecord.VTotal = VTotal
						self.lastFormatRecord.VActive = VActive
						self.lastFormatRecord.interlaced = interlaced
						return true
					end
				end
				]]

				if self:_AssumeHDMICustomFormat( src, pxclk, HTotal, HActive, VTotal, VActive, interlaced ) then
						self.videoFormat._valid = true
						self.lastFormatRecord.pxClk = pxclk
						self.lastFormatRecord.HTotal = HTotal
						self.lastFormatRecord.HActive = HActive
						self.lastFormatRecord.VTotal = VTotal
						self.lastFormatRecord.VActive = VActive
						self.lastFormatRecord.interlaced = interlaced
						return true
				else
					return false	--Not detect
				end
			else
				self.videoFormat._valid = true
				return true
			end
		else
			return false
		end
	end

	------------------------------------------------------
	-- Configure HDMI format
	------------------------------------------------------
	function ITE68051:_ConfigureHDMIFormat(src)
		self.videoFormat._lastFormatChanged = false
		self:SetVideoMode( self.videoFormat.signal_locked, self.videoFormat.video_mode, self.videoFormat.colorSpace )
		return true
	end

	-- 0xc4:Clock Phase Adjusting InVert:7 Delay 6:4 0x00
	-- 0xc1:
	-- Reg_OutDDR 0 1: TTL half bus DDR mode R/W 0
	-- Reg_DDRMode 1 1: TTL full bus DDR mode R/W 0
	-- Reg_DDREdgeSel 2 DE edge select when full bus DDR mode R/W 0
	-- Reg_HSyncPol 3 TTL output HSync output polarity R/W 0
	-- Reg_VSyncPol 4 TTL output VSync output polarity R/W 0
	-- Reg_SyncEmb 5[A 1: Enable sync embedded R/W 0
	-- Reg_BTA1004 6 1: Enable BTA1004 mode R/W 0
	-- Reg_ChgSyncPol 7 1: Output Sync’s polarity setting by reg1C1h[4:3]	
	
	
	------------------------------------------------------
	function ITE68051:SetVideoMode(sig,vidMod,colorSpace)
		if not sig then
			self:WriteReg( BASE_MAP,0x0F,{0x01} )
			self:WriteReg( BASE_MAP,0xC1,{0x98} )
			--self:WriteReg( BASE_MAP,0xC4,{0X20} ) 		
			self:hdmirxset(BASE_MAP,0xbd,0x30,{0x00} )
			self:WriteReg( BASE_MAP,0x0F,{0x00} )
		elseif vidMod == "4k60" then
			self:WriteReg( BASE_MAP,0x0F,{0x01} )
			self:WriteReg( BASE_MAP,0xC1,{0x9a} )
			--self:WriteReg( BASE_MAP,0xC4,{0x20} ) 		
			self:hdmirxset(BASE_MAP,0xbd,0x30,{0x10} )
			self:WriteReg( BASE_MAP,0x0F,{0x00} )
		elseif vidMod == "4k" then
			if colorSpace == "rgb444" then --Clock phase adjustment
				self:WriteReg( BASE_MAP,0x0F,{0x01} )
				self:WriteReg( BASE_MAP,0xC1,{0x9b} )
				--self:WriteReg( BASE_MAP,0xC4,{0X20} ) 		
				self:hdmirxset(BASE_MAP,0xbd,0x30,{0x00} )
                                self:WriteReg( BASE_MAP,0x0F,{0x00} )
			elseif colorSpace == "yuv444" then
				self:WriteReg( BASE_MAP,0x0F,{0x01} )
				self:WriteReg( BASE_MAP,0xC1,{0x9b} )
				--self:WriteReg( BASE_MAP,0xC4,{0X20} ) 		
				self:hdmirxset(BASE_MAP,0xbd,0x30,{0x00} )
				self:WriteReg( BASE_MAP,0x0F,{0x00} )		
			else
				self:WriteReg( BASE_MAP,0x0F,{0x01} )
				self:WriteReg( BASE_MAP,0xC1,{0x9b} )
				--self:WriteReg( BASE_MAP,0xC4,{0X20} ) 		
				self:hdmirxset(BASE_MAP,0xbd,0x30,{0x00} )
				self:WriteReg( BASE_MAP,0x0F,{0x00} )
			end		
		elseif vidMod == "sd" then
			self:WriteReg( BASE_MAP,0x0F,{0x01} )
			self:WriteReg( BASE_MAP,0xC1,{0x98} )
			--self:WriteReg( BASE_MAP,0xC4,{0X00} ) 		
			self:hdmirxset(BASE_MAP,0xbd,0x30,{0x00} )
			self:WriteReg( BASE_MAP,0x0F,{0x00} )
		else --Assume all others be 'hd'
			self:WriteReg( BASE_MAP,0x0F,{0x01} )
			self:WriteReg( BASE_MAP,0xC1,{0x98} )
			--self:WriteReg( BASE_MAP,0xC4,{0X00} ) 		
			self:hdmirxset(BASE_MAP,0xbd,0x30,{0x00} )
			self:WriteReg( BASE_MAP,0x0F,{0x00} )
		end
	end
	
	------------------------------------------------------
	
	ITE68051:Reset()
	return ITE68051
end --END of CreateITE68051()

return CreateITE68051

