local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

require "sii9136drv"

local format_name_map = {
	["720x480@60i"] = "480i",
	["720x480@59i"] = "480i",
	["720x480@59.94i"] = "480i",
	["720x576@50i"] = "576i",
	["720x480@60 Hz"] = "480p",
	["720x480@59 Hz"] = "480p",
	["720x480@59.94 Hz"] = "480p",
	["720x576@50 Hz"] = "576p",
	["1280x720@60 Hz"] = "720p",
	["1280x720@59 Hz"] = "720p",
	["1280x720@59.94 Hz"] = "720p",
	["1280x720@50 Hz"] = "720p50",
	["1920x1080@60i"] = "1080i",
	["1920x1080@59i"] = "1080i",
	["1920x1080@59.94i"] = "1080i",
	["1920x1080@50i"] = "1080i50",
	["1920x1080@60 Hz"] = "1080p",
	["1920x1080@59 Hz"] = "1080p",
	["1920x1080@59.94 Hz"] = "1080p",
	["1920x1080@50 Hz"] = "1080p50",
	["1920x1080@25 Hz"] = "1080p25",
	["1920x1080@24 Hz"] = "1080p24",
	["1920x1080@23 Hz"] = "1080p24",
	["1920x1080@23.98 Hz"] = "1080p24",
	["1920x1080@30 Hz"] = "1080p30",
	["1920x1080@29 Hz"] = "1080p30",
	["1920x1080@29.97 Hz"] = "1080p30",
	["2048x1080@60 Hz"] = "1080p",
	["2048x1080@59 Hz"] = "1080p",
	["2048x1080@59.94 Hz"] = "1080p",
	["2048x1080@50 Hz"] = "1080p50",
	["2048x1080@25 Hz"] = "1080p25",
	["2048x1080@24 Hz"] = "1080p24",
	["2048x1080@23 Hz"] = "1080p24",
	["2048x1080@23.98 Hz"] = "1080p24",
	["2048x1080@30 Hz"] = "1080p30",
	["2048x1080@29 Hz"] = "1080p30",
	["2048x1080@29.97 Hz"] = "1080p30",
	["3840x2160@30 Hz"] = "4k30",
	["3840x2160@29 Hz"] = "4k30",
	["3840x2160@29.97 Hz"] = "4k30",
	["3840x2160@25 Hz"] = "4k25",
	["3840x2160@24 Hz"] = "4k24",
	["3840x2160@23.98 Hz"] = "4k24",
	["4096x2160@24 Hz"] = "4k24:smpte",
	["3840x2160@50 Hz"] = "4k50",
	["3840x2160@60 Hz"] = "4k60",
	["4096x2160@50 Hz"] = "4k50",
	["4096x2160@60 Hz"] = "4k60"		
}

local function CreateSII9136( i2c_dev, callback_fn )
-------------------------------------------------------
--{ The constructor wrap function.
-------------------------------------------------------

i2c_dev = i2c_dev or "/dev/i2c-1"

local _M = {
	sii9136 = nil
}

----

function _M:Init()
	if type(callback_fn) == "function" then
		SII9136.setPinControlCallback(callback_fn)
	end

	_M.sii9136 = SII9136.createNew( i2c_dev )
	if not _M.sii9136 then
		_dbg.Error( "SII9136", "Fail to create the chip driver at I2C device:", i2c_dev )
		return false
	end

	_M.sii9136:RefreshEDID() --Try read EDID first!
	_M.sii9136:SetVideoFormat( "custom" ) -- As default first.
	_M.sii9136:SetAudioFormat( 48 ) -- 48KHz sample as default.
	_M.sii9136:UpdateVideoSet()
	_M.sii9136:UpdateAudioSet()
	return true
end

----

function _M:Step()
	if _M.sii9136 then
		_M.sii9136:Poll()
	end
end

----

function _M:UpdateVideoFormat(
	frm_or_vic, 		--Video format name, e.g. "1080i"/"1080p"/"1920x1080p 60Hz"/....
	cea_id,				--CEA format ID, this is get from HDMI InfoFrame.
	colorSpace,			--eg: yuv422,yuv444,rgb444,yuv420.	
	video_mode,			--eg: 4k 4k60 sd hd							
	format_type,		--(Optional) string, "CEA", "HDMI", "4K", "3D", "PC"

	options				--(Optional) table:
						--{
						--	 colorSpace="RGB"/"YCBCR444"/"YCBCR422_16BITS"/"YCBCR422_8BITS"/"XVYCC444",
						--	 colorDepth=8/10/12/16,
						--	 sync=0/1/2, [0=External sync, 1=Internal DE, 2=Embedded]
						--
						--	 ** NOTE: The next 3 arguments can be auto calculated **
						--   aspectRatio="4:3"/"16:9",
						--   colorimetry="601"/"709" [SD=601, others=709]
						--   tclkSel=0.5/1/2/4 [Clock-rate, default is 1; SD should be 2]
						--}
)
	
	if not _M.sii9136 then
		_dbg.Error( "SII9136", "Invalid chip driver!" )
		return false
	end

	if not frm_or_vic then
		_dbg.Error( "SII9136", "Invalid video format name, or CEA ID" )
		return false
	end

	if type(frm_or_vic) == "string" then
		local drv_format = format_name_map[ frm_or_vic ]
		if drv_format then
			frm_or_vic = drv_format
		else
			if frm_or_vic:match("@") then
				frm_or_vic = "custom"
			--otherwise, keep the passed value.
			end
		end
	elseif type(frm_or_vic) ~= "number" then
		frm_or_vic = "custom"
	end

	if not format_type then
		if frm_or_vic == "custom" then
			format_type = "PC"
			frm_or_vic = cea_id
		elseif video_mode:match("4k") and not video_mode:match("4k60") then
			format_type = "4K"
		else 
			format_type = "CEA"
		end
	end	
	_M.sii9136:SetVideoFormat( frm_or_vic, format_type, options )
	_M.sii9136:UpdateVideoSet()
	return true
end

----

function _M:UpdateAudioFormat(
	sample_rate, 	--Number(double), should be: 32, 44.1, 48, 88.2, 96, 176.4, 192, 768
	options			--(Optional) table:
	--{
	--		  mclk_fs = 128/192/256/384/512/768/1024/1152
	--		  channels = N(2~8)
	--		  word_length = N(16~24)      
	--		  edge = rising/falling
	--}
)
	if not _M.sii9136 then
		_dbg.Error( "SII9136", "Invalid chip driver!" )
		return false
	end

	if not sample_rate or type(sample_rate) ~= "number" then
		_dbg.Error( "SII9136", "Invalid audio sampling rate" )
		return false
	end

	if sample_rate >= 8000 then -- get me original sampling, so need /1000
		sample_rate = sample_rate/1000
	end

	_M.sii9136:SetAudioFormat( sample_rate, options )
	_M.sii9136:UpdateAudioSet()
	return true
end

--TODO: More functions, e.g. Read and parse EDID, ...


_M:Init()
return _M

-------------------------------------------------------
--} End of the constructor wrap function.
-------------------------------------------------------
end

return CreateSII9136

