local coaf=require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local client = require "coaf.client"

local periphery = require "periphery"
local GPIO = periphery.GPIO
local posix_time = require "posix.time"

local IO_STEP_TIME = 0.166 --second
local RESTORE_TIME = 3000 --ms.

local FLAG_RESTORE = false

local _M = require "chips.iocfg"

_M.tally_flags = {
	tally = false,
	main = false,
	preview = false
}

local function onTallyEvent( topic, msg, msgType )
	if msgType ~= "table" then
		_dbg.Warn( "IO-LEDS", "Got invalid tally event message of type:", msgType )
		return
	end

	_dbg.Info( "TALLY", "Control", msg.led, " to state:", msg.state )
	if msg.led == "TALLY" then
		_M.tally_flags.tally = (msg.state == "on")
	elseif msg.led == "MAIN" then
		_M.tally_flags.main = (msg.state == "on")
	elseif msg.led == "PREVIEW" then
		_M.tally_flags.preview = (msg.state == "on")
	end

	if _M.tally_flags.main then
		_M.TallyLightControl( nil, "red", "on" )
		_M.TallyLightControl( nil, "green", "off" )
	elseif _M.tally_flags.tally or _M.tally_flags.preview then
		_M.TallyLightControl( nil, "red", "off" )
		_M.TallyLightControl( nil, "green", "on" )
	else
		_M.TallyLightControl( nil, "red", "off" )
		_M.TallyLightControl( nil, "green", "off" )
	end
end

local function io_leds_thread( thrObj )
	local last_button_hold = nil
	_dbg.Info( "IO-LED", "Control thread is started..." )
	while not thrObj.terminated do
		--Check button
		if _M.reset_button.drv then
			local val = _M.reset_button.drv:read()

			if _M.reset_button.revert then
				val = not val
			end

			if not val then
				local now = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )
				if not last_button_hold then
					last_button_hold = now
					_M.on_demo.on = "restore"
					_M.on_demo.stage = 1
					_M.on_demo.stage_led_id = 1
				else
					local holdtime = (now.tv_sec - last_button_hold.tv_sec) * 1000
					holdtime = holdtime + (now.tv_nsec - last_button_hold.tv_nsec)/1000000
					
					if holdtime >= RESTORE_TIME then
						FLAG_RESTORE = true
						_M.on_demo.on = nil
						local i
						for i=1, #_M.tally_leds do
							_M.tally_leds[i].red.state = "off"
							_M.tally_leds[i].green.state = "off"
							thrObj.terminated = true --Exit current loop after LED control.
						end
					end
				end
			else
				last_button_hold = nil
				if _M.on_demo.on ~= "startup" then
					_M.on_demo.on = nil
				end
			end
		end
		
		--On board LED.
		if _M.board_led.drv then 
			if _M.board_led.state == "blink" then
				if not _M.board_led.blink_id then 
					_M.board_led.blink_id = 1
				end
				local level = _M.board_led.blink_steps[ _M.board_led.blink_id ]
				if _M.board_led.revert then
					level = not level
				end
				_M.board_led.drv:write(level)
				_M.board_led.blink_id = _M.board_led.blink_id + 1
				if _M.board_led.blink_id > #_M.board_led.blink_steps then
					_M.board_led.blink_id = 1
				end
			else
				if _M.board_led.revert then
					_M.board_led.drv:write( _M.board_led.state ~= "on" )
				else
					_M.board_led.drv:write( _M.board_led.state == "on" )
				end
			end
		end

		--Tally lights.
		if _M.on_demo.on and _M.on_demo.stage then
			local cur_color = _M.on_demo.steps[_M.on_demo.on][ _M.on_demo.stage ][1]
			local state = _M.on_demo.steps[_M.on_demo.on][ _M.on_demo.stage ][2]

			local prv_color = _M.on_demo.steps[_M.on_demo.on][ _M.on_demo.stage ][3]
			local prv_state = _M.on_demo.steps[_M.on_demo.on][ _M.on_demo.stage ][4]

			if prv_color then
				local drv = _M.tally_leds[ _M.on_demo.stage_led_id ][prv_color].drv
				local revert = _M.tally_leds[ _M.on_demo.stage_led_id ][prv_color].revert
				if drv then
					if revert then
						drv:write( not prv_state )
					else
						drv:write( prv_state )
					end
				end
			end

			local drv = _M.tally_leds[ _M.on_demo.stage_led_id ][cur_color].drv
			local revert = _M.tally_leds[ _M.on_demo.stage_led_id ][cur_color].revert
			if drv then
				if revert then
					drv:write( not state )
				else
					drv:write( state )
				end
			end

			_M.on_demo.stage_led_id = _M.on_demo.stage_led_id + 1
			if _M.on_demo.stage_led_id > #_M.tally_leds then
				_M.on_demo.stage_led_id = 1
				_M.on_demo.stage = _M.on_demo.stage + 1
				if _M.on_demo.stage > #_M.on_demo.steps[_M.on_demo.on] then
					_M.on_demo.on = nil
					_M.on_demo.stage = nil
				end
			end
		else
			local i
			for i=1, #_M.tally_leds do
				if _M.tally_leds[i].red.drv then
					if  _M.tally_leds[i].red.state == "blink" then
						if not _M.tally_leds[i].red.blink_id then
							_M.tally_leds[i].red.blink_id = 1
						end
						
						local steps = _M.tally_leds[i].red.blink_steps or _M.tally_leds[i].blink_steps or _M.tally_leds.blink_steps
						local revert = _M.tally_leds[i].red.revert
						if revert then
							_M.tally_leds[i].red.drv:write( not steps[ _M.tally_leds[i].red.blink_id ] )
						else
							_M.tally_leds[i].red.drv:write( steps[ _M.tally_leds[i].red.blink_id ] )
						end

						_M.tally_leds[i].red.blink_id = _M.tally_leds[i].red.blink_id + 1
						if _M.tally_leds[i].red.blink_id > #steps then
							_M.tally_leds[i].red.blink_id = 1
						end
					else
						local revert = _M.tally_leds[i].red.revert
						if revert then
							_M.tally_leds[i].red.drv:write( (_M.tally_leds[i].red.state == "on") )
						else
							_M.tally_leds[i].red.drv:write( not (_M.tally_leds[i].red.state == "on") )
						end
					end
				end

				if _M.tally_leds[i].green.drv then
					if _M.tally_leds[i].green.state == "blink" then
						if not _M.tally_leds[i].green.blink_id then
							_M.tally_leds[i].green.blink_id = 1
						end
						
						local steps = _M.tally_leds[i].green.blink_steps or _M.tally_leds[i].blink_steps or _M.tally_leds.blink_steps
						local revert = _M.tally_leds[i].green.revert
						if revert then
							_M.tally_leds[i].green.drv:write( not steps[ _M.tally_leds[i].green.blink_id ] )
						else
							_M.tally_leds[i].green.drv:write( steps[ _M.tally_leds[i].green.blink_id ] )
						end

						_M.tally_leds[i].green.blink_id = _M.tally_leds[i].green.blink_id + 1
						if _M.tally_leds[i].green.blink_id > #steps then
							_M.tally_leds[i].green.blink_id = 1
						end
					else
						local revert = _M.tally_leds[i].green.revert
						if revert then
							_M.tally_leds[i].green.drv:write( (_M.tally_leds[i].green.state == "on") )
						else
							_M.tally_leds[i].green.drv:write( not (_M.tally_leds[i].green.state == "on") )
						end
					end
				end
			end
		end

		coaf.Sleep( IO_STEP_TIME )
	end
	_dbg.Info( "IO-LED", "Control thread is terminated!" )
end

function _M.Init()
	if _M.board_led.io then
		_M.board_led.drv = GPIO( _M.board_led.io, "out" )
		_M.board_led.state = "blink"
		_M.board_led.blink_id = 1
		if _M.board_led.revert then
			_M.board_led.drv:write( false )
		else
			_M.board_led.drv:write( true )
		end
	end

	local i
	for i=1, #_M.tally_leds do
		if _M.tally_leds[i].red.io then
			local drv_mode = _M.tally_leds[i].red.revert and "low" or "high"
			_M.tally_leds[i].red.drv = GPIO( _M.tally_leds[i].red.io, drv_mode )
			_M.tally_leds[i].red.state = "off"
		end
		if _M.tally_leds[i].green.io then
			local drv_mode = _M.tally_leds[i].red.revert and "low" or "high"
			_M.tally_leds[i].green.drv = GPIO( _M.tally_leds[i].green.io, drv_mode )
			_M.tally_leds[i].green.state = "off"
		end
	end

	if _M.reset_button.io then
		_M.reset_button.drv = GPIO( _M.reset_button.io, "in" )
	end

	_G.g_event:Subscribe( "tally", onTallyEvent )

	_M.ioThread = thread.New( io_leds_thread )
	_M.ioThread:Start()
end

--Period check.
function _M.Step()
	if FLAG_RESTORE then
		FLAG_RESTORE = false
		local sys_client = client.New( "systemctrl", "local", nil, 3 )
		if not sys_client then
			_dbg.Error( "RESTORE", "Fail to connect systemctrl service!" )
		else
			local proxy = sys_client:CreateProxy( "/", true ) --Exclusive.
			if not proxy then
				_dbg.Error( "RESTORE", "Fail to create systemctrl:/ service proxy" )
				sys_client:Destroy()
			else
				_dbg.Info( "RESTORE", "Restored factory settings since pressed restore button!" )
				proxy.RestoreFactory()
				proxy:Destroy()
			end
		end
	end
end

function _M.TallyLightControl(id, color, status)
	id = tonumber(id)
	if status then
		if not id or id < 1 or id > #_M.tally_leds then
			--_M.on_demo.on = nil
			local i
			for i=1, #_M.tally_leds do
				if color ~= "red" and color ~= "green" then
					_M.tally_leds[i].red.state = status
					_M.tally_leds[i].green.state = status
				else
					_M.tally_leds[i][color].state = status
				end
			end
		else
			--_M.on_demo.on = nil
			if color ~= "red" and color ~= "green" then
				_M.tally_leds[id].red.state = status
				_M.tally_leds[id].green.state = status
			else
				_M.tally_leds[id][color].state = status
			end
		end
	end

	if _M.signal and _M.signal.tally_light_id then
		if not (_M.tally_flags.tally or _M.tally_flags.main or _M.tally_flags.preview) then
			_M.tally_leds[ _M.signal.tally_light_id ].red.state = _M.signal.red
			_M.tally_leds[ _M.signal.tally_light_id ].green.state = _M.signal.green
			_M.tally_leds[ _M.signal.tally_light_id ].red.blink_id = 1
			_M.tally_leds[ _M.signal.tally_light_id ].green.blink_id = 1
		end
	end
end

function _M.SetLightStatus( status )
	if status == "reset" then
		_M.on_demo.on = "reset"
		_M.on_demo.stage = 1
		_M.on_demo.stage_led_id = 1
	elseif status == "reboot" or status == "restore" or status == "off" then
		_M.on_demo.on = nil
		local i
		for i=1, #_M.tally_leds do
			_M.tally_leds[i].red.state = "off"
			_M.tally_leds[i].green.state = "off"
		end
	end
end

function _M.SetSignalStatus( status )
	if not _M.signal then
		return
	end

	if status == "sd" or status == "hd" then
		_M.signal.green = "blink"
		_M.signal.red = "off"
	elseif status == "4k" then
		_M.signal.green = "blink"
		_M.signal.red = "blink"
	elseif status == "4k60" then
		_M.signal.green = "off"
		_M.signal.red = "blink"
	else
		_M.signal.green = "off"
		_M.signal.red = "off"
	end

	_M.TallyLightControl() --Just update.
end

return _M

