local coaf = require "coaf.core"
local thread = require "coaf.thread"
local _dbg = require "coaf.debug"
-- local pretty = require "pl.pretty"
local evdev = require "evdev"
local socket = require "cosocket.core"
local lfs = require "lfs"
local ev = require "ev"

local _M = {
	callback_fn = nil,
	callback_obj = nil,
	keyboards = {},
	probe_task = nil,
	symbols = {}
}

local function new_keyboard(device)
	local keyboard = {
		id = device,
		CAPSL = nil,
		NUML = nil,
		SCROLLL = nil,
		downing_keys = {}
	}

	local dev = evdev.Device("/dev/input/by-id/" .. device, true) --Make it to be writable.

	--Make the NUML be on by default.
	dev:write(evdev.EV_LED, evdev.LED_NUML, 1)
	dev:write(evdev.EV_SYN, evdev.SYN_REPORT, 0)
	---------------------------------

	keyboard.CAPSL = dev:getstatus(evdev.EV_LED, evdev.LED_CAPSL)
	keyboard.NUML = dev:getstatus(evdev.EV_LED, evdev.LED_NUML)
	keyboard.SCROLLL = dev:getstatus(evdev.EV_LED, evdev.LED_SCROLLL)

	keyboard.dev = dev
	keyboard.dev_fd = dev:pollfd()
	keyboard.dev_watcher = coaf.NewIO( function(w, revents)
		local timestamp, eventType, eventCode, value = dev:tryRead()
		if not timestamp then
			_dbg.Error("KEYBOARD", "The keyboard /dev/input/by-id/" .. device .. " read error since EOF. Is it disconnected?" )
			w:destroy()
			_M.keyboards[device] = nil
			keyboard.dev:close()
			return
		else
			if eventType == evdev.EV_LED then
				if eventCode == evdev.LED_CAPSL then
					keyboard.CAPSL = value
				elseif eventCode == evdev.LED_NUML then
					keyboard.NUML = value
				elseif eventCode == evdev.LED_SCROLLL then
					keyboard.SCROLLL = value
				end
			elseif eventType == evdev.EV_KEY then
				if value == 0 then --Key released.
					keyboard.downing_keys[eventCode] = nil
					--if _M.key_callback then
						_M.key_callback(keyboard, eventCode, value, timestamp)
					--end
				else
					--if _M.key_callback then
						_M.key_callback(keyboard, eventCode, value, timestamp)
					--end
					keyboard.downing_keys[eventCode] = timestamp
				end
			end
		end
	end, keyboard.dev_fd, ev.READ)

	_M.keyboards[device] = keyboard
	keyboard.dev_watcher:start()
	return keyboard
end

local function remove_keyboard(device)
	local keyboard = _M.keyboards[device or "#invalid"]
	if keyboard then
		_M.keyboards[device] = nil --Remove the record first.
		if keyboard.dev_watcher then
			keyboard.dev_watcher:destroy()
		end
		if keyboard.dev then
			keyboard.dev:close()
		end
	end
end

--[[Probe keyboard devices task]]
local function probe_keyboards(threadObj)
	_dbg.Info("KEYBOARD", "Keyboard probe task running...")

	while not threadObj.terminated do

		local found_keyboards = {}
		local new_keyboards = {}
		local loss_keyboards = {}

		local attr = lfs.attributes("/dev/input/by-id")
		if attr and attr.mode == "directory" then
			for dev in lfs.dir("/dev/input/by-id") do
				if dev ~= "." and dev ~= ".." and dev:match("-kbd$") then
					found_keyboards[dev] = true
				end
			end
		end

		for item,_v in pairs(_M.keyboards) do
			if not found_keyboards[item] then
				table.insert(loss_keyboards, item)
			end
		end

		for item,_v in pairs(found_keyboards) do
			if not _M.keyboards[item] then
				table.insert(new_keyboards, item)
			end
		end

		if #loss_keyboards > 0 then
			for _i, item in ipairs(loss_keyboards) do
				_dbg.Info( "KEYBOARD", "Keyboard of /dev/input/by-id/" .. item .. " is disconnected.")
				remove_keyboard(item)
			end
		end

		if #new_keyboards > 0 then
			for _i, item in ipairs(new_keyboards) do
				_dbg.Info( "KEYBOARD", "A new keyboard of /dev/input/by-id/" .. item .. " connected.")
				new_keyboard(item)
			end
		end

		coaf.Sleep(2)
	end
	_dbg.Info("KEYBOARD", "Keyboard probe task is terminated.")
end

-------------------------------------------------------

function _M.key_callback(keyboard, key_code, key_val, timestamp)
	if _M.callback_fn then
		if _M.callback_obj then
			_M.callback_fn(_M.callback_obj, keyboard, key_code, key_val, timestamp)
		else
			_M.callback_fn(keyboard, key_code, key_val, timestamp)
		end
	end
end

-------------------------------------------------------

function _M.Init()
	if not _M.probe_task then
		_M.probe_task = thread.New(probe_keyboards)
		_M.probe_task:Start()
	end
end

function _M.RegisterCallback(callback, callback_obj)
	_M.callback_fn = callback
	_M.callback_obj = callback_obj
end

setmetatable(_M.symbols, {
	__index = require "evdev.constants"
})

return _M
