#!/usr/bin/env lua

local coaf=require "coaf.core"
local _dbg = require "coaf.debug"

coaf.ChDir(coaf.GetAppDir(arg))

local socket = require "cosocket.core"
local service = require "coaf.service"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
local thread = require "coaf.thread"

local withLog = false
local _x
for _x=1, #arg do
	if arg[_x] == "--log" then
		withLog = true
		break
	end
end

if withLog then
	-- ** initialize the log file **
	local roll = require "coaf.log.rollfile"
	os.execute( "/bin/mkdir -p /var/log/" )
	local logF = roll.Create("/var/log/codec_mate.log")
	if not logF or not logF:init() then
		_dbg.Error("CODEC-MATE", "Fail to open log file: /var/log/codec_mate.log, so the log messages should be output to console.")
		withLog = false
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
	-- ** log **
end

coaf.Init( "codec_mate" )

if withLog then
	local stdio = require "coaf.log.stdio"
	stdio.SetDefaultFilterName( "CODEC-MATE" )
	stdio.Start()
end

service.InitSocket()
client.InitSocket( socket, adapter )

-- ** CoService **
local cosvc_CodecMate
local sure_codecServiceWorking = false
local max_trytimes = 5
local min_trytimes = 2
local try_times = max_trytimes
------------------
-- ** A simple service **
------------------
local mate_service = {}

function mate_service.NotifyCodecReady()
	_dbg.Info( "CODEC-MATE", "Codec service is online!" )
	sure_codecServiceWorking = true
	try_times = min_trytimes
	return true
end

local created_proxy = nil

local function GetCodecCheckProxy()
	if created_proxy then
		return created_proxy
	end

	local proxy
	local cli
	cli = client.New( "codec", "local", nil, 5 )
	if not cli then
		_dbg.Error( "CODEC-MATE", "Fail to create connection to service 'local/codec'" )
		return nil
	end
	proxy = cli:CreateProxy("/", true) --Exclusive
	if not proxy then
		_dbg.Error( "CODEC-MATE", "Fail to create object proxy for '/: local/codec'" )
		cli:Destroy()
		return nil
	end

	created_proxy = proxy
	return proxy
end

-- ** period checker**

local function for_PeriodCheck()
	local proxy = GetCodecCheckProxy()
	if not proxy then
		_dbg.Error( "CODEC-MATE", "Fail to get codec proxy!" )
		return
	end
	local done,be_blocked = proxy.IsBlocked()
	if not done or be_blocked then
		try_times = try_times - 1
		if try_times <= 0 then
			_dbg.Info( "CODEC-MATE", "Codec service looks to be dead or blocked (tried", (sure_codecServiceWorking and min_trytimes or max_trytimes), "times), reset it!" )
			sure_codecServiceWorking = false
			try_times = max_trytimes
			os.execute( "/usr/local/bin/coctrl.lua stop codec" )
			os.execute( "/usr/local/bin/coctrl.lua start codec" )
		end
	else
		try_times = (sure_codecServiceWorking and min_trytimes or max_trytimes)
	end
end

-- ** Timers **
local forTimer

-- **Main entry**
local function Main(args)

	cosvc_CodecMate = service.New( "codec_mate", "unix:/var/run/codec_mate.sock" )
	if cosvc_CodecMate then
		_dbg.Info( "CODEC-MATE", "CoService of 'codec_mate' is created." )
		cosvc_CodecMate:CreateCoServerProxy()
		cosvc_CodecMate:ExposeObject("/", mate_service )
	else
		_dbg.Error( "CODEC-MATE", "Fail created CoService!" )
	end

	forTimer = thread.NewTimer( for_PeriodCheck, 1, 2 )
	if forTimer then
		forTimer:start()
	end
end

coaf.Start(Main, arg)

if forTimer then
	forTimer:stop()
end

if created_proxy then
	created_proxy:Destroy()
end

_dbg.Print( "CODEC-MATE", "Service terminated." )

