-- The downloader
local coaf = require "coaf.core"
local thread = require "coaf.thread"
local run_daemon = require "coaf.run_daemon"

local _M = {
	tasks = {}
}

---------------------------------------------------

local down_tools = {
	http = require "http_wget",
	https = require "http_wget"
}

local downmap = require "downmap"

---------------------------------------------------

function _M.Init()
	run_daemon.Init()
end

---------------------------------------------------

function _M.StartDownload(down_type, name,url,dir)
	local status = _M.tasks[name]
	if not status then
		status = {}
		_M.tasks[name] = status
	end

	if status.running_task then
		return nil, "Downloading task " .. name .. " is in progress!"
	elseif not url then
		return nil, "Invalid download URL"
	else
		local type_flag = url:match("^(.*)://.*")
		if not type_flag then
			return nil, "Invalid downloading URL"
		end

		local down = down_tools[ type_flag ]
		if not down then
			return nil, "Unsupported download type: " .. type_flag
		end

		local file = url:match(".+/(.*)$")
		if not file then
			return nil, "For file downloading, you must tell me the filename in URL"
		end

		dir = dir or "/tmp"
		if down_type then
			local down_item = downmap[ down_type ]
			if down_item and type(down_item.beforeProcess) == "function" then
				local r, err = down_item.beforeProcess(dir,file,url)
				if not r then
					status.down_type = down_type
					status.status = "error"
					status.message = err or "prepare downloading fail"
					status.progress = "0%"
					return nil, err
				end
			end
		end

		status.down_type = down_type
		status.status = "pending"
		status.message = "Starting download"
		status.progress = "0%"
		status.dir = dir
		status.file = file

		status.running_task = run_daemon.Create({
			name = name,

			run = function(event, name, url, dir, file )
				local r = down( event, name, url, dir, file )
				os.exit(r or 0)
			end,

			args = { name, url, dir, file },

			on_started = function(task_obj)
				local status = task_obj.owner
				if status then
					status.status = "downloading"
					status.message = "Downloading started"
				end
			end,

			on_terminate = function(task_obj,flag,reason,code)
				local status = task_obj.owner
				task_obj.owner = nil
				status.running_task = nil

				if flag and reason == "exit" and code == 0 then
					if status.status == "pending" or status.status == "downloading" then
						status.status = "finished"
					end
					if status.down_type then
						local down_item = downmap[ status.down_type ]
						if down_item and type(down_item.afterProcess) == "function" then
							local r,err = down_item.afterProcess( status.status, status.dir, status.file )
							if not r then
								status.status = "error"
								status.message = err or "Download after processing fail"
							end
						end
					end
				elseif reason == "kill" then
					status.status = "aborted"
					status.message = "Downloading task is aborted"
				else
					if status.status == "pending" or status.status == "downloading" then
						status.status = "error"
						status.message = "Downloading is error"
					end
				end
				return nil
			end,

			on_event = function(task_obj, event_id, msg)
				local status = task_obj.owner
				if status then
					status.status = msg.status or status.status
					status.message = msg.message or status.message
					status.progress = msg.progress or status.progress
				end
			end
		})

		if status.running_task then
			status.running_task.owner = status
			status.running_task:Start()
			return true
		else
			status.status = "error"
			status.message = "Fail to start download task"
			return nil, "Fail to start download task for: " .. name
		end
	end
end

------------------------------------------

function _M.AbortDownload(name)
	local status = _M.tasks[name]
	if status and status.running_task then
		status.running_task:Kill()
		status.running_task:Destroy()
		status.running_task = nil
	end
	return true
end

------------------------------------------

function _M.GetStatus(name)
	return _M.tasks[name]
end

------------------------------------------

return _M

