#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local service = require "coaf.service"
local roll = require "coaf.log.rollfile"

-- Change current dir to app location
coaf.ChDir(coaf.GetAppDir(arg))

local config = require "netcfg"
local funcs = require "nm_funcs"
local _did, special = pcall( require, "nm_special" )
if not _did then
	special = nil
end

require "SimpleNetHelper"

-- First, initialize the log file.
os.execute( "/bin/mkdir -p /var/log/nm" )
local logF = roll.Create("/var/log/nm/networkmanager.log")
if not logF or not logF:init() then
	_dbg.Error("NM", "Fail to open log file: /var/log/nm/networkmanager.log, so the log messages should be output to console.")
else
	_dbg.AddLogger("logfile", logF, nil, nil)
end
----------------------------------

coaf.Init( "networkmanager" )
-- Initialize service.
service.InitSocket()

local g_service

-- Get the configure file.
local nm_config = config.Load("/etc/network/nm.default") -- 'nm.default' is fail-safe configure.

-- Mark it is on patching.
config.PATCHING = true
if special and special.PatchConfigs then
	local _tmp_r = special.PatchConfigs(nm_config)
	if _tmp_r then
		nm_config = _tmp_r
	end
end
config.PATCHING = false

----------------------------------
local nmService = {
	configuredIntfs = {},
	interfaceStatus = {}
}

local function TableExists(tb,key)
	local i
	for i=1, #tb do
		if tb[i] == key then
			return i
		end
	end
	return nil
end

local function GetDNSList()
	local f = io.open( "/etc/resolv.conf", "r" )
	if not f then
		return nil
	else
		local rslt = {}
		local l = f:read("*l")
		local addr
		while l do
			addr = l:match("^%s*nameserver%s+([%d%.]+)")
			if addr then
				table.insert(rslt, addr)
			end
			l = f:read("*l")
		end
		f:close()
		return rslt
	end
end

function nmService.Reset()
	if special and special.BeforeReset then
		special.BeforeReset( nmService )
	end
	nmService.StopMDNS()
	nmService.configuredIntfs = {}
	nmService.ConfigNetwork()
	nmService.StartMDNS()
	if special and special.AfterReset then
		special.AfterReset( nmService )
	end
end


function nmService.StartMDNS()
	if nm_config.enable_mdns and special and special.StartAvahiDaemon then
		local ignoreAddrs = {}
		if nm_config.aliases then
			local k,v
			for k,v in pairs(nm_config.aliases) do
				if v.enable then
					if v.address and v.address ~= "" and v.address ~= "0.0.0.0" then
						table.insert( ignoreAddrs, v.address )
					end

					if v.address2 and v.address2 ~= "" and v.address2 ~= "0.0.0.0" then
						table.insert( ignoreAddrs, v.address2 )
					end
				end
			end
		end

		if  nm_config.interlink 
			and nm_config.interlink.enable 
			and nm_config.interlink.address 
			and nm_config.interlink.address ~= "" 
			and nm_config.interlink.address ~= "0.0.0.0" then
			table.insert( ignoreAddrs, nm_config.interlink.address )
		end

		special.StartAvahiDaemon( ignoreAddrs )
	end
end

function nmService.StopMDNS()
	if nm_config.enable_mdns and special and special.StopAvahiDaemon then
		special.StopAvahiDaemon()
	end
end

function nmService.ConfigNetwork(intfs)
	if not intfs then
		intfs = funcs.EnumNetworks()
		if special and special.FixEnumeratedInterfaces then
			local _tmp_r = special.FixEnumeratedInterfaces( nmService, intfs )
			if _tmp_r then
				intfs = _tmp_r
			end
		end
	end

	local r, err
	if nm_config.interlink then
		r = TableExists( intfs, nm_config.interlink.ifname )
		--local carrier = intfs[nm_config.interlink.ifname]
		if r then
			table.remove( intfs, r )
			nmService.configuredIntfs[ nm_config.interlink.ifname ] = true -- Mark it configured.
			r, err = funcs.ConfigInterlink( nm_config.interlink )
			if not r then
				_dbg.Error( "NM", "Fail to configure inter-link, with error:", err )
			end
		end
	end

	local i, ifname
	for i, ifname in ipairs(intfs) do
		nmService.configuredIntfs[ ifname ] = true -- Mark it configured.

		if nm_config.ethernets then
			local cfg = nm_config.ethernets[ifname]
			if not cfg then
				nm_config:S_ADD( ifname, {} ) -- Not configured, add it.
				cfg = nm_config.ethernets[ifname]	--Default config
			end

			r, err = funcs.ConfigEthernet( ifname, cfg )
			if not r then
				_dbg.Error( "NM", "Fail to configure interface of '" .. ifname .. "', error:", err )
			end
		end

		if nm_config.aliases then
			local cfg = nm_config.aliases[ifname]
			if cfg then
				r, err = funcs.ConfigEthernetAlias( ifname, cfg )
				if not r then
					_dbg.Error( "NM", "Fail to configure interface alias of '" .. ifname .. "', error:", err )
				end
			end
		end

		if special and special.InterfaceSpecialConfig then
			special.InterfaceSpecialConfig( nmService, ifname, intfs[ifname] and intfs[ifname]~=0 or false )
		end
	end -- for.
end

function nmService.EnumNetworks()
	local intfs = funcs.EnumNetworks()

	if special and special.FixEnumeratedInterfaces then
		local _tmp_r = special.FixEnumeratedInterfaces( nmService, intfs )
		if _tmp_r then
			intfs = _tmp_r
		end
	end
	return intfs
end

function nmService.GetAllInterfaces()
	return SnH.GetAllInterfaces()
end

function nmService.GetLinkStatus(ifname)
	return funcs.GetLinkStatus(ifname)
end

function nmService.GetLinkDetail(ifname)
	if nm_config.ethernets and nm_config.ethernets[ifname] then
		local cfg = nm_config.ethernets[ifname]
		local ret = funcs.GetLinkDetail(ifname)
		ret.method = cfg.method
		if cfg.method ~= "dhcp" then
			if not ret.address or ret.address == "0.0.0.0" then
				ret.address = cfg.address
				ret.netmask = cfg.netmask
			end
			if not ret.gw or ret.gw == "" or ret.gw == "0.0.0.0" then
				ret.gw = cfg.gw
			end
			ret.dns = cfg.dns
		else
			ret.dns = GetDNSList()
		end
		return ret
	else
		return nil
	end
end

----------------------------------

local periodChecker
local check_is_working = false

local function checkInterfaces()
	if check_is_working then
		return
	end

	check_is_working = true
	pcall( function()

		local intfs = funcs.EnumNetworks()
		if special and special.FixEnumeratedInterfaces then
			local _tmp_r = special.FixEnumeratedInterfaces( nmService, intfs )
			if _tmp_r then
				intfs = _tmp_r
			end
		end

		local new_or_update_intfs={}

		local i,v
		--Check carriers
		for i,v in ipairs(intfs) do
			if not nmService.configuredIntfs[ v ] then
				_dbg.Info( "NM", "New interface:", v, "found, configure it." )
				table.insert( new_or_update_intfs, v )
			elseif nmService.interfaceStatus and nmService.interfaceStatus[v] ~= intfs[v] then --Carrier changed.
				if intfs[v] and intfs[v] ~= 0 then
					_dbg.Info( "NM", "Interface:", v, "carrier on, update the network configuration." )
				else
					_dbg.Info( "NM", "Interface:", v, "carrier off, disable it." )
				end

				table.insert( new_or_update_intfs, v)
			end
		end

		nmService.ConfigNetwork( new_or_update_intfs )

		for i,v in pairs(nmService.configuredIntfs) do
			if not TableExists( intfs, i ) then
				nmService.configuredIntfs[i] = nil -- Interface lost/no carrier, drop it.
			end
		end

		nmService.interfaceStatus = intfs
	end ) --End of pcall.

	check_is_working = false
end

----------------------------------

local function Main(args)
	if special and special.LoadDrivers then
		special.LoadDrivers()
	end

	funcs.Init()
	nmService.Reset()
	periodChecker = coaf.NewTimer( checkInterfaces, 1, 2 )
	if periodChecker then
		periodChecker:start()
	end

	g_service = service.New( "networkmanager", "unix:/var/run/networkmanager.sock" )
	if g_service then
		g_service:CreateCoServerProxy()
		g_service:ExposeObject("/", nmService )
		g_service:ExposeObject("/config/nm", nm_config )
	else
		_dbg.Error( "NM", "Fail to create CoService for networkmanager" )
	end
end

coaf.Start(Main,arg)

if periodChecker then
	periodChecker:destroy()
end

if special and special.RemoveDrivers then
	special.RemoveDrivers()
end

_dbg.Print( "NM", "Network manager service is terminated." )

