local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local run_daemon = require "coaf.run_daemon"

local _M = {
	avahi_daemon_task = nil,
	dhcp_service = nil,
	dhcp_pubs = {}
}

local function udhcpc_action( action, ifname, pidfile )
	if action == "exec" or action == "execp" then
		return "/sbin/udhcpc", { "/sbin/udhcpc", "-p", pidfile or "/var/run/dhclient-" .. (ifname or "eth0") .. ".pid", "-f", "-i", ifname or "eth0" }
	else
		return "/sbin/udhcpc -i " .. (ifname or "eth0")
	end
end

function _M.GetTools()
--NOTE: declare your network tool commands here.
--
--NOTE: DHCLIENT:
--      if DHCLIENT is a string (command), that means it's comaptible with 'dhclient', so
--      the system will invoke it as 'DHCLIENT -pf <pid-file> interface'.
--
--      if DHCLIENT is a function, the proto type of function is:
--
--      funciton DHCLIENT( command_type, interface_name, pid_file_name )
--
--          -- This function should return 2 results:
--          --     1) tool_command: if your dhcp client is not dhclient comaptible, 
--          --        you can return your special command, e.g. 'udhcpc'
--          --     2) command_arguments: you can combine 'interface_name', 'pid_file_name'
--          --        and any other options, as the command arguments of your 'tool_command'
--          --        according to 'command_type', the 'command_arguments' should
--          --        be an array of a string, reference follow:
--
--      	-- The 'command_type' has two cases:
--      	--     1) 'exec': in this case, 'command_arguments' need returns as
--      	--        an array, each item is an argument.
--      	--     2) 'cmd': in this case, 'command_arguments' need returns as a 
--      	--        string, sperated by space.
--
--      end
--
	return {
		IFCONFIG = "/sbin/ifconfig",
		ROUTE = "/sbin/route",
		IP = "/sbin/ip",
		DHCLIENT = udhcpc_action
	}
end

function _M.LoadDrivers()
	--TODO: Before network service start, it will load network drivers.
	--      by default, this will do nothing, if your OS drivers is loaded
	--      automatically.
	--      But, if you want to load your special driver(s), or do some
	--      special actions before start network service, you can write
	--      your action codes here.

	--NOTE: This is a bug fixing, for modify MAC address.
	os.execute( "/sbin/ifconfig eth0 down" )
	run_daemon.Init()
	os.execute( "/bin/ln -sf /usr/sbin/dnsmasq /usr/sbin/dhcp_server" )
	os.execute( "/usr/bin/killall dhcp_server" )
	os.execute( "/usr/bin/killall avahi-daemon" )
end

function _M.RemoveDrivers()
	--TODO: After network service is stopped, it will call this, so
	--      you can write your driver cleanup codes here.
end

function _M.PatchConfigs(nm_config)
	--TODO: The argument 'nm_config' is the network configure table, (see 
	--      netcfg.lua). If you want to change the configurations before
	--      setup network interfaces (e.g. you want to change the MAC addr
	--      or specify a fixed IP address, ...), you can change it here.
	--      DON'T FORGET return the network configuration table back.

	if nm_config.ethernets and nm_config.ethernets.eth0 then
		local cfg = nm_config.ethernets.eth0
		if cfg.mac == "" then
			local f = io.popen( "/usr/bin/uuidgen", "r" )
			if f then
				local l = f:read("*l")
				f:close()
				if l then
					local r1,r2,r3,r4,r5,r6 = l:match(".+%-(%x%x)(%x%x)(%x%x)(%x%x)(%x%x)(%x%x)")
					local x1 = tonumber( "0x" .. r1 )
					x1 = x1 & 0x7e
					cfg.mac = string.format( "%02x:", x1 ) .. r2 .. ":" .. r3 .. ":" .. r4 .. ":" .. r5 .. ":" .. r6
					nm_config:__TRIGGER(true)
				else
					_dbg.Error( "NM", "Failure to get uuidgen result, failure to generate randomize MAC address" )
				end
			else
				_dbg.Error( "NM", "No uuidgen found or failure to execute uuidgen, failure to generate randomize MAC address" )
			end
		end
	end

	--Always enable MDNS service.
	if not nm_config.enable_mdns then
		nm_config.enable_mdns = true
	end

	--Always enable alias (if disabled)
	if nm_config.aliases and nm_config.aliases.eth0 and not nm_config.aliases.eth0.enable then
		nm_config.aliases.eth0.enable = true
	end

	return nm_config
end

function _M.BeforeReset(nmService)
	--TODO: Before reset network service, it will invoke this.
	--      Argument 'nmService' is the service 
end

function _M.AfterReset(nmService)
	--TODO: After network service reset done, it will invoke this.
end

function _M.FixEnumeratedInterfaces(nmService,intfs)
	--TODO: After system enumerated network interfaces (or updated),
	--      it will invoke this function, 'intfs' is an array contains
	--      all interfaces name. For some purpose, you may need to
	--      drop (disable) some interfaces, so you can change the 
	--      intfs array. DON'T FORGET to return 'intfs' back.
	return intfs
end

function _M.InterfaceSpecialConfig(nmService,intf)
	--TODO: After the interface of 'intf' is configured by default
	--      processing, if you wan't to do some other configure, e.g.
	--      using 'ethtool' to set the interface parameters, you can
	--      write your codes here.
end

function _M.StartAvahiDaemon(ignoreAddrs)
	--TODO: (this is the new feature): If you configured to enable
	--      avahi-daemon (mDNS) service, after network interfaces
	--      are configured, it will call this function to start 
	--      avahi-daemon. So you can run that service program by
	--      yourself; or if you do nothing, the avahi-daemon 
	--      service is ignored.
	--
	--      'ignoreAddrs' is an array contains all ignored IP addresses
	--      for avahi-daemon service. Typically, these addresses
	--      are fail-safe addresses (alias addresses).
	
	if _M.avahi_daemon_task then
		_M.avahi_daemon_task:Kill()
		_M.avahi_daemon_task:Destroy()
		_M.avahi_daemon_task = nil
	end

	local args = {}
	if ignoreAddrs then
		local i
		for i=1, #ignoreAddrs do
			table.insert( args, "-B" )
			table.insert( args, ignoreAddrs[i] )
		end
	end

	_M.avahi_daemon_task = run_daemon.Create({
		name = "avahi-daemon",
		run = "/usr/sbin/avahi-daemon",
		args = args,
		on_terminate = "restart"
	})

	if _M.avahi_daemon_task then
		_M.avahi_daemon_task:Start()
	end

end

function _M.StopAvahiDaemon()
	--TODO: (this is the new feature): Call this to stop avahi-daemon
	--      (mDNS) service.
	
	if _M.avahi_daemon_task then
		_M.avahi_daemon_task:Kill()
		_M.avahi_daemon_task:Destroy()
		_M.avahi_daemon_task = nil
	end
end

local function update_dhcp_service()
	if _M.dhcp_service then
		_M.dhcp_service:Kill()
		_M.dhcp_service:Destroy()
		_M.dhcp_service = nil
	end

	if #_M.dhcp_pubs > 0 then
		local args = {}
		local i, j
		for i=1, #_M.dhcp_pubs do
			for j=1, #_M.dhcp_pubs[i] do
				table.insert( args, _M.dhcp_pubs[i][j] )
			end
		end
		table.insert( args, "-l" )
		table.insert( args, "/tmp/dhcp.leases" )
		table.insert( args, "-p" ) -- -p 0 to disable DNS features.
		table.insert( args, "0" )
		table.insert( args, "-d" )

		_M.dhcp_service = run_daemon.Create({
			name = "dhcp-server-dnsmasq",
			run = "/usr/sbin/dhcp_server",
			args = args,
			on_terminate = "restart"
		})
	
		if _M.dhcp_service then
			_M.dhcp_service:Start()
		end
	end
end

function _M.StartDHCPService(ifname, ifaddr, addr_prefix, options)
	--TODO: (this is the new feature):
	-- If you configured to enable DHCP service for specified interface 
	-- ('ifname'), it will call this function to start DHCP service.
	-- 'ifaddr' is the configured IP address for 'ifname', and 
	-- 'addr_prefix' is the address prefix (number).
	--
	-- 'options' is a table, contains:
	-- {
	--    dynamic_start = "DHCP start IP address",
	--    dynamic_end = "DHCP end IP address",
	--    gw = "gateway",
	--    dns = { array of dns list },
	--
	--    --NOTE: below is according to DHCP ruler to set DHCP option
	--    dhcp_options = {
	--    	[option-number] = "value",
	--    	...
	--    }
	-- }
	--
	-- you can run the DHCP service program by yourself according to the
	-- passed arguments, or do nothing to ignore that.
	local i, item
	for i=1, #_M.dhcp_pubs do
		if _M.dhcp_pubs[i].ifname == ifname then
			item = { ifname = ifname }
			_M.dhcp_pubs[i] = item
			break
		end
	end
	if not item then
		item = { ifname = ifname }
		table.insert( _M.dhcp_pubs, item )
	end
	item[1] = "-i"
	item[2] = ifname
	item[3] = "--listen-address"
	item[4] = ifaddr
	item[5] = "--dhcp-host"
	item[6] = ifaddr
	item[7] = "--dhcp-range"
	item[8] = options.dynamic_start .. "," .. options.dynamic_end .. ",8h"
	update_dhcp_service()
end

function _M.StopDHCPService(ifname)
	--TODO: (this is the new feature)
	--      Stop the DHCP service for specified interface 'ifname'.
	
	for i=1, #_M.dhcp_pubs do
		if _M.dhcp_pubs[i].ifname == ifname then
			table.remove(_M.dhcp_pubs,i)
			break
		end
	end

	update_dhcp_service()
end

return _M

