local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local sysinfo = require "system.sysInfo"
local TOOLS = require "system.cmdtools"
local web_serv = require "system.web_server"

local with_sysconf, sysconf = pcall(require,"sys_config")
if with_sysconf then
	if type(sysconf) ~= "table" then
		sysconf = _G
	end
else
	sysconf = _G
end

local sysctrlService = {}

------------------------------------------------

function sysctrlService.Init()
end

function sysctrlService.ProcessFirmware(file, fwType)
	_dbg.Info( "SYSCTRL", "Processing firmware: ", file, ", with firmware type:", fwType or "default" )
	local f = io.open( file, "r" )
	if not f then
		_dbg.Error( "SYSCTRL", "Invalid firmware file of: ", file )
		return false, "invalid", "Invalid firmware file"
	end
	f:close()

	if not fwType or fwType == "" or fwType == "default" then
		local up_proc = require "system.upgradeProc"
		local r, reason, msg = up_proc.ProcessUpgradeFirmware(file, "/tmp/.fw_upgrade" )
		if not r then
			_dbg.Error( "SYSCTRL", "Fail to process upgrading firmware file, with reason and error: ", reason, msg )
		else
			_dbg.Info( "SYSCTRL", "Firmware processing DONE." )
		end
		return r, reason, msg
	else --Other firmware types, not sure what processing type.
		local done, up_proc = pcall( require, "system.upgrade_" .. fwType )
		if not done or not up_proc then
			_dbg.Error( "SYSCTRL", "Unsupport firmware type of", fwType )
			return false, "unsupport", "Unsupport firmware type"
		end
		local r, reason, msg = up_proc.ProcessUpgradeFirmware(file)
		if not r then
			_dbg.Error( "SYSCTRL", "Fail to process upgrading firmware file (type=", fwType, "), with reason and error: ", reason, msg )
		else
			_dbg.Info( "SYSCTRL", "Firmware processing DONE. Type:", fwType )
		end
		return r, reason, msg
	end
end

----

function sysctrlService.DoUpgrade(loc)
	_dbg.Info( "SYSCTRL", "Upgrading from location: ", loc or "<default>" )
	local up_proc = require "system.upgradeProc"
	local r, msg = up_proc.DoUpgrade(loc)
	if not r then
		_dbg.Error( "SYSCTRL", "Fail to do upgrading, with error: ", msg )
	else
		_dbg.Info( "SYSCTRL", "Upgrading DONE." )
		if r == 9 then
			_dbg.Info( "SYSCTRL", "Upgrading needs reboot device!" )
			sysctrlService.Reboot(3) -- 3 seconds later.
		end
	end
	return r, msg
end

----

function sysctrlService.Reset()
	local r, action = pcall(require, "system.reset")
	if not r or not action then
		_dbg.Error( "SYSCTRL", "Invalid reset action declared for this board!" )
		return false
	else
		_dbg.Info( "SYSCTRL", "Reset the system." )
		return action()
	end
end

function sysctrlService.Reboot(delay)
	local r, action = pcall(require, "system.reboot" )
	if not r or not action then
		_dbg.Info( "SYSCTRL", "No special reboot action declared, so reboot directly!" )
	else
		action()
	end
	_G.g_action = "reboot"
	_G.g_actionDelay = delay
	coaf.Exit(0)
	return true
end

function sysctrlService.RestoreFactory()
	local r, action = pcall(require, "system.restore")
	if not r or not action then
		_dbg.Error( "SYSCTRL", "No special restore factory action declared!" )
		return false
	else
		action()
	end
	_G.g_action = "reboot"
	coaf.Exit(0)
	return true
end

function sysctrlService.GetDeviceInfo(lang)
	local done, oem, oeminfo
	done, oem = pcall(require, "system.oem")
	if not done or type(oem) ~= "table" then
		oem = _G
	end

	done, oeminfo = pcall(require, "system.oeminfo")
	if not done or type(oeminfo) ~= "table" then
		oeminfo = _G.OEM_INFO
	end

	local r = {
		DEVICE_TYPE = oem.DEVICE_TYPE or "UNKNOWN",
		OEM_TYPE = oem.OEM_TYPE or "UNKNOWN",
		DEVICE_NAME = "CODEC",
		DEVICE_VENDOR = "KILOVIEW LLC",
		DEVICE_COPY = "Copyright(C) 2018",
		DEVICE_URL = "http://www.kiloview.com/"
	}

	local info = oeminfo and oeminfo[ oem.OEM_TYPE or "UNKNOWN" ] or nil
	if info then
		info = info[lang or "cn"] or info["cn"]
	end

	if info then
		r.DEVICE_NAME = info.DEVICE_NAME or "CODEC"
		r.DEVICE_VENDOR = info.DEVICE_VENDOR or ""
		r.DEVICE_COPY = info.DEVICE_COPY or "Copyright(C) 2018"
		r.DEVICE_URL = info.DEVICE_URL or ""
	end
	return r
end

function sysctrlService.GetVersion()
	local done,ver = pcall(require, "version")
	if not done or type(ver) ~= "table" then
		ver = _G
	end

	local done,mod = pcall(require, "system.hardwareVersion")
	local hwVer
	if done and mod then
		hwVer = mod.GetHardwareVersion()
	end
	return {
		HARDWARE_VERSION = hwVer or "1.0",
		SOFTWARE_VERSION = ver.SOFTWARE_VERSION or "1.0.0000",
		FIRMWARE_VERSION = ver.FIRMWARE_VERSION or "0.00"
	}
end

function sysctrlService.GetSerialNo()
	local done,mod = pcall(require, "system.serialNo")
	if done and mod then
		return mod.GetSerialNo()
	else
		return sysconf.SERIAL_NO or "00000000", sysconf.SERIAL_NO_LIST
	end
end

function sysctrlService.CreateSerialNo(serialNo, ...)
	local done,mod = pcall(require, "system.serialNo")
	if done and mod then
		return mod.CreateSerialNo(serialNo, ...)
	end

	if type(serialNo) == "string" then
		local ofile = io.open( _G.globalCfgLocation .. "/sys_config.lua", "w" )
		if not ofile then
			return false, "Fail to create 'sys_config.lua'"
		end

		if not sysconf.SERIAL_NO_LIST then
			sysconf.SERIAL_NO_LIST = {}
		end
		table.insert( sysconf.SERIAL_NO_LIST, serialNo )
		sysconf.SERIAL_NO = serialNo
		ofile:write("local SYSCONF={}\n\n")
		ofile:write("SYSCONF.SERIAL_NO_LIST={\n")
		local i
		for i=1, #sysconf.SERIAL_NO_LIST do
			ofile:write("[[" .. sysconf.SERIAL_NO_LIST[i] .. "]],\n")
		end
		ofile:write( "}\n\n" )
		ofile:write( "SYSCONF.SERIAL_NO=[[" .. serialNo .. "]]\n\n" )
		ofile:write( "return SYSCONF\n\n" )
		ofile:close()
		return true
	else
		return false, "Invalid serial no"
	end
end

function sysctrlService.GetCpuLoad()
	return sysinfo.GetCpuLoad()
end

function sysctrlService.GetMemUse()
	return sysinfo.GetMemUse()
end

function sysctrlService.GetDiskSpaces()
	return sysinfo.GetDiskSpaces()
end

function sysctrlService.GetPerisTime()
	return sysinfo.GetPersisTime()
end

local function onPingNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.lossRate = msg.lossRate
end

function sysctrlService.Ping( targetAddr )
	if not targetAddr then
		return false, false, 100
	end
	local ctx = {
		status = "wait",
		lossRate = 100
	}

	local topic = tostring(ctx)
	local ticket = g_event:Subscribe( topic, onPingNotify, ctx )
	local pid = coaf.Fork( function()
		local pfile = io.popen( "/bin/ping " .. targetAddr .. " -W 1 -c 2 -q", "r" )
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local pt, pr, lr
			while line do
				--2 packets transmitted, 2 packets received, 0% packet loss
				pt, pr, lr = line:match( "(%d+).*transmitted.*(%d+).*received.*(%d+)%%.*$" )
				if pt and pr and lr then
					break
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else
				g_event:NotifyBack( topic, { status="finished", code = code, lossRate = tonumber(lr) } )
			end
		end
		os.exit(0)
	end)

	local r = coaf.WaitPID( pid )
	if not r then
		g_event:Unsubscribe( ticket )
		return false, false, 100
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.5)	--Wait a while
		end
		g_event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return true, ctx.code == 0, ctx.lossRate
		else
			return true, false, 100
		end
	end
end


local function oniperf3Notify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.ret = msg.ret
end

function sysctrlService.iperf3( targetAddr )
	if not targetAddr then
		return false, false, ""
	end
	local ctx = {
		status = "wait",
		ret = "error"
	}

	local topic = tostring(ctx)
	local ticket = g_event:Subscribe( topic, oniperf3Notify, ctx )
	local pid = coaf.Fork( function()
		-- local pfile = io.popen( "/bin/iperf3 -c " .. targetAddr .. " -i 1 -w 8k -t 1", "r" )
		local pfile = io.popen( "/usr/bin/iperf3 -c " .. targetAddr .. " -i 1 -t 2", "r" )
		
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local ret=""
			local start_i, end_j
			local unit = "M"
			local line = pfile:read("*l")
			local pt
			while line do
				--[  5]   0.00-1.00   sec  5.69 MBytes  47.8 Mbits/sec    0   14.3 KBytes
				pt = line:match( "([0-9%.]+)%s*.bits/sec" )					
				if pt then
					start_i, end_j = string.find(line, 'bits/sec')
					unit = string.sub(line, start_i-1, start_i-1)
					ret = pt .. unit .. "bits/sec"
					break
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else			
				g_event:NotifyBack( topic, { status="finished", code = code, ret = ret } )
			end
		end
		os.exit(0)
	end)

	local r = coaf.WaitPID( pid )
	if not r then
		g_event:Unsubscribe( ticket )
		return false, false, 100
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.5)	--Wait a while
		end
		g_event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return true, ctx.code == 0, ctx.ret
		else
			return true, false, ""
		end
	end
end

function sysctrlService.EnumDisks()
	local r,lfs = pcall(require, "lfs")
	if r and lfs then
		local rslt = {}
		local name
		for name in lfs.dir( "/sys/class/block" ) do
			if name ~= "." and name ~= ".." then
				if name:match("loop") or name:match("mtd") or name:match("ram") or name:match("rom") then
					--_dbg.Info( "SYSCTRL", "Enum skip block device of", name )
				else
					local pline
					local pfile = io.popen( TOOLS.UDEVADM .. " info --query=all --name=/dev/" .. name, "r" )
					if pfile then
						local line = pfile:read("*l")
						while line do
							pline = line:match("^P:%s+(.+)$")
							if pline then
								break
							end
							line = pfile:read("*l")
						end
						pfile:close()
					end

					if pline then
						table.insert( rslt, {
							device = "/dev/" .. name,
							devPath = pline
						})
					end
				end
				--Found block device.
			end
			--Valid name
		end --for.
		return true, rslt
	else
		_dbg.Error( "SYSCTRL", "Could not enum device files", r, lfs )
		return false, nil
	end
end

local function onDDNotify( ctx, topic, msg, msgType )
	ctx.status = msg.status
	ctx.code = msg.code
	ctx.rRate = msg.rRate
end

function sysctrlService.TestDiskRead(dev, blkSize, blkCnt )
	if not dev then
		return false, "0"
	end
	local ctx = {
		status = "wait",
		rRate = "0"
	}

	local topic = tostring(ctx)
	local ticket = g_event:Subscribe( topic, onDDNotify, ctx )
	local pid = coaf.Fork( function()
		local pfile = io.popen( "/bin/dd if=" .. dev .. " of=/dev/null bs=" .. (blkSize or 4096) .. " count=" .. (blkCnt or 1024) .. " 2>&1", "r" )
		if not pfile then
			g_event:NotifyBack( topic, { status="fail" } )
		else
			local line = pfile:read("*l")
			local rate
			while line do
				--40960000 bytes (39.1MB) copied, 5.025621 seconds, 7.8MB/s
				rate = line:match( ".*copied.*second.-(%d.*)/s" )
				if rate then
					break
				end
				line = pfile:read("*l")
			end
			local done, status, code = pfile:close()
			if status ~= "exit" then
				g_event:NotifyBack( topic, { status="fail" } )
			else
				g_event:NotifyBack( topic, { status="finished", code = code, rRate = rate } )
			end
		end
		os.exit(0)
	end)

	local r = coaf.WaitPID( pid )
	if not r then
		g_event:Unsubscribe( ticket )
		return false, "0"
	else
		if ctx.status == "wait" then
			coaf.Sleep(0.5)	--Wait a while
		end
		g_event:Unsubscribe( ticket )
		if ctx.status == "finished" then
			return ctx.code == 0, ctx.rRate
		else
			return false, "0"
		end
	end
end


function sysctrlService.download_bin_file(url,bin_file)
	local socket = require "socket"
	local http = require "socket.http"
	local request_body = ""
	local response_body = {}
	local res, code, response_headers = http.request{
		url = url ,
		method = "GET",
		headers={
			["Content-Type"] = "application/octet-stream",
		},
		source = ltn12.source.string(request_body),
		sink = ltn12.sink.table(response_body)
	}
	local_bin_file = "/tmp/" .. bin_file
	local f = io.open(local_bin_file,"w+")
	for k,v in pairs(response_body) do
		f:write(v)
	end
	f:close()
	local  done, reason, msg = sysctrlService.ProcessFirmware( local_bin_file )
	if not done then
		os.execute( "/bin/rm -f " .. local_bin_file )
		return true
	end

	sysctrlService.Reboot(1)
	return true
end

function sysctrlService.SetWebPort(http_port, https_port, http_open, https_open)
	local f = io.open( web_serv.config_file, "w")
	if not f then
		return false
	end

	f:write("HTTP_PORT=" .. (http_port or "80") .. "\n")
	f:write("HTTPS_PORT=" .. (https_port or "443") .. "\n")
	f:write("DISABLE_HTTP=" .. (http_open and '' or '#') .. "\n")
	f:write("DISABLE_HTTPS=" .. (https_open and '' or '#'))
	f:close()

	return true
end

function sysctrlService.GetWebPort()
	local f = io.open(web_serv.config_file, "r")
	local data = {HTTP_PORT = 80, HTTPS_PORT = 443, DISABLE_HTTP = "", DISABLE_HTTPS = ""}
	if f then
		for line in f:lines() do
			local key, value = line:match("([^=]*)=(.*)$")
			if key then
				data[key] = value
			end
		end
	end

	return { 
		http_port = data.HTTP_PORT or 80,
		https_port = data.HTTPS_PORT or 443,
		http_enable = ((not data.DISABLE_HTTP) or data.DISABLE_HTTP == ""),
		https_enable = ((not data.DISABLE_HTTPS) or data.DISABLE_HTTPS == "")
	}
end

local APIAuthorization_config = require("APIAuthorization_config").Load()
function sysctrlService.SetHTTPAPIAuthorization(skip)
	APIAuthorization_config.APIAuthorization = skip
	APIAuthorization_config:__TRIGGER(true)
	return true
end

function sysctrlService.GetHTTPAPIAuthorization()
	return APIAuthorization_config.APIAuthorization
end
---------------------------------------------------------

return sysctrlService

