local coaf = require "coaf.core"
local thread = require "coaf.thread"
local _dbg = require "coaf.debug"
local event = require "coaf.event"

local support_posix, posix_unistd = pcall(require,"posix.unistd")

local daemon_event  --Init by '_M.Init'

----------------------------------------------
local function on_coevent_notify( daemon_task, topic, msg, msgType )
	if msg then
		if msg.event_type == "~TERMINATE" then
			daemon_task.terminate_info = msg.args or {
				term_normal = true,
				exit_status = "exit",
				exit_code = 1
			}
		else
			if type(daemon_task.options.on_event) == "function" then
				daemon_task.options.on_event( daemon_task, msg.event_type, msg.args )
			end
		end
	end
end

----------------------------------------------
local event_obj_meta = {}

function event_obj_meta:Notify( event_type, args )
	local msg = { event_type = event_type, args = args }
	daemon_event:NotifyBack( self.topic, msg)
end

event_obj_meta.__index = event_obj_meta

local nofork_event_obj_meta = {}

function nofork_event_obj_meta:Notify( event_type, args )
	local msg = { event_type = event_type, args = args }
	on_coevent_notify( self.daemon_task, self.topic, msg ) --"table", but no care.
end

nofork_event_obj_meta.__index = nofork_event_obj_meta
----------------------------------------------

local function daemon_run_thread( thrObj, daemon_task )
	_dbg.Info( "CoAF.Daemon", "Daemon:", daemon_task.options.name, " task is running...")

	local eventObj = {
		topic = daemon_task.options.unique_id
	}

	if daemon_task.options.no_fork then
		eventObj.daemon_task = daemon_task
		eventObj.thread = thrObj
		setmetatable( eventObj, nofork_event_obj_meta )
	else
		setmetatable( eventObj, event_obj_meta )
	end

	while not thrObj.terminated do
		local ticket = daemon_event:Subscribe( daemon_task.options.unique_id, on_coevent_notify, daemon_task )	
		local pid = nil
		if type(daemon_task.options.on_before_start) == "function" then
			daemon_task.options.on_before_start( daemon_task )
		end

		if type(daemon_task.options.run) == "string" then
			if support_posix then
				daemon_event:CloseSlaveWrite("close-on-exec", false)
				pid = coaf.Fork( function()
					--Don't monitor the slave read in sub-process.
					daemon_event:CloseSlaveRead()
					if not daemon_task.options.args then
						daemon_task.options.args = { [0] = daemon_task.options.run }
					else
						daemon_task.options.args[0] = daemon_task.options.run
					end
					local flag, msg = posix_unistd.execp( daemon_task.options.run, daemon_task.options.args )
					_dbg.Error( "CoAF.Daemon", "ERROR of execute:", daemon_task.options.run, "\nWith message:\n", msg )
					eventObj:Notify( "~TERMINATE", {
						term_normal = false,
						exit_status = "exec-error",
						exit_code = 255,
						--msg = msg
					})
					os.exit(1)
				end )
				daemon_event:CloseSlaveWrite("close-on-exec", true)
			else
				daemon_event:CloseSlaveWrite("close-on-exec", false)
				pid = coaf.Fork( function()
					--Don't monitor the slave read in sub-process.
					daemon_event:CloseSlaveRead()

					local cmd = daemon_task.options.run
					if daemon_task.options.args then
						local i, a
						for i,a in ipairs(daemon_task.options.args) do
							cmd = cmd .. " " .. a
						end
					end
					local term_normal, exit_status, exit_code = os.execute( cmd )
					eventObj:Notify( "~TERMINATE", {
						term_normal = term_normal,
						exit_status = exit_status,
						exit_code = exit_code
					})
					os.exit(exit_code)
				end )
				daemon_event:CloseSlaveWrite("close-on-exec", true)
			end
		else
			if daemon_task.options.no_fork then
				pid = "NOFORK"
			else
				daemon_event:CloseSlaveWrite("close-on-exec", false)
				pid = coaf.Fork( function()
					--Don't monitor the slave read in sub-process.
					daemon_event:CloseSlaveRead()

					local args = daemon_task.options.args or {}
					local term_normal, exit_status, exit_code = daemon_task.options.run( eventObj, table.unpack( args ) )
					eventObj:Notify( "~TERMINATE", {
						term_normal = term_normal,
						exit_status = exit_status or "exit",
						exit_code = exit_code or 0
					})
					os.exit(exit_code or 0)
				end)
				daemon_event:CloseSlaveWrite("close-on-exec", true)
			end
		end

		if not pid then
			daemon_task.pid = nil
			daemon_task.status = "error"

			daemon_event:Unsubscribe( ticket )	

			_dbg.Error( "CoAF.Daemon", "Fail to create task for daemon:", daemon_task.options.name )
			if daemon_task.options.on_terminate then
				local act
				if type(daemon_task.options.on_terminate) == "function" then
					act = daemon_task.options.on_terminate( daemon_task, nil, "exec-error", 255 )
				else
					act = daemon_task.options.on_terminate
				end
				if act ~= "restart" then
					daemon_task.status = "terminated"
					break
				end
				daemon_task.status = "restart"
			else
				daemon_task.status = "terminated"
				break
			end

			daemon_task.options.terminate_info = nil

		else
			if pid == "NOFORK" then
				daemon_task.pid = 0
			else
				daemon_task.pid = pid
			end

			daemon_task.status = "running"
			if type(daemon_task.options.on_started) == "function" then
				daemon_task.options.on_started( daemon_task, pid == "NOFORK" and 0 or pid )
			end

			local abort, need_kill

			if pid == "NOFORK" then
				local args = daemon_task.options.args or {}
				local term_normal, exit_status, exit_code = daemon_task.options.run( eventObj, table.unpack( args ) )
				eventObj:Notify( "~TERMINATE", {
					term_normal = term_normal,
					exit_status = exit_status or "exit",
					exit_code = exit_code or 0
				})
			else
				local r, err = coaf.WaitPID( pid )
				if not r and err == "#CoAF-KILL#" then
					need_kill = true
				end
			end

			abort = need_kill or thrObj.terminated

			daemon_task.status = "terminated"
			daemon_task.pid = nil

			if need_kill then
				--First set this status, but may changed by killed task.
				if not daemon_task.terminate_info then
					daemon_task.terminate_info = {
						term_normal = false,
						exit_status = "kill",
						exit_code = 0
					}
				end

				if support_posix then
					local sig = require "posix.signal"
					sig.kill( pid, daemon_task.options.kill_signal or sig.SIGTERM )
				else
					if daemon_task.options.kill_signal then
						os.execute( "/bin/kill -" .. daemon_task.options.kill_signal .. " " .. pid )
					else
						os.execute( "/bin/kill " .. pid )
					end
				end
				coaf.Sleep(0.5) --Wait a while, for getting notify message if possible.
			else
				coaf.Sleep(0.1) --Wait a while, for getting notify message if possible.
			end

			daemon_event:Unsubscribe( ticket )	

			if daemon_task.options.on_terminate then
				local act
				if type(daemon_task.options.on_terminate) == "function" then
					if not daemon_task.terminate_info then
						act = daemon_task.options.on_terminate( daemon_task, true,
														   abort and "kill" or "exit",
														   0 )
					else
						act = daemon_task.options.on_terminate( daemon_task, daemon_task.terminate_info.term_normal,
														   abort and "kill" or daemon_task.terminate_info.exit_status,
														   daemon_task.terminate_info.exit_code )
					end
				else
					act = daemon_task.options.on_terminate
				end
				if abort or act ~= "restart" then
					break
				end
				daemon_task.status = "restart"
			else
				break
			end
			daemon_task.options.terminate_info = nil
		end
	end
	daemon_task.task_thread = nil
	_dbg.Info( "CoAF.Daemon", "Daemon:", daemon_task.options.name, " task is terminated.")
end

----------------------------------------------
local daemon_meta = {}

----
function daemon_meta:Wait( timeout )
	if self.task_thread then
		--TODO: Improve this: now is simple query the status.
		while self.task_thread and (not timeout or timeout > 0) do
			local waittime = 0.1
			if timeout then
				if waittime > timeout then
					waittime = timeout
				end
				timeout = timeout - waittime
			end

			coaf.Sleep(waittime)
			if not self.task_thread then
				return "done"
			end
		end
		return "timeout"
	else
		return "done"
	end
end

----
function daemon_meta:Kill( signal )
	if self.task_thread then
		local old_set_signal = self.options.kill_signal
		if signal then
			self.options.kill_signal = signal
		end
		self.task_thread:Terminate()
		self.options.kill_signal = old_set_signal
	end
end

----
function daemon_meta:Destroy()
	if self.task_thread then
		self.task_thread:Terminate()
	end
end

----
function daemon_meta:Status()
	return self.status
end

----
function daemon_meta:GetPID()
	return self.pid
end

----
function daemon_meta:Start()
	if not self.task_thread then
		self.task_thread = thread.New( daemon_run_thread )
		if not self.task_thread then
			_dbg.Error( "CoAF.Daemon", "Fail to create daemon task for:", self.options.name )
			return false
		else
			self.task_thread:Start( self )
			return true
		end
	else
		return true
	end
end

daemon_meta.__index = daemon_meta

----------------------------------------------

local _M = {}

function _M.Init()
	if not daemon_event then
		-- Don't start slave thread, since we never use it.
		-- Using the new feature: "pipe" event instead of "coevent"
		daemon_event = event.createNew( "_coaf_run_daemon", "pipe" )
		daemon_event:CloseSlaveRead("close-on-exec", true)
		--NOTE: as default, the 'slave-write' socket is close-on-exec too,
		--      but when forking it, it will temporary change to 0.
		daemon_event:CloseSlaveWrite("close-on-exec", true)
	end
end

-- Create a daemon run task
-- options:
-- {
--     name = "optional-task-name",
--     run = "command", --or
--     run = function(eventObj, ...),
--           --NOTE: run should return term_normal, exit_status, exit_code
--           --      like 'os.execute()'
--
--     args = {
--         command-arguments-ARRAY,
--         passed to 'command' or as '...' arguments
--         for run function.
--     },
--
--     no_fork = true | false, --Don't fork sub-process, that means run
--                               in the current process as coroutine.
--
--     kill_signal = <SIGNAL>, --default is SIGTERM
--
--     --optional callbacks:
--     on_before_start = callback( daemon_task_obj ),
--     on_started = callback( daemon_task_obj, pid ),
--
--     on_terminate = callback( daemon_task_obj, 
--                              term_normal, --These 3 values is most 
--                                           --like the return values 
--                                           --of 'os.execute'
--                              exit_status,
--                              exit_code
--                            ), 
--
--     --or:
--
--     on_terminate = "restart", --up 'callback' also can returns 
--                               --'restart' to restart task.
--     
--     on_event = callback( daemon_task_obj, event_id, event_args )
-- }

function _M.Create( options )
	if not options or
	   ( type(options.run) ~= "string" and type(options.run) ~= "function" ) then
	   _dbg.Error( "CoAF.Daemon", "Invalid daemon 'run' argument!" )
	   return nil
	end

	if options.no_fork and type(options.run) ~= "function" then
	   _dbg.Error( "CoAF.Daemon", "While in set 'no_fork' mode, the 'run' must be a function!" )
		return nil
	end

	local daemon_task = {
		options = options,
		status = "idle"
	}

	options.unique_id = tostring(daemon_task):match(":%s+(.+)$")

	--Make a readable name.
	if not options.name then
		if type(options.run) == "string" then
			options.name = options.run
		else
			options.name = options.unique_id
		end
	end

	return setmetatable( daemon_task, daemon_meta )
end

return _M

