local compa = require "coaf.compatible"
local coaf = require "coaf.core"
local support_posix, unistd = pcall(require, "posix.unistd")
local fcntl
local signal
local errno

if support_posix then
	fcntl = require "posix.fcntl"
	signal = require "posix.signal"
	errno = require "posix.errno"
else
	unistd = nil
end

local _M = {
}

local xfile_meta = {}

--This is most like io.file:read(), but:
--  does not support 'n' option.
--

function xfile_meta:__wait(time,coro)
	if coro then 
		self.waitting_coro = coro
	end
	local abort, reason = coaf.Sleep(time)
	if coro then
		self.waitting_coro = nil
	end
	if abort and reason ~= "got-data" then
		self.EOF = true
		return false
	else
		return true
	end
end

function xfile_meta:__checkline()
	local i, buf, ret, got
	local pre, nxt
	for i,buf in ipairs(self.buffers) do
		pre,nxt = buf:match("(.-)[\r]?[\n](.*)$")
		if pre then
			if #nxt > 0 then
				self.buffers[i] = nxt --Replace current line.
				got = i-1
			else
				self.buffers[i] = ''
				got = i
			end
			break
		end
	end
	if got then
		ret=''
		for i=1,got do
			ret = ret .. self.buffers[1]
			table.remove(self.buffers,1)
		end
		return ret .. pre
	else
		return nil
	end
end

function xfile_meta:__checkout(bytes)
	if bytes <= 0 then
		return ''
	end

	local cnt = bytes
	local i, buf, ret, got
	local pre, nxt
	for i,buf in ipairs(self.buffers) do
		if #buf >= cnt then
			cnt = 0
			pre = buf:sub(1,cnt)
			nxt = buf:sub(cnt+1)
			if #nxt > 0 then
				self.buffers[i] = nxt --Replace current line.
				got = i-1
			else
				self.buffers[i] = ''
				got = i
			end
			break
		else
			cnt = cnt - #buf
		end
	end
	if got then
		ret=''
		for i=1,got do
			ret = ret .. self.buffers[1]
			table.remove(self.buffers,1)
		end
		return ret .. pre
	else
		return nil
	end
end

function xfile_meta:read(...)
	local opts = {...}
	local rets
	if #opts == 0 then
		opts[1] = "l"
	end

	local coro, ismain = coroutine.running()
	if ismain then
		coro = nil
	end

	local i,opt  
	local abort, reason
	for i,opt in ipairs(opts) do
		if opt == "a" or opt == "*a" then
			while not self.EOF do
				if not self:__wait(self.timeout or 1, coro) then
					break
				end
			end

			local full_cnt = table.concat(self.buffers)
			if #self.buffers > 0 then
				self.buffers = {}
			end

			if rets then
				return table.unpack(rets), full_cnt
			else
				return full_cnt
			end

		elseif opt == "l" or opt == "*l" or opt == "L" or opt == "*L" then
			repeat
				local line = self:__checkline()
				if line then
					if opt == "L" or opt == "*L" then
						line = line .. "\n"
					end
					if not rets then
						rets = {line}
					else
						table.insert(rets,line)
					end
					break
				elseif self.EOF then
					if #self.buffers > 0 then
						full_cnt = table.concat(self.buffers)
						if opt == "L" or opt == "*L" then
							full_cnt = full_cnt .. "\n"
						end
						self.buffers = {}
						if rets then
							return table.unpack(rets), full_cnt
						else
							return full_cnt
						end
					else
						if rets then
							return table.unpack(rets)
						else
							return nil
						end
					end
				else
					if not self:__wait(self.timeout or 1, coro) then
						break
					end
				end
			until false;

		elseif type(opt) == "number" and opt >= 0 then
			repeat
				local line = self:__checkout(opt)
				if line then
					if not rets then
						rets = {line}
					else
						table.insert(rets,line)
					end
					break
				elseif self.EOF then
					if #self.buffers > 0 then
						full_cnt = table.concat(self.buffers)
						self.buffers = {}
						if rets then
							return table.unpack(rets), full_cnt
						else
							return full_cnt
						end
					else
						if rets then
							return table.unpack(rets)
						else
							return nil
						end
					end
				else
					if not self:__wait(self.timeout or 1, coro) then
						break
					end
				end
			until false;
			
		else
			table.insert(rets, nil)
		end
	end
	if not rets then
		return nil
	else
		return table.unpack(rets)
	end
end

function xfile_meta:close(kill_signal)
	if self.read_fd then
		unistd.close(self.read_fd)
		self.read_fd = nil
	end

	if self.EXIT then
		return true, self.EXIT, self.EXIT_CODE
	end

	if self.pid then
		self.EXIT = "exit"
		signal.kill( self.pid, kill_signal or signal.SIGTERM )
		self.pid = nil
	end

	if not self.EOF then
		self.EOF = true
	end

	if self.waitting_coro then
		coroutine.resume( self.waitting_coro, nil, "#CoAF-KILL#" )
	end

	self.buffers = {}

	return true, self.EXIT or "error", self.EXIT_CODE or 255
end

function xfile_meta:get_pid()
	return self.pid
end

function xfile_meta:write()
	--Nothing, only stub.
end

xfile_meta.__index = xfile_meta

function _M.xpopen(cmd,args,rw,timeout)
	if not support_posix then
		return nil, "does not support"
	end

	if type(cmd) ~= "string" then
		return nil, "invalid command"
	end

	local ev = require "ev"

	local read_fd, write_fd = unistd.pipe()
	if not read_fd or not write_fd then
		return nil, "fail to create IO pipe for 'xpopen'"
	end

	if not args then
		args = {}
	end
	args[0] = cmd

	--Non-blocking read pipe.
	local flag = fcntl.fcntl( read_fd, fcntl.F_GETFL )
	if compa.lua_version >= 5.3 then
		flag = flag | fcntl.O_NONBLOCK
	elseif compa.bit then
		flag = compa.bit.bor( flag, fcntl.O_NONBLOCK )
	end
	fcntl.fcntl( read_fd, fcntl.F_SETFL, flag )
	fcntl.fcntl( read_fd, fcntl.F_SETFD, 1 )

	--Non-blocking write pipe.
	flag = fcntl.fcntl( write_fd, fcntl.F_GETFL )
	if compa.lua_version >= 5.3 then
		flag = flag | fcntl.O_NONBLOCK
	elseif compa.bit then
		flag = compa.bit.bor( flag, fcntl.O_NONBLOCK )
	end
	fcntl.fcntl( write_fd, fcntl.F_SETFL, flag )

	local pid = coaf.Fork( function()
		unistd.dup2( write_fd, 1 )	-- stdout
		--unistd.dup2( write_fd, 2 )	-- stderr
		unistd.close( write_fd )
		unistd.close( read_fd )
		unistd.execp( cmd, args )
		--In normally, never here.
		os.exit(1)
	end)

	--Write fd is no used for me.
	unistd.close(write_fd)
	if not pid then
		unistd.close(read_fd)
		return nil, "fail to execute command"
	end

	local xfile = {
		pid = pid,
		read_fd = read_fd,
		command = cmd,
		args = args,
		timeout = timeout,
		
		EOF = false,
		EXIT = nil,
		EXIT_CODE = nil,

		buffers = {
		}
	}

	local task = coaf.NewIO( function(taskobj)
		while true do
			local buf,e,code = unistd.read( read_fd, 1024 )
			if not buf then
				if code ~= errno.EAGAIN and code ~= errno.EWOULDBLOCK then
					xfile.EOF = true
					taskobj:destroy()
					xfile.read_task = nil
				end
				break
			elseif #buf == 0 then
				xfile.EOF = true
				taskobj:destroy()
				xfile.read_task = nil
				break
			end
			table.insert( xfile.buffers, buf )
		end

		if xfile.waitting_coro then
			coroutine.resume( xfile.waitting_coro, nil, "#CoAF-KILL#", "got-data" )
		end
	end, read_fd, ev.READ )

	if not task then
		signal.kill( pid, 1 ) --Force kill, TODO here.
		unistd.close(read_fd)
		return nil, "fail to start read"
	end

	xfile.read_task = task

	task = coaf.NewChild( function(cur_task, revent)
		xfile.EOF = true
		xfile.EXIT = xfile.EXIT or "exit"
		xfile.EXIT_CODE = xfile.EXIT_CODE or 0 --TODO

		if xfile.read_task then
			xfile.read_task:destroy()
			xfile.read_task = nil
		end
		cur_task:stop()

		if xfile.waitting_coro then
			coroutine.resume( xfile.waitting_coro, nil, "#CoAF-KILL#" )
		end

		xfile.child_task = nil
	end, pid )

	--TODO: not check task yet. I ASSUME it will always success.
	xfile.child_task = task

	xfile.read_task:start()
	xfile.child_task:start()

	setmetatable(xfile, xfile_meta)
	return xfile
end

return _M

