--[==[

The Co-Application Framework (CoAF) core.

The CoAF modules *MUST* depend on 'lua-ev' module.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local ev = require "ev"

----------------------------------------------------------
-- Internal, signal name mapping.
-----------------------------------------------------------

local function getSignal(sig)
	if not sig then
		_dbg.Debug("CoAF", "Invalid signal specified: <nil>")
		return nil
	end

	if type(sig) == "string" then
		sig = sig:upper()
		if not sig:match("^SIG") then
			sig = "SIG" .. sig
		end
		sig = ev[sig]
		if not sig then
			_dbg.Debug("CoAF", "Invalid signal specified" )
			return nil
		else
			return sig
		end
	else
		return sig
	end
end

----------------------------------------------------------

local _M = {
	CoAF_mainThread = nil,
	CoAF_pidLocation = "/var/run",
	CoAF_pathDelim = "/",
	CoAF_pidFile = nil,
	CoAF_LOOP = nil,
	CoAF_osType = nil,
	CoAF_appName = nil,
	CoAF_deadSignals = {},
	CoAF_fork = nil,
	CoAF_sys_sleep = nil,
	CoAF_namedDelayTasks = {},

	__exit_callbacks = {}
}

----------------------------------------------------------
-- SO you can use CoAF.ev.READ, CoAF.ev.WRITE, ...
----------------------------------------------------------
_M.ev = ev

----------------------------------------------------------

local _dbg = require "coaf.debug"

--_G.CoAF_LOOP = ev.Loop.default	-- as default. (don't init it directly!)

-----------------------------------------------------------
-- ** CoAF.OSType():
--
-- Get OS Type.
-----------------------------------------------------------

function _M.OSType()
	if not _M.CoAF_osType then
		local bf = package.cpath:match("%p[\\|/]?%p(%a+)")
		if bf == "so" then
			_M.CoAF_osType = "Linux"
		elseif bf:lower() == "dll" then
			_M.CoAF_osType = "Windows"
			_M.CoAF_pathDelim = "\\"
		elseif bf:lower() == "dylib" then
			_M.CoAF_osType = "MacOS"
		else --Assume it be Linux
			_M.CoAF_osType = "Linux"
		end 
	end
	return _M.CoAF_osType
end

function _M.GetAppDir(progArg)
	if progArg and progArg[0] then
		local dir = progArg[0]:match("^(.+)/.*")
		if dir then
			return dir
		else
			return "."
		end
	else
		return "."
	end
end

function _M.ChDir(dir)
	if not dir then
		return true
	end
	local status, unistd = pcall(require, "posix.unistd")
	if not status then
		return nil, unistd
	end
	local err
	status, err = unistd.chdir(dir)
	if not status then
		return nil, err
	else
		return true
	end
end

function _M.GetCwd()
	local status, unistd = pcall(require, "posix.unistd")
	if not status then
		return nil, unistd
	end
	return unistd.getcwd()
end

-----------------------------------------------------------
function _M.GetHostname()
	if _M.OSType() == "Windows" then
		return nil
	end

	local r
	local f = io.popen( "/bin/hostname", "r" )
	if f then
		r = f:read("*l")
		f:close()
	end
	if not r then
		f = io.open( "/etc/hostname", "r" )
		if f then
			r = f:read("*l")
			f:close()
		end
	end
	if not r or r == "" then
		return "none"
	else
		return r
	end
end

-----------------------------------------------------------
function _M.SetHostname(name)
	if _M.OSType() == "Windows" then
		return nil, "Unsupported hostname set in Windows"
	end

	if name then
		local f = io.open( "/etc/hostname", "w" )
		if f then
			f:write(name .. "\n")
			f:close()
		end
		os.execute( "/bin/hostname '" .. name .. "'" )
		return name
	else
		return nil, "Invalid host name"
	end
end
-----------------------------------------------------------
-- ** CoAF.Init([name[, loop]]):
--
-- Initialize, specify application name (optional), and 
-- the event loop object (optional)
-----------------------------------------------------------

function _M.Init(name, loop)
	--Detect OS type first.
	_M.OSType()

	if not name then
		name = _G.arg[0]
	end
	
	if name then
		local x = name:match(".+[\\|/](.*)")
		if x then
			_M.CoAF_appName = x
		else
			_M.CoAF_appName = name
		end
	end

	if not _M.CoAF_LOOP then
		if loop then
			_M.CoAF_LOOP = loop
		else
			_M.CoAF_LOOP = ev.Loop.default
		end
	end
	
	--For some compatiable reason, but strongly suggest don't
	--use the global variant.

	_G.CoAF_LOOP = _M.CoAF_LOOP

	-- Clear return code.
	_M.CoAF_returnCode = nil

	local st
	local unistd
	st, unistd = pcall(require,"posix.unistd")
	if not st or not unistd then
		_dbg.Warn("CoAF", "The 'posix.unistd' library is not found, so the CoAF.Fork() is not supported.")
	else
		_M.CoAF_fork = unistd.fork
	end

--[[
	local socket
	st, socket = pcall(require,"socket")
	if st and socket then
		_dbg.Info("CoAF", "Select socket.sleep() for CoAF.Sleep()")
		_M.CoAF_sys_sleep = socket.sleep
	else
		st, unistd = pcall(require,"posix.time")
		if unistd then
			_dbg.Info("CoAF", "Select posix for CoAF.Sleep()")
			_M.CoAF_sys_sleep = function(seconds)
				local sec = math.floor(seconds)
				seconds = seconds - sec
				local spec = {tv_sec=sec, tv_nsec = math.floor(seconds*1000000000) }
				unistd.nanosleep(spec)
			end
		else
			_dbg.Warn("CoAF", "No supported system sleep routines found ('posix' or 'socket')")
		end
	end
]]
end

-----------------------------------------------------------
-- ** CoAF.AddDeadSignal( signal[, returnValue] ):
--
-- Add a dead signal (Which signal could cause application
-- terminate.). Optionally, you can specify what return
-- value (Exit value) when this signal handled.
--
-- The dead signal can be number (signal num), or standard
-- signal name, e.g. SIGINT, SIGTERM, ... (Also you can
-- not write the 'SIG' prefix, so it's more simple like 
-- INT, TERM, ... )
-----------------------------------------------------------

function _M.AddDeadSignal( sig, retVal )
	if not _M.CoAF_LOOP then
		_dbg.Warn("CoAF", "CoAF is not initialized yet.")
		return false
	end
	local signum = getSignal(sig)
	if not signum then
		return false
	end
	if _M.CoAF_deadSignals[signum] then
		if retVal then
			_M.CoAF_deadSignals[signum].returnValue = retVal
		end
	else
		local task = ev.Signal.new(function(loop,sig)
			-- sig:stop(loop)
			_dbg.Warn("CoAF", "!! Got dead signal of ", signum )
			local ctx = _M.CoAF_deadSignals[signum]
			if ctx then
				_M.Exit(ctx.returnValue)
			else
				_M.Exit(1)
			end
		end, signum)
		if not task then
			_dbg.Error("CoAF", "Fail to create dead signal handler:", signum)
			return false
		end
		_M.CoAF_deadSignals[signum] = {
			task = task,
			returnValue = retVal or 1
		}
		task:start(_M.CoAF_LOOP)
		return true
	end
end

-----------------------------------------------------------
-- ** CoAF.RemoveDeadSignal( ... signals ):
--
-- Remove dead signal(s). Multi arguments accepted, each one
-- is a signal number or signal name which want to remove.
-- 
-- If signal specified '*', that's remove all dead signals
-- have already added.
--
-- By default, CoAF added these signals as dead signals:
-- INT, TERM, QUIT, SEGV, PIPE
-----------------------------------------------------------
function _M.RemoveDeadSignals(...)
	if not _M.CoAF_LOOP then
		_dbg.Warn("CoAF", "CoAF is not initialized yet.")
		return false
	end
	local args = {...}
	local i
	for i=1, #args do
		local sig = args[i]
		if sig == '*' then
			local k,v
			for k,v in pairs(_M.CoAF_deadSignals) do
				if v.task then
					v.task:stop(_M.CoAF_LOOP)
				end
			end
			_M.CoAF_deadSignals = {}
			break
		else
			sig = getSignal(sig)
			if sig and _M.CoAF_deadSignals[sig] then
				local wa = _M.CoAF_deadSignals[sig].task
				if wa then
					wa:stop(_M.CoAF_LOOP)
				end
				_M.CoAF_deadSignals[sig] = nil
			end
		end
	end
	return true
end

-----------------------------------------------------------
-- ** CoAF.SetPIDFileLocation(location)
-- 
-- Specifiy the default PID file create location. Default 
-- is "/var/run".
-----------------------------------------------------------

function _M.SetPIDFileLocation(loc)
	_M.CoAF_pidLocation = loc or "/var/run"
end

-----------------------------------------------------------
-- ** CoAF.CreatePIDFile( [pid-filename] )
--
-- Create a pid file. If 'pid-filename' specified, the created
-- pid file name is that; otherwise, it create the pid file
-- locate at <PIDFileLocation>/<appname>.pid
-----------------------------------------------------------
function _M.CreatePIDFile( file )
	if _M.CoAF_pidFile then
		return _M.CoAF_pidFile
	end

	if _M.OSType() ~= "Linux" then
		_dbg.Warn("CoAF", "Unsupport PID file for OS: " .. _M.OSType())
		return nil
	end

	if not file then
		local name = _M.CoAF_appName
		if not name then
			if type(_G.arg) ~= "table" or not _G.arg[0] then
				_dbg.Error("CoAF", "Invalid application name to create PID file")
				return nil
			end
			name = _G.arg[0]
			if name then
				local x = name:match(".+[\\|/](.*)")
				if x then name = x end
			end
		end

		if not name then
			return nil
		end

		file = (_M.CoAF_pidLocation or ".") .. "/" .. name .. ".pid"
	end

	local pidf = io.open("/proc/self/stat","r")
	if not pidf then
		_dbg.Error("CoAF","Could not get current PID")
		return nil
	end

	local info = pidf:read("*l")
	if not info then
		pidf:close()
		_dbg.Error("CoAF", "Could not get current PID")
		return nil
	end

	pidf:close()
	local pid = info:match("^(%d*)")
	if not pid then
		_dbg.Error("CoAF", "Could not get current PID")
		return nil
	end

	pidf = io.open(file, "w")
	if not pidf then
		_dbg.Error("CoAF", "Fail to create pid file '" .. file .. "'")
		return nil
	end
	pidf:write('' .. pid)
	pidf:close()

	_M.CoAF_pidFile = file
	return true
end

-----------------------------------------------------------
-- ** CoAF.Exit([returnCode]):
-- 
-- Call this to stop current app.
-----------------------------------------------------------

function _M.Exit(code)
	_M.CoAF_returnCode = code or 0
	
	-- Try terminate main thread, if possible.
	if _M.CoAF_mainThread then
		_M.CoAF_mainThread:Terminate()
	end

	local i,item
	for i,item in ipairs(_M.__exit_callbacks) do
		if type(item.callback) == "function" then
			item.callback(item.uData)
		end
	end

	if _M.CoAF_LOOP then
		--_dbg.Info("CoAF", "Exit with code:", code)
		_M.CoAF_LOOP:unloop("ALL")
	end
end

-----------------------------------------------------------
-- ** CoAF.OnExit(callback[,userData])
-- 
-- Handle a callback when program exit (by coaf.Exit())
-- Allow multiple times.
-----------------------------------------------------------
function _M.OnExit(callback, uData)
	local key = tostring(callback) .. tostring(uData)
	if _M.__exit_callbacks[key] then
		return
	end
	local item = {
		callback = callback,
		uData = uData
	}
	_M.__exit_callbacks[key] = item
	table.insert(_M.__exit_callbacks, item)
end

-----------------------------------------------------------
-- ** CoAF.Start([entryFunc[,...]):
-- 
-- Start the CoAF application, enter the event loop.
--
-- If 'entryFunc' is not specified, first it try search the
-- global function 'CoMain'. If found 'CoMain', pick it as
-- the main entry function. Otherwise, no main entry, but it
-- continue starts the event loop.
--
-- The 'entryFunc' should get at least 1 argument: an array 
-- contains all command line arguments; Or if you passed argu-
-- ments to Start, the arguments are passed to 'entryFunc'
-----------------------------------------------------------

function _M.Start(mainEntry, passArgs, ...)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Invalid main loop. Init() first before start!")
		return 1
	end

	_M.AddDeadSignal("INT", 1)
	_M.AddDeadSignal("QUIT", 1)
	_M.AddDeadSignal("TERM", 1)
	_M.AddDeadSignal("SEGV", 1)
	--_M.AddDeadSignal("PIPE", 1)

	if not mainEntry then
		mainEntry = _G.CoMain
	end

	if type(mainEntry) == "function" then
		local thread = require "coaf.thread"
		_M.CoAF_mainThread = thread.New(mainEntry, true)  --Don't pass self.
	end

	if _M.CoAF_mainThread then
		local args = {...}
		if not passArgs and #args == 0 then
			_M.CoAF_mainThread:Start(_G.arg)
		else
			_M.CoAF_mainThread:Start(passArgs, ...)
		end
	end

	_M.CoAF_LOOP:loop()

	_M.CoAF_mainThread = nil

	if _M.CoAF_pidFile then
		os.remove(_M.CoAF_pidFile)
		_M.CoAF_pidFile = nil
	end

	return _M.CoAF_returnCode or 0
end

----------------------------------------------------------------
local evWatcher_wrap_meta = {
}

function evWatcher_wrap_meta:__index(key)
	return evWatcher_wrap_meta[key] or self.watcher[key]
end

--[[ nothing
function evWatcher_wrap_meta:__newindex(key,val)
end
]]

function evWatcher_wrap_meta:start(is_daemon)
	return self.watcher:start( self.loop, is_daemon )
end

function evWatcher_wrap_meta:destroy()
	if self.watcher:is_pending() then
		self.watcher:clear_pending(self.loop)
	end
	return self.watcher:stop(self.loop)
end

function evWatcher_wrap_meta:stop()
	return self.watcher:stop(self.loop)
end

function evWatcher_wrap_meta:again(seconds)
	return self.watcher:again(self.loop, seconds)
end

local function wrapWatcher(wa, loop)
	if not wa then
		return nil
	end
	if not loop then
		loop = _M.CoAF_LOOP
	end

	return setmetatable({
		watcher = wa,
		loop = loop
	}, evWatcher_wrap_meta)
end
----------------------------------------------------------------
-- CoAF.NewTimer

function _M.NewTimer(func, after_seconds, repeat_seconds)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Fail to create new timer: CoAF is not initialized.")
		return nil
	end

	if type(func) ~= "function" then
		_dbg.Error("CoAF", "Invalid callback function")
		return nil
	end

	local wrapper = {
		watcher = nil,
		loop = _M.CoAF_LOOP
	}
	local wa = ev.Timer.new( function(loop, tmr, revents)
		func(wrapper, revents)
	end, after_seconds, repeat_seconds )
	if wa then
		wrapper.watcher = wa
		setmetatable(wrapper, evWatcher_wrap_meta)
		return wrapper
	else
		_dbg.Error("CoAF", "Fail to create new timer.")
		return nil
	end
end


----------------------------------------------------------------
-- CoAF.NewIO
function _M.NewIO(func, fd, revents)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Fail to create new IO watcher: CoAF is not initialized.")
		return nil
	end

	if type(func) ~= "function" then
		_dbg.Error("CoAF", "Invalid callback function")
		return nil
	end

	local wrapper = {
		watcher = nil,
		loop = _M.CoAF_LOOP
	}
	local wa = ev.IO.new( function(loop, io, revents)
		func(wrapper, revents)
	end, fd, revents )
	if wa then
		wrapper.watcher = wa
		setmetatable(wrapper, evWatcher_wrap_meta)
		return wrapper
	else
		_dbg.Error("CoAF", "Fail to create new IO watcher.")
		return nil
	end
end

----------------------------------------------------------------
-- CoAF.NewIdle
function _M.NewIdle(func)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Fail to create new idle task: CoAF is not initialized.")
		return nil
	end

	if type(func) ~= "function" then
		_dbg.Error("CoAF", "Invalid callback function")
		return nil
	end

	local wrapper = {
		watcher = nil,
		loop = _M.CoAF_LOOP
	}
	local wa = ev.Idle.new( function(loop, idle, revents)
		func(wrapper, revents)
	end )

	if wa then
		wrapper.watcher = wa
		setmetatable(wrapper, evWatcher_wrap_meta)
		return wrapper
	else
		_dbg.Error("CoAF", "Fail to create new idle task.")
		return nil
	end
end

----------------------------------------------------------------
-- CoAF.NewSignal
function _M.NewSignal(func, sig)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Fail to create new signal watcher: CoAF is not initialized.")
		return nil
	end

	sig = getSignal(sig)
	if not sig or type(func) ~= "function" then
		_dbg.Error("CoAF", "Invalid arguments")
		return nil
	end

	local wrapper = {
		watcher = nil,
		loop = _M.CoAF_LOOP
	}
	local wa = ev.Signal.new( function(loop, sig, revents)
		func(wrapper, revents)
	end, sig )
	if wa then
		wrapper.watcher = wa
		setmetatable(wrapper, evWatcher_wrap_meta)
		return wrapper
	else
		_dbg.Error("CoAF", "Fail to create new signal watcher.")
		return nil
	end
end

----------------------------------------------------------------
-- CoAF.NewStat

function _M.NewStat(func, path, interval)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Fail to create new stat watcher: CoAF is not initialized.")
		return nil
	end

	if type(func) ~= "function" then
		_dbg.Error("CoAF", "Invalid callback function")
		return nil
	end

	local wrapper = {
		watcher = nil,
		loop = _M.CoAF_LOOP
	}
	local wa = ev.Stat.new( function(loop, stat, revents)
		func(wrapper, revents)
	end, path, interval )
	if wa then
		wrapper.watcher = wa
		setmetatable(wrapper, evWatcher_wrap_meta)
		return wrapper
	else
		_dbg.Error("CoAF", "Fail to create new stat watcher.")
		return nil
	end
end

----------------------------------------------------------------
-- CoAF.NewChild

function _M.NewChild(func, pid, trace)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Fail to create new child watcher: CoAF is not initialized.")
		return nil
	end

	if type(func) ~= "function" then
		_dbg.Error("CoAF", "Invalid callback function")
		return nil
	end

	local wrapper = {
		watcher = nil,
		loop = _M.CoAF_LOOP
	}
	local wa = ev.Child.new( function(loop, child, revents)
		func(wrapper, revents)
	end, pid, trace or false )
	if wa then
		wrapper.watcher = wa
		setmetatable(wrapper, evWatcher_wrap_meta)
		return wrapper
	else
		_dbg.Error("CoAF", "Fail to create new child watcher.")
		return nil
	end
end

-----------------------------------------------------------------
-- CoAF.Fork([child_call, ...args])
function _M.Fork( child_call, ... )
	if not _M.CoAF_fork then
		_dbg.Error("CoAF", "CoAF.Fork() not supported")
		return nil, "unsupport"
	end

	local pid
	pid, status = _M.CoAF_fork()
	if not pid then
		return nil, status
	elseif pid == 0 then -- In child
		_M.CoAF_LOOP:fork()
		if child_call then
			child_call(...)
		end
		return 0
	else
		return pid
	end
end

-----------------------------------------------------------------
-- CoAF.GetOpenedFDs()
-- Get all opened fd(s)
function _M.GetOpenedFDs()
	local x, dirent = pcall(require, "posix.dirent")
	if not x then
		return nil
	end
	local item, ret
	ret = {}
	for item in dirent.files("/proc/self/fd") do
		if item ~= "." and item ~= ".." and item ~= "0" and item ~= "1" and item ~= "2" then
			x = tonumber(item)
			if x then
				table.insert(ret,x)
			end
		end
	end
	return ret
end

--[[
--NOTE: TODO: This is for verification only, no used.
function _M.__CloseOnExec()
	local compa = require "coaf.compatible"
	local posix_fcntl = compa.CheckPosix( "posix.fcntl" )
	local posix_unistd = compa.CheckPosix( "posix.unistd" )
	if not posix_fcntl or not posix_unistd then
		return false
	end
	local fds = _M.GetOpenedFDs()
	if not fds then
		return false
	end

	local FD_CLOEXEC = posix_fcntl.FD_CLOEXEC or 1
	local i
	for i=1, #fds do
		local flag = posix_fcntl.fcntl( fds[i], posix_fcntl.F_GETFD )
		if flag then
			if compa.lua_version >= 5.3 then
				flag = flag & FD_CLOEXEC
			else
				flag = compa.bit.band( flag, FD_CLOEXEC )
			end
			if flag == FD_CLOEXEC then
				print( "**** closed:", fds[i] )
				posix_unistd.close( fds[i] )
			end
		end
	end
	return true
end
--]]

-----------------------------------------------------------------
-- CoAF.WaitPID(child_pid[, trace])

function _M.WaitPID(child, trace)
	if not _M.CoAF_LOOP then
		_dbg.Error("CoAF", "Fail to create new child watcher: CoAF is not initialized.")
		return nil
	end
	
	trace = trace or false

	if child then
		local coro, ismain = coroutine.running()
		local rslt
		local task = ev.Child.new(function(loop,wa,revent)
			wa:stop(loop)
			if ismain then
				--_dbg.Info("CoAF", "!!! Detail of WaitPID: in main task and unloop")
				loop:unloop()
				rslt = true
			else
				local r, st = coroutine.resume(coro,true)
				if not r then
					_dbg.Error("CoAF", "Thread error:", st)
				end
			end
		end, child, trace )
		task:start(_M.CoAF_LOOP)
		if ismain then
			loop:loop()
			return rslt
		else
			local r,err = coroutine.yield(coro)
			if not r and err == "#CoAF-KILL#" then
				if task:is_pending() then task:clear_pending(_M.CoAF_LOOP) end
				task:stop(_M.CoAF_LOOP)
			end
			return r,err
		end
	else
		_dbg.Error("CoAF", "Invalid child pid")
		return nil
	end
end

-----------------------------------------------------------------
-- CoAF.Sleep(seconds)

function _M.Sleep(seconds)
	-- You know, Sleep(0) is important for task switch.
	if seconds <= 0 then 
		seconds = 0 
	end

	local coro, ismain = coroutine.running()
	if ismain then
		ev.Timer.new(function(loop,tmr)
			tmr:stop(loop)
			--_dbg.Info("CoAF", "!!! Detail of Sleep: in main task and unloop")
			loop:unloop()
		end, seconds, 0):start(_M.CoAF_LOOP)
		
		-- recurise loop()
		_M.CoAF_LOOP:loop()
		--[[
		if not _M.CoAF_sys_sleep then
			_dbg.Error("CoAF", "No system sleep() support for main thread calling CoAF.Sleep()")
			return
		else
			_M.CoAF_sys_sleep(seconds)
		end
		]]
	else
		local task = ev.Timer.new(function(loop,tmr)
			tmr:stop(loop)
			local r, st = coroutine.resume(coro,true)
			if not r then
				_dbg.Error("CoAF", "Thread error:", st)
			end
		end, seconds, 0)
		task:start(_M.CoAF_LOOP)
		local r,flag,reason = coroutine.yield(coro)
		if not r and flag == "#CoAF-KILL#" then
			if task:is_pending() then task:clear_pending(_M.CoAF_LOOP) end
			task:stop(_M.CoAF_LOOP)
			--Aborted.
			return true, reason
		end
	end
end

-----------------------------------------------------------------
-- CoAF.ScheduleDelayedTask(func[, delay_seconds[, name]])

function _M.ScheduleDelayedTask(func, delay_seconds, name)
	if not func then
		return nil
	end
	if name and _M.CoAF_namedDelayTasks[name] then
		return nil
	end

	local task
	if not delay_seconds or delay_seconds <= 0 then
		task = ev.Idle.new(function(loop,wa)
			wa:stop(loop)
			if name then
				_M.CoAF_namedDelayTasks[name] = "#RUN#"
			end
			pcall( func )
			if name then
				_M.CoAF_namedDelayTasks[name] = nil
			end
		end) 
	else
		task = ev.Timer.new(function(loop,wa)
			wa:stop(loop)
			if name then
				_M.CoAF_namedDelayTasks[name] = "#RUN#"
			end
			pcall( func )
			if name then
				_M.CoAF_namedDelayTasks[name] = nil
			end
		end, delay_seconds, 0)
	end
	if not task then
		return nil
	end
	if name then
		_M.CoAF_namedDelayTasks[name] = task
	end
	task:start(_M.CoAF_LOOP)
	return true
end

-----------------------------------------------------------------
-- CoAF.ScheduleDelayedTask(func[, delay_seconds[, name]])
function _M.UnscheduleDelayedTask(name)
	if name and _M.CoAF_namedDelayTasks then
		local task = _M.CoAF_namedDelayTasks[name]
		_M.CoAF_namedDelayTasks[name] = nil
		if task then
			if task == "#RUN#" then --It's already running.
				return
			end

			if task:is_pending() then task:clear_pending(_M.CoAF_LOOP) end
			task:stop(_M.CoAF_LOOP)
		end
	end
end

return _M
