--[==[

The Co-Application Framework (CoAF) debug module.

The CoAF modules *MUST* depend on 'lua-ev' module.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local def_stdout = {
	logger = io.stdout,
	INFO = true,
	WARN = false,
	ERR = false,
	filter = {".*"}
}

local def_stderr = {
	logger = io.stderr,
	INFO = false,
	WARN = true,
	ERR = true,
	filter = {".*"}
}

local _M = {
	info_level = 3,
	debug_level = 3,
	loggers = {
		[1] = def_stdout,
		[2] = def_stderr
	},
	LF = "\r\n",
	SPACE = " ",
	global_disabled_tags = {}
}

-----------------------------------------------------------------------
-- dbg.SetLF(lf) : Set line terminator.
-----------------------------------------------------------------------

function _M.SetLF(lf)
	_M.LF = lf or "\r\n"
end

-----------------------------------------------------------------------
-- dbg.SetSpace(sp) : Set fields space char.
-----------------------------------------------------------------------

function _M.SetSpace(sp)
	_M.SPACE = sp
end

-----------------------------------------------------------------------
-- dbg.EnableConsole() : Enable output to console.
-----------------------------------------------------------------------

function _M.EnableConsole()
	_M.loggers[1] = def_stdout
	_M.loggers[2] = def_stderr
end

-----------------------------------------------------------------------
-- dbg.DisableConsole() : Disable output to console.
-----------------------------------------------------------------------
function _M.DisableConsole()
	_M.loggers[2] = nil
	_M.loggers[1] = nil
end

-----------------------------------------------------------------------
-- dbg.GlobalDisableTags( tags )
-----------------------------------------------------------------------
function _M.GlobalDisableTags(tags)
	if type(tags) ~= "table" then
		tags = { tags }
	end
	local i, tag
	for i,tag in ipairs(tags) do
		if type(tag) == "string" then
			_M.global_disabled_tags[ tag ] = true
		end
	end
end

-----------------------------------------------------------------------
-- dbg.GlobalEnableTags( tags )
-----------------------------------------------------------------------
function _M.GlobalEnableTags(tags)
	if type(tags) ~= "table" then
		tags = { tags }
	end
	local i, tag
	for i,tag in ipairs(tags) do
		if type(tag) == "string" then
			_M.global_disabled_tags[ tag ] = nil
		end
	end
end

-----------------------------------------------------------------------
-- dbg.AddLogger(name, logger [, filter_type, filter_tags ])
-- 
-- Add (register) a new logger. The logger shoud like the 'io.file',
-- support 'write' function.
--
-- 'name' is required, you should specify a name for this logger, so
-- if you want to remove it, you can specify this name.
--
-- 'filter_type' can be nil, a 'string' or an array. This argument
-- determine this logger logs what type ('err','info','warn') of 
-- information. If it's nil or string '*', means all.
--
-- 'filter_tags' can be nil, a 'string' or an array. This argument
-- determine this logger matchs those log tags (tag is a argument
-- for Error(), Info(), Warn(), ... functions). If it's nil or '*',
-- means all. 
-----------------------------------------------------------------------
function _M.AddLogger(name, logger, filtTypes, filtTags)
	if not name or not logger then
		return false
	end
	local logItem = {
		logger = logger
	}

	if not filtTypes then
		logItem.ERR = true
		logItem.INFO = true
		logItem.WARN = true
	else
		if type(filtTypes) == "string" then
			filtTypes = {filtTypes}
		end
		local i
		for i=1, #filtTypes do
			if not filtTypes[i] or filtTypes[i] == "" or filtTypes[i] == "*" then
				logItem.ERR = true
				logItem.INFO = true
				logItem.WARN = true
				break
			else
				local ft = filtTypes[i]:lower()
				if ft == "err" or ft == "error" then
					logItem.ERR = true
				elseif ft == "warn" or ft == "warning" then
					logItem.WARN = true
				elseif ft == "info" then
					logItem.INFO = true
				end
			end
		end
	end

	if not filtTags or filtTags == "" or filtTags == "*" then
		filtTags = nil
	elseif type(filtTags) == "string" then
		filtTags = { filtTags }
	elseif type(filtTags) ~= "table" then
		filtTags = nil
	end

	logItem.filter = filtTags

	_M.loggers[name] = logItem

	return true
end

-----------------------------------------------------------------------
-- dbg.RemoveLogger(name) : Remove a registered logger, by name.
-----------------------------------------------------------------------
function _M.RemoveLogger(name)
	if name then
		_M.loggers[name] = nil
	end
end

-----------------------------------------------------------------------
-- dbg.InfoLevel(level) : Set dbg.Info()'s default level. (0~3)
-----------------------------------------------------------------------
function _M.InfoLevel(level)
	if level >= 0 then
		_M.info_level = level
	end
end

-----------------------------------------------------------------------
-- dbg.DebugLevel(level) : Set debug level (0~3)
-----------------------------------------------------------------------
function _M.DebugLevel(level)
	if level >= 0 then
		_M.debug_level = level
	end
end

-----------------------------------------------------------------------
-----------------------------------------------------------------------

-- Internal using.
local function matchTag(tag, filters)
	if not tag or not filters then return true end

	local i
	for i=1, #filters do
		if tag:match(filters[i]) then
			return true
		end
	end
	return false
end

-- Internal using.
local function formatTag(pfx, tag, auxflag)
	return string.format("[%s]\t%7s: %s - ", os.date("%m%d.%H%M%S"), pfx .. (auxflag or ""), tag or "<none>" )
end

-- Internal using.
local function output(pfx,tag,auxflag,...)
	local k, logItem, i

	for k, i in pairs(_M.global_disabled_tags) do
		if tag and tag:match(k) then
			return
		end
	end

	local fmtStr
	for k, logItem in pairs(_M.loggers) do
		if logItem[pfx] and matchTag(tag, logItem.filter) then
			if not fmtStr then
				if auxflag == "@" then
					fmtStr = ''
				else
					fmtStr = formatTag(pfx,tag,auxflag)
				end
				local fmtArgs = {...}
				for i = 1, #fmtArgs do
					fmtArgs[i] = tostring(fmtArgs[i])
				end
				fmtStr = fmtStr .. table.concat( fmtArgs, _M.SPACE )
			end

			logItem.logger:write(fmtStr)
			logItem.logger:write(_M.LF)
		end
	end
end

-----------------------------------------------------------------------
-- All debug output routines:
-----------------------------------------------------------------------
function _M.Print(tag,...)
	output("INFO", tag, "@", ...)
end

function _M.Error(tag,...)
	output("ERR", tag, nil, ...)
end

function _M.Warn(tag,...)
	output("WARN", tag, nil, ...)
end

function _M.Info(tag, ...)
	if _M.debug_level >= _M.info_level then
		output("INFO", tag, nil, ...)
	end
end

function _M.Info1(tag, ...)
	if _M.debug_level >= 1 then
		output("INFO", tag, "/1", ...)
	end
end

function _M.Info2(tag, ...)
	if _M.debug_level >= 2 then
		output("INFO", tag, "/2", ...)
	end
end

function _M.Info3(tag, ...)
	if _M.debug_level >= 3 then
		output("INFO", tag, "/3", ...)
	end
end

function _M.Detail(tag, ...)
	if _M.debug_level >= 3 then
		output("INFO", tag, "/D", ...)
	end
end

return _M

