--[==[
The Co-Application Framework (CoAF) RPC connection module.
Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.
]==]

local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local protocol = require "coaf.rpc.protocol"
local debug_filt = require "coaf.rpc.debug_filter"

--Try require 'coaf.core', but not necessary.
local _flag, coaf = pcall(require, "coaf.core")
if not _flag then
	coaf = nil
end

--Try require "socket"
local socket
_flag, socket = pcall(require, "cosocket.core")
if not _flag then
	if ngx and ngx.socket then
		socket = ngx.socket
	else
		_flag, socket = pcall(require, "socket")
	end
end

local GETTIME = os.time
--[[ NOTE: socket.gettime() may return a wrong value of double
if socket and socket.gettime then
	GETTIME = socket.gettime
else
	GETTIME = os.time
end
]]

----@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

local KEEP_ALIVE_INTERVAL_DIV = 3

local _M = {
	default_keepAlive = false,
	default_keepAliveTimeout = 60
}

-------------------------------------------------------
-- connection
-------------------------------------------------------
local connection_meta = {}

function connection_meta:SetKeepAlive(opts)
	self.keepAliveOptions = opts
end

function connection_meta:Start()
	if self.thread then
		self.thread:StartNow(self)
	end
end

function connection_meta:Destroy()
	if self.thread then
		self.thread:Terminate()
	end
end

function connection_meta:Send(data, i, j) -- Try send out.
	if self.debug_TX and debug_filt.MatchStringContent(data) then
		_dbg.Print(self.debug_TAG, self.debug_NAME, "[" .. (self.Type or "LINK") .. "]", " (TX) >>>>>>" )
		_dbg.Print(self.debug_TAG, data)
	end

	while true do
		self.lastKeepAliveRequest = GETTIME()
		local x, e = self.sock:send(data, i, j)
		if not x then
			if e ~= "busy" then -- If busy, try again.
				return x, e
			else -- Only occus while under 'CoAF' environment.
				_dbg.Info("CoAF.RPC", "Connection of '" .. self.name .. "' sending is busy, try again.")
				if coaf then
					coaf.Sleep(0.01)
				end
			end
		else
			return x,e
		end
	end			
end

connection_meta.__index = connection_meta

local function ConnectionThreadTask(thrObj, connObj)
	--_dbg.Info("CoAF.RPC", "Connection task of '" .. connObj.name .. "' running.")
	
	local parser = protocol.NewParser()
	parser:EnableOriginal(connObj.debug_RX)
	local defaultTimeout = 30

	local keepAlive
	local keepAliveInterval
	local keepAliveTimeout

	local lastRxTime = GETTIME()

	while not thrObj.terminated do
		keepAlive = connObj.keepAliveOptions and connObj.keepAliveOptions.keepAlive or _M.default_keepAlive
		keepAliveTimeout = connObj.keepAliveOptions and connObj.keepAliveOptions.keepAliveTimeout or _M.default_keepAliveTimeout
		keepAliveInterval = keepAliveTimeout/KEEP_ALIVE_INTERVAL_DIV
		if keepAliveInterval <= 0 then
			keepAliveInterval = 1
		end

		if keepAlive and keepAliveInterval > 0 then
			connObj.sock:settimeout(keepAliveInterval)
		else
			connObj.sock:settimeout(defaultTimeout)
		end

		if keepAlive and (GETTIME() - connObj.lastKeepAliveRequest) >= keepAliveInterval then
			if connObj.handler.KeepAliveRequest then
				connObj.handler.KeepAliveRequest( connObj.handler, connObj )
			else
				connObj.lastKeepAliveRequest = GETTIME()
			end
		end

		local data, err, partial = connObj.sock:receive(parser.next_fetch)
		if not data then
			if err ~= "timeout" and err ~= "busy" then
				--_dbg.Info("CoAF.RPC", "Connection of '" .. connObj.name .. "' error:", err)
				break
			end
			if partial and #partial > 0 then
				data = partial
			elseif err == "timeout" then
				-- parser:Clear() -- So current parsing is terminated.
				if keepAlive and ( GETTIME() - lastRxTime ) > keepAliveTimeout then
					_dbg.Warn("CoAF.RPC", "Connection of '" .. connObj.name .. "' keep alive checking fail, aborted.")
					break
				end
			else
			--NOTE: here! Very important: (BUG found)
			-- In coaf environment, 'busy' means the socket is using by others.
			-- In this case, if this task (coroutine, or main task) does not yield
			-- out, it will fall in a dead loop: Other task will never be waked up
			-- and this loop will always continue ('busy' is always get).
				if coaf then
					coaf.Sleep(0.01)
				end
			end
		end

		if data then
			lastRxTime = GETTIME()
			local done, flag = parser:Push(data)
			if done and flag == "complete" then
				local content = parser.content
				local orig_content = parser.orig_content
				parser:Clear() -- To accept new contents.
				parser:EnableOriginal(connObj.debug_RX)

				if connObj.debug_RX and debug_filt.MatchContent(content) then
					_dbg.Print(connObj.debug_TAG, connObj.debug_NAME, "[" .. (connObj.Type or "LINK") .. "]", " (RX) <<<<<<")
					_dbg.Print(connObj.debug_TAG, table.concat( orig_content ))
				end

				if not content.SN or not content.CMD then
					_dbg.Warn("CoAF.RPC", "Invalid request got (without SN or CMD), connection:", connObj.name)
				elseif not connObj.handler.Process then
					_dbg.Warn("CoAF.RPC", "Invalid protocol handler, for connection:", connObj.name )
					if not content.CMD:match("^ACK/") then --Ignore ACK answer
						local ack = {
							SN = content.SN,
							CMD = "ACK/" .. content.CMD,
							ARGS = content.ARGS,
							headers = {
								RESULT = "ERROR",
								MSG = "Internal error: invalid handler",
								SVC = content.headers["SVC"],
								VIA = content.headers["VIA"],
								LOC = content.headers["LOC"],
								OBJ = content.headers["OBJ"]
							}
						}
						connObj:Send( table.concat(protocol.FormatContent(ack)) )
					end
				else
					connObj.handler.Process( connObj.handler, connObj, content)
				end 
			end -- End of if completed.
		end --End of get line.

	end --Thread loop.

	--NOTE: Very important: must close it, else it will leak(keep the connection long time in system)
	if connObj.sock then
		connObj.sock:close()
	end

	--_dbg.Info("CoAF.RPC", "Connection task of '" .. connObj.name .. "' terminated.")
	connObj.thread = nil
	if connObj.handler.DestroyConnection then
		connObj.handler.DestroyConnection( connObj.handler, connObj )
	end
end

----------------------------------------------------------------------------
--
-- Create a new connection:
-- 
-- sock:	A socket;
-- handler:	A object(table), should contains 2 methods:
--			1) Process(handler_self, connection, protocol_content)
--			2) DestroyConnection(handler_self, connection)
--			3) KeepAliveRequest(handler_self, connection)
-- name:	Optional. connection name.

function _M.New(sock, handler, name)

	if not sock or not handler then
		_dbg.Error("CoAF.RPC", "Invalid socket/handler for connection" )
		return nil
	end

	if not name then
		local ip,port = sock:getpeername()
		if ip or port then
			name = ip or "*"
			name = name .. (port and (":" .. port) or "*")
		end
	end
	local connectionThread = thread.New(ConnectionThreadTask)
	if not connectionThread then
		_dbg.Error("CoAF.RPC", "Fail to create connection work thread of:", name)
		return nil
	end
	local obj = {
		name = name,
		sock = sock,
		handler = handler,
		thread = connectionThread,
		keepAliveOptions = nil,
		lastKeepAliveRequest = 0,
		subscribes = {}
	}
	setmetatable(obj, connection_meta)
	return obj
end

---------------------------------------------------------------

return _M
