--[==[
The Co-Application Framework (CoAF) RPC service daemon module.
Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.
]==]

local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local connection = require "coaf.rpc.connection"
local protocol = require "coaf.rpc.protocol"
local streamer = require "coaf.rpc.streamer"
local compa = require "coaf.compatible"

local _ready, coaf = pcall( require, "coaf.core" )
if not _ready then
	coaf = nil
end

local _M = {
	__last_invoke_connection = nil
}

function _M.GetInvokingConnection()
	return _M.__last_invoke_connection
end

local function NOTIFY_invoker(current_obj, current_service, arg_str, consumer)
	local func, err = load(arg_str)
	if not func then
		return nil, "#CoAF-RPC-ERROR#", err
	end
	local status, call_args = pcall(func)
	if not status then
		return nil, "#CoAF-RPC-ERROR#", "Bad parameters"
	end
	if not call_args then
		call_args = {}
	end
	if current_obj[consumer] then
		return current_obj[consumer]( current_obj, table.unpack(call_args))
	else
		return nil
	end
end

local function CALL_invoker(current_obj, current_service, arg_str, static, target)
	local func, err = load(arg_str)
	if not func then
		return nil, "#CoAF-RPC-ERROR#", err
	end
	local status, call_args = pcall(func)
	if not status then
		return nil, "#CoAF-RPC-ERROR#", "Bad parameters"
	end
	if not call_args then
		call_args = {}
	end
	if current_obj[target] then
		if static then
			return current_obj[target]( table.unpack(call_args) )
		else
			return current_obj[target]( current_obj, table.unpack(call_args))
		end
	else
		return nil, "#CoAF-RPC-ERROR#", "Invalid call function"
	end
end

local function GET_invoker(current_obj, current_service, arg_str)
	local func, err = load(arg_str)
	if not func then
		return nil, "#CoAF-RPC-ERROR#", err
	end
	local status, call_args = pcall(func)
	if not status then
		return nil, "#CoAF-RPC-ERROR#", "Bad parameters"
	end
	if not call_args then
		call_args = {}
	end

	local rslts = {}
	local i, v
	for i,v in ipairs(call_args) do
		if type(v) == "string" then
			if type(current_obj["get_" .. v]) == "function" then
				local flag, retVal = pcall( current_obj["get_" .. v], current_obj )
				if not flag then
					_dbg.Error( "CoAF.RPC", "Call property 'get_" .. v .. "' with error:", retVal )
					table.insert( rslts, nil )
				else
					table.insert( rslts, retVal )
				end
				--table.insert( rslts, current_obj["get_" ..v](current_obj) )
			else
				table.insert( rslts, current_obj[v])
			end
		else
			table.insert(rslts, nil)
		end
	end

	return table.unpack(rslts)
end

local function SET_invoker(current_obj, current_service, arg_str)
	local func, err = load(arg_str)
	if not func then
		return nil, "#CoAF-RPC-ERROR#", err
	end
	local status, call_args = pcall(func)
	if not status then
		return nil, "#CoAF-RPC-ERROR#", "Bad parameters"
	end
	if not call_args then
		call_args = {}
	end

	local arg1_type = type(call_args[1])
	if arg1_type == "table" then
		local k,v,t
		for k, v in pairs(call_args[1]) do
			t = type(k)
			if t == "string" or t == "number" then
				if type( current_obj["set_" .. k] ) == "function" then
					--current_obj["set_"..k]( current_service, v )
					local flag, err = pcall( current_obj["set_"..k], current_obj, v )
					if not flag then
						_dbg.Error( "CoAF.RPC", "Call property 'set_" .. k .. "' error:", err )
					end
				else
					current_obj[k] = v
				end
			end
		end
	elseif arg1_type == "string" or arg1_type == "number" then
		if type( current_obj[ "set_" .. call_args[1] ] ) == "function" then
			--current_obj[ "set_" .. call_args[1] ](current_obj, call_args[2])
			local flag, err = pcall( current_obj["set_"..call_args[1]], current_obj, call_args[2] )
			if not flag then
				_dbg.Error( "CoAF.RPC", "Call property 'set_" .. k .. "' error:", err )
			end
		else
			current_obj[ call_args[1] ] = call_args[2]
		end
	end
end

local invokers = {
	["NOTIFY"] = NOTIFY_invoker,
	["CALL"] = CALL_invoker,
	["STATIC_CALL"] = CALL_invoker,
	["GET"] = GET_invoker,
	["SET"] = SET_invoker
}
-------------------------------------------------------
-- Service daemon
-------------------------------------------------------
local daemon_meta = {}

function daemon_meta:RegisterService(svcId, svcObj)
	self.services[svcId] = svcObj
end

function daemon_meta:RemoveService(svcId)
	self.services[svcId] = nil
end

function daemon_meta:SetServiceAlias(alias,svcId)
	self.aliases[alias] = svcId
end

function daemon_meta:RemoveServiceAlias(alias)
	self.aliases[alias] = nil
end

function daemon_meta:Start()
	if self.thread then
		self.thread:Start(self)
	end
end

function daemon_meta:Stop()
	if self.thread then
		self.thread:Terminate()
	else
		-- Kill all peers.
		local k,v 
		for k,v in pairs(self.peers) do
			v:Destroy()	
		end
		self.peers = {}
	end
end

function daemon_meta:AddSpecialConnection(sock,name,deadCallback)
	if not name then
		local ip,port = sock:getsockname()
		if ip or port then
			name = ip or "*"
			name = name .. (port and (":" .. port) or "*")
		end
	end
	
	if self.peers[name] then
		self.peers[name]:Destroy()
		self.peers[name] = nil
	end

	local conn = connection.New(sock, self, name)
	if not conn then
		_dbg.Error("CoAF.RPC", "Fail to add special connection of:", name)
		return nil
	else
		self.peers[name] = conn
		self.deadCallbacks[name] = deadCallback
	end
	conn:Start()
	return conn
end

local function processSTREAM( svc, connObj, content )
	local invalid_args
	local action
	if not content.ARGS or not content.headers.OBJ or not content.headers.VIA or content.headers.VIA == "" then
		_dbg.Error( "CoAF.RPC", "Invalid STREAM protocol contents - ARGS:", content.ARGS, ", OBJ:", content.headers.OBJ, ", VIA:", content.headers.VIA )
		invalid_args = "Invalid STREAM protocol contents"
	elseif content.ARGS == "connect" or content.ARGS == "disconnect" or content.ARGS == "keep-alive" or content.ARGS == "content" then
		action = content.ARGS
	else
		_dbg.Error( "CoAF.RPC", "Invalid STREAM action:", content.ARGS )
	end

	--Check args or action flag.
	if invalid_args or not action then
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "ERROR",
				MSG = invalid_args or "Invalid STREAM action"
			}
		}
		
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )
		return
	end

	--Get service.
	local obj = svc:GetObject( content.headers.OBJ )
	if not obj then
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "ERROR",
				MSG = "Invalid object"
			}
		}
		
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )
		return
	end

	--Create streamer.
	local s,create_flag = streamer.CreateStreamer( content.headers.VIA, connObj )
	if not s then
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "ERROR",
				MSG = "Fail to create streamer object:" .. (create_flag or "UNKNOWN ERROR")
			}
		}
		
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )
		return
	end

	--Router(VIA) changed for connection
	if create_flag == "new" and action ~= "connect" then
		s:Destroy(false) --Not send disconnect message.
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "ERROR",
				MSG = "New router for STREAM, must reconnect"
			}
		}
		
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )
		return
	end

	if action == "connect" then
		local call_rets, flag, msg
		if obj.OnConnect then
			local genCodes = { "return " }
			if type(content.data) == "table" then
				table.insert( genCodes, table.concat( content.data ) )
			elseif type(content.data) == "string" then
				table.insert( genCodes, content.data )
			else
				table.insert(genCodes, "{}")
			end

			local func, err = load(table.concat(genCodes))
			if not func then
				call_rets, flag, msg = nil, "#CoAF-RPC-ERROR#", err
			else
				local status, call_args = pcall(func)
				if not status then
					call_rets, flag, msg = nil, "#CoAF-RPC-ERROR#", "Bad parameters"
				else
					if not call_args then
						call_args = {}
					end

					s.LAST_SN = content.SN
					s.CMD = "ACK/" .. content.CMD
					s.OBJ = content.headers.OBJ
					s.SVC = content.headers.SVC
					s.LOC = content.headers.LOC
					s.VIA = content.headers.VIA

					call_rets, flag, msg = pcall( obj.OnConnect, s, table.unpack(call_args) )
				end
			end
		else
			call_rets, flag, msg = nil, "#CoAF-RPC-ERROR#", "Invalid OnConnect function"
		end

		if not call_rets or not flag then
			s:Destroy(false) --Not send disconnect message.
			local ack = {
				SN = content.SN,
				CMD = "ACK/" .. content.CMD,
				ARGS = content.ARGS,
				headers = {
					SVC = content.headers.SVC,
					OBJ = content.headers.OBJ,
					LOC = content.headers.LOC,
					VIA = content.headers.VIA,
					RESULT = "ERROR",
					MSG = msg or flag or "Connection refused"
				}
			}
			
			connObj:Send( table.concat( protocol.FormatContent(ack) ) )
			return
		end

		--Now ACK OK.
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "OK",
				MSG = ""
			}
		}
		
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )

		local r, msg = pcall(obj.OnStart, s)
		if not r then
			_dbg.Error( "CoAF.RPC", "Call 'OnStart' fail with message:", msg )
		end

	elseif action == "disconnect" then
		s.LAST_SN = content.SN
		s.CMD = "ACK/" .. content.CMD
		s.OBJ = content.headers.OBJ
		s.SVC = content.headers.SVC
		s.LOC = content.headers.LOC
		s.VIA = content.headers.VIA
		local r, flag = pcall(obj.OnDisconnect, s)
		s:Destroy() --Destroy connection.


		local rslt = "OK"
		local msg = ""

		if not r then
			_dbg.Error( "CoAF.RPC", "Call 'OnDisconnect' with error message:", flag )
		--	rslt = "ERROR"
		--	msg = flag or "ERROR"
		elseif flag then
			return
		end

		--If OnDisconnect returns false, means no disconnect message sent, so I'll ack it.
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = rslt,
				MSG = msg 
			}
		}
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )

	elseif action == "content" then
		if content.data and #content.data > 0 then
			s.LAST_SN = content.SN
			s.CMD = "ACK/" .. content.CMD
			s.OBJ = content.headers.OBJ
			s.SVC = content.headers.SVC
			s.LOC = content.headers.LOC
			s.VIA = content.headers.VIA
			s:__PutData( content.data )
			local r,msg = pcall(obj.OnData, s)
			if not r then
				_dbg.Error( "CoAF.RPC", "Call 'OnData' fail with message:", msg )
			end
		end
	else --keep-alive
		s.LAST_SN = content.SN
		s.CMD = "ACK/" .. content.CMD
		s.OBJ = content.headers.OBJ
		s.SVC = content.headers.SVC
		s.LOC = content.headers.LOC
		s.VIA = content.headers.VIA
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "OK",
				MSG = ""
			}
		}
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )
	end
end

local function processThreadTask(thrObj, daemon, connObj, content )
	local svc = daemon.services[content.headers.SVC or "#NIL#"]
	local rslt = "OK", msg

	if content.headers.SVC and not svc then
		--New feature: Check alias
		svc = daemon.services[ daemon.aliases[content.headers.SVC] or "#NIL#" ]
		if not svc then
			rslt = "ERROR"
			msg = "Invalid service"
		end
	end

	--Store last invoking connection.
	svc.__last_invoke_connection = connObj
	_M.__last_invoke_connection = connObj

	if content.CMD == "CONNECT" or content.CMD == "PING" or rslt ~= "OK" then 
		if svc and content.headers.OBJ then
			local obj = svc:GetObject(content.headers.OBJ)
			if not obj then
				rslt = "ERROR"
				msg = "Invalid object"
			end
		end

		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				LOC = content.headers.LOC,
				OBJ = content.headers.OBJ,
				VIA = content.headers.VIA,
				RESULT = rslt,
				MSG = msg
			}
		}
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )

	elseif not svc then
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "ERROR",
				MSG = "Invalid service"
			}
		}
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )

	elseif content.CMD == "SUBS" or content.CMD == "UNSUBS" then -- Subscribe/Unsubscribe
		if not content.data then
			rslt = "ERROR"
			msg = "Invalid topic arguments"
		elseif not content.headers.FROM then
			rslt = "ERROR"
			msg = "Invalid SUBS/UNSUBS 'FROM' field"
		elseif not content.headers.OBJ then
			rslt = "ERROR"
			msg = "Invalid subscribe/unsubscribe object"
		else
			local trunk
			if type(content.data) == "table" then
				trunk = "return " .. table.concat( content.data )
			elseif type(content.data) == "string" then
				trunk = "return " .. content.data
			else
				rslt = "ERROR"
				msg = "Invalid topic arguments"
			end
			if trunk then
				local func, err = load(trunk)
				if not func then
					rslt = "ERROR"
					msg = "Bad argument: " .. msg
				else
					local stat, topics = pcall( func )
					if not stat or type(topics) ~= "table" then
						rslt = "ERROR"
						msg = "Bad/null topics"
					else
						local done
						if content.CMD == "SUBS" then
							done, err = svc:Subscribe( connObj, content.headers.OBJ, content.headers.FROM, table.unpack(topics) )
						else
							done, err = svc:Unsubscribe( connObj, content.headers.OBJ, content.headers.FROM, table.unpack(topics) )
						end
						if not done then
							rslt = "ERROR"
							msg = err or "Subscribe/Unsubribe fail"
						end
					end -- Got valid args.
				end -- Load args done.
			end -- Got args.
		end

		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = rslt,
				MSG = msg
			}
		}
		
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )

	elseif content.CMD == "STREAM" then --For streamer using
		processSTREAM( svc, connObj, content )

	elseif content.CMD == "NOTIFY" or content.CMD == "CALL" or content.CMD == "STATIC_CALL" or content.CMD == "GET" or content.CMD == "SET" then
		local ready = true
		local genCodes

		local callTarget
		if not content.headers.OBJ then
			rslt = "ERROR"
			msg = "Invalid object"
			ready = false
		else
			if content.CMD == "NOTIFY" then
				callTarget = content.ARGS
				if not callTarget or callTarget == "" then
					callTarget = "OnNotify"
				end
			elseif content.CMD == "CALL" or content.CMD == "STATIC_CALL" then
				if not content.ARGS then
					rslt = "ERROR"
					msg = "Invalid call method"
					ready = false
				else
					callTarget = content.ARGS
				end
			end

			genCodes = { "return " }

			if type(content.data) == "table" then
				table.insert( genCodes, table.concat( content.data ) )
			elseif type(content.data) == "string" then
				table.insert( genCodes, content.data )
			else
				ready = false
				rslt = "ERROR"
				msg = "Invalid call arguments"
			end
		end

		local rets
		if ready then
			local obj = svc:GetObject( content.headers.OBJ )
			if not obj then
				rslt = "ERROR"
				msg = "Invalid object"
			else
				local func = invokers[content.CMD]
				if content.CMD == "NOTIFY" then
					rets = { pcall(func, obj, svc, table.concat( genCodes ), callTarget) }
				elseif content.CMD == "CALL" then
					rets = { pcall(func, obj, svc, table.concat( genCodes ), false, callTarget) }
				elseif content.CMD == "STATIC_CALL" then
					rets = { pcall(func, obj, svc, table.concat( genCodes ), true, callTarget) }
				else
					rets = { pcall(func, obj, svc, table.concat( genCodes ) ) }
				end

				if not rets[1] then
					_dbg.Error( "CoAF.RPC", content.CMD .. ": " .. content.headers.OBJ, callTarget, "With ERROR:", rets[2] )
					rslt = "ERROR"
					msg = "Execute fail: " .. (rets[2] or "unknown")
					rets = nil
				elseif not rets[2] and rets[3] == "#CoAF-RPC-ERROR#" then
					rslt = "ERROR"
					msg = "Invoke error: " .. (rets[4] or "unknown")
					rets = nil
				elseif content.CMD == "NOTIFY" then
					return --Don't send ACK for NOTIFY.
				else
					table.remove(rets, 1)
				end

			end
		end

		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = rslt,
				MSG = msg
			},
			parameters = rets
		}
		--NOTE: For good performance, maybe I need create a new co-thread.
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )
	elseif content.CMD:match("^ACK/") then
		return -- ACK message, ignore it.
	else -- Unsupported
		local ack = {
			SN = content.SN,
			CMD = "ACK/" .. content.CMD,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				OBJ = content.headers.OBJ,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "ERROR",
				MSG = "Unsupport command"
			}
		}
		--NOTE: For good performance, maybe I need create a new co-thread.
		connObj:Send( table.concat( protocol.FormatContent(ack) ) )
	end 
end

function daemon_meta:Process(connObj, content)
	local task = thread.New(processThreadTask)
	task:StartNow(self, connObj, content)
end

function daemon_meta:KeepAliveRequest(connObj)
	local task = thread.New(function()
		connObj:Send("--#KEEP-ALIVE#--\r\n") 
	end)
	task:StartNow()
end

function daemon_meta:DestroyConnection(connObj)
	self.peers[connObj.name] = nil
	if self.deadCallbacks[connObj.name] then
		local func = self.deadCallbacks[connObj.name]
		self.deadCallbacks[connObj.name] = nil
		if type(func) == "function" then
			func(connObj.name, connObj)
		end
	end
end

daemon_meta.__index = daemon_meta

-------------

local function DaemonThreadTask(thrObj, daemonObj)
	_dbg.Info("CoAF.RPC", "Service daemon of '" .. daemonObj.name .. "' is running...")
	while not thrObj.terminated do
		local conn, err = daemonObj.sock:accept()
		if not conn then
			if err ~= "timeout" and err ~= "busy" then
				_dbg.Info("CoAF.RPC", "Service daemon of'" .. daemonObj.name .. "' accept error:", err)
				break
			end
			if err == "busy" then
			--NOTE: here! Very important: (BUG found)
			-- In coaf environment, 'busy' means the socket is using by others.
			-- In this case, if this task (coroutine, or main task) does not yield
			-- out, it will fall in a dead loop: Other task will never be waked up
			-- and this loop will always continue ('busy' is always get).
				coaf.Sleep(0.01)
				if coaf then
					coaf.Sleep(0.01)
				end
			end
		else
			local ip, port = conn:getpeername()
			local name = ip or "*"
			name = name .. (port and (":" .. port) or "")
			_dbg.Info("CoAF.RPC", "Got a new peer connection:", name)
			local peerObj = connection.New(conn, daemonObj, name)
			if not peerObj then
				_dbg.Error("CoAF.RPC", "Fail to create peer task for:", name)
				conn:Close()
			else
				daemonObj.peers[name] = peerObj
				peerObj:Start()
			end
		end
	end

	-- Kill all peers.
	local k,v 
	for k,v in pairs(daemonObj.peers) do
		v:Destroy()	
	end
	daemonObj.peers = {}
	daemonObj.sock:close()
	daemonObj.sock = nil
	_dbg.Info("CoAF.RPC", "Service daemon of '" .. daemonObj.name .. "' is terminated.")
end

---------------------------------

function _M.New(sock, name)
	if not name then
		name = "<none>"
	end

	local daemonThread

	if sock then 
		daemonThread = thread.New(DaemonThreadTask)
		if not daemonThread then
			_dbg.Error("CoAF.RPC", "CreateServiceDaemon: Fail to create daemon thread for:", name)
			return nil
		end
	end

	local obj = {
		name = name,
		sock = sock,
		services = {},
		aliases = {},
		peers = {},
		deadCallbacks = {},
		thread = daemonThread
	}
	setmetatable(obj, daemon_meta)
	return obj
end

return _M
