local coaf = require "coaf.core"
local thread = require "coaf.thread"
local _dbg = require "coaf.debug"
local connection = require "coaf.rpc.connection"
local protocol = require "coaf.rpc.protocol"
local socket = require "cosocket.core"
local debug_filt = require "coaf.rpc.debug_filter"

require "cocrypto"

--New feature: route to coserver itself.
--local/_CO_DAEMON.coserver
--local/_CO_DAEMON.discovery

local coaf_service = require "coaf.service"

local _M = {
	daemonTasks = {},
	allowAccepts = {},
	connections = {},
	debug_TX = false,
	debug_RX = false
}

-------------------

function _M.EnableDebugTx(en)
	_M.debug_TX = en
	local i,c
	for i,c in ipairs(_M.connections) do
		c.debug_TX = en
	end
end

function _M.EnableDebugRx(en)
	_M.debug_RX = en
	local i,c
	for i,c in ipairs(_M.connections) do
		c.debug_RX = en
	end
end

-------------------
local function conn_PROXY_MatchNotify(_self, loc, svc, objPath, topic, notifyIdr)
	return false
end

local function conn_SERVICE_MatchNotify(_self, loc, svc, objPath, topic, notifyIdr)
	if not notifyIdr then --No IDR for duplication check
		return true
	else
		if not _self.lastNotified then
			_self.lastNotified = { notifyIdr }
			return true
		else
			local lastNoti = _self.lastNotified
			local i
			for i = #lastNoti, 1, -1 do
				if lastNoti[i] == notifyIdr then
					return false --Already notified (duplicated)
				end
			end
			if #lastNoti >= 16 then
				table.remove(lastNoti, 1)
			end
			table.insert( lastNoti, notifyIdr )
			return true
		end
	end
end

local function conn_DEFAULT_MatchNotify(_self, loc, svc, objPath, topic, notifyIdr)
	if notifyIdr then
		if not _self.lastNotified then
			_self.lastNotified = { notifyIdr }
		else
			local lastNoti = _self.lastNotified
			local i
			for i = #lastNoti, 1, -1 do
				if lastNoti[i] == notifyIdr then
					return false --Already notified (duplicated)
				end
			end
			if #lastNoti >= 16 then
				table.remove(lastNoti, 1)
			end
			table.insert( lastNoti, notifyIdr )
		end
	end

	if not _self.subscribes then
		return false
	end

	local subs = _self.subscribes
	local loc_subs = subs[loc]
	if not loc_subs then
		return false
	end

	local svc_subs = loc_subs[svc]
	if not svc_subs then
		return false
	end

	local obj_subs = svc_subs[objPath]
	if not obj_subs then
		return false
	end

	if obj_subs[topic] then
		return true
	else
		return false
	end
end

-------------------

local function serviceSocketTask(thrObj, daemonObj, sock_name)
	_dbg.Info("CoServer", "Connection daemon of '" .. sock_name .. "' is running...")
	while not thrObj.terminated do
		local conn, err = daemonObj.sock:accept()
		if not conn then
			if err ~= "timeout" and err ~= "busy" then
				_dbg.Info("CoServer", "Connection daemon of '" .. sock_name .. "' accept error:", err)
				break
			end
			if err == "busy" then	--NOTE: here! Very important: (BUG found)
			-- In coaf environment, 'busy' means the socket is using by others.
			-- In this case, if this task (coroutine, or main task) does not yield
			-- out, it will fall in a dead loop: Other task will never be waked up
			-- and this loop will always continue ('busy' is always get).

				coaf.Sleep(0.01)
			end
		else
			local ip, port, socket_type = conn:getpeername()
			if ip and socket_type == "unix" then
				ip = socket_type .. ":" .. ip
			end

			local name = ip or "*"
			name = name .. (port and (":" .. port) or "")
			_dbg.Info("CoServer", "Got a new coserver connection:", name)
			local peerObj = connection.New(conn, _M, name)
			if not peerObj then
				_dbg.Error("CoServer", "Fail to create peer task for:", name)
				conn:Close()
			else
				peerObj.debug_TAG = "CoServer"
				peerObj.debug_NAME = name
				peerObj.debug_TX = _M.debug_TX
				peerObj.debug_RX = _M.debug_RX

				peerObj.connected_addr = ip or "unknown"
				peerObj.connected_port = port

				peerObj.daemon = daemonObj
				peerObj.bind_from = sock_name
				peerObj.MatchNotify = conn_DEFAULT_MatchNotify
				peerObj.via_name = tostring(peerObj):match(".*:%s*(.*)$")
				daemonObj.peers[name] = peerObj
				table.insert( _M.connections, peerObj )
				peerObj:Start()
			end
		end
	end

	-- Kill all peers.
	local k,v 
	for k,v in pairs(daemonObj.peers) do
		v:Destroy()	
	end
	daemonObj.peers = {}
	daemonObj.sock:close()
	daemonObj.sock = nil
	_dbg.Info("CoServer", "Connection daemon of '" .. sock_name .. "' is terminated.")
end

------------------------

function _M.Bind(sock_addr)
	local addr, port
	if not sock_addr:match("^unix:") then
		addr, port = sock_addr:match("(.*):([0-9]+)")
		if not addr then
			_dbg.Error("CoServer", "Invalid binding socket address of:", sock_addr)
			return false
		end
		if addr == '' then
			addr = '*'
		end
		port = tonumber(port)
		sock_addr = addr .. ":" .. port  -- Formal format.
	else
		addr = sock_addr
	end

	if _M.daemonTasks[sock_addr] then
		_dbg.Warn("CoServer", "Address is already bound:", sock_addr )
		return true
	end

	local unix, path = sock_addr:match("(unix:)(.*)")
	if unix then
		os.remove(path)	-- Try remove exist unix domain file.
	end
	local sock = socket.tcp()
	if not sock then
		_dbg.Error("CoServer", "Fail to create service socket for: ", sock_addr )
		return false
	end

	local x, err = sock:bind(addr, port)
	if not x then
		_dbg.Error("CoServer", "Fail to bind address at ", sock_addr)
		sock:close()
		return false
	end
	sock:listen()

	local thr = thread.New(serviceSocketTask)
	local daemonObj = {
		thread = thr,
		sock = sock,
		peers = {},
		name = sock_addr
	}
	thr:Start(daemonObj, sock_addr)
	_M.daemonTasks[sock_addr] = daemonObj
	_dbg.Info("CoServer", "Connection service bound at:", sock_addr)
	return true
end

function _M.Unbind(address)
	if not sock_addr:match("^unix:") then
		addr, port = sock_addr:match("(.*):([0-9]+)")
		if not addr then
			_dbg.Error("CoServer", "Invalid socket address while Unbind():", sock_addr)
			return false
		end
		if addr == '' then
			addr = '*'
		end
		port = tonumber(port)
		sock_addr = addr .. ":" .. port  -- Formal format.
	end
	local daemonObj = _M.daemonTasks[sock_addr]
	if daemonObj then
		_dbg.Info("CoServer", "Connection service un-bound:", sock_addr)
		daemonObj.thread:Terminate()
		_M.daemonTasks[sock_addr] = nil
	end
	return true
end

----------------------

function _M.AllowAcceptProxy(loc, serviceId, addressPattern)
	local idr = (loc or "ANY") .. "/" .. (serviceId or "ANY") .. ":" .. (addressPattern or "*")
	_M.allowAccepts[idr] = {loc, serviceId, addressPattern}
	return idr
end

function _M.DisallowAcceptProxy(acceptId, optArg)
	if _M.allowAccepts[acceptId] then
		_M.allowAccepts[acceptId] = nil
	else
		optArg = optArg or "ANY"
		if (acceptId == "*" or acceptId == "ANY") and optArg == "ANY" then
			_M.allowAccepts = {}
		else
			local k,v
			for k,v in pairs(_M.allowAccepts) do
				if (acceptId == v[1] or acceptId == "*" or acceptId == "ANY" or v[1] == "ANY") and 
					(optArg == v[2] or optArg == "ANY" or v[2] == "ANY") then
					_M.allowAccepts[k] = nil
				end
			end
		end
	end
end

function _M.IsProxyAllowed(loc, serviceId, bindFrom)
	if not loc or loc == "local" then
		if bindFrom:match("^unix") then --Always allow local service proxy from unix socket.
			return true
		end
	end

	local k,v
	for k,v in pairs(_M.allowAccepts) do
		if v and 
		   (loc == v[1] or loc == "ANY" or v[1] == "ANY") and
		   (serviceId == v[2] or serviceId == "ANY" or v[2] == "ANY")
		then
			if not v[3] or v[3] == "*" or not bindFrom or ((v[3] == "unix" or v[3] == "local") and bindFrom:match("^unix")) or
				((v[3] == "ip" or v[3] == "remote") and not bindForm:match("^unix")) or
				bindForm:match(v[3]) then
				return true
			end
		end
	end
	return false
end

-----------------------------------

local function NewConnection(addr)
	local port
	if not addr:match("^unix:") then
		local ip
		ip, port = addr:match("(.*):([0-9]+)")
		if ip then
			if ip == '' or ip == '*' then
				addr = "127.0.0.1"
			else
				addr = ip
			end
			port = tonumber(port) or 14000
		else
			port = 14000
		end
		port = tonumber(port) or 14000
	end 

	local sock = socket.tcp()
	if not sock then
		_dbg.Error("CoServer", "Fail to create proxy/request-proxy connection! To:", addr, port)
		return nil
	end

	local status, err = sock:connect(addr, port)
	if not status then
		_dbg.Error("CoServer", "Fail to connect proxy/request-proxy to:", addr, port)
		sock:close()
		return nil
	end
	return sock, addr, port
end

-----------------------------------

local function connection_WaitDie(conn)
	if not conn or not conn.thread then
		return "dead"
	end
	local coro, isMain = coroutine.running()
	if coro and not isMain then
		if not conn.DieWaitters then
			conn.DieWaitters = {coro}
		else
			table.insert(conn.DieWaitters, coro)
		end
		local r,err = coroutine.yield( coro )
		if not r then
			return err
		else
			return "dead"
		end
	else
		return "#fail#"
	end
end

-----------------------------------

function _M.RequestProxy(address, loc, serviceId, timeout, authKey)
	local sock, target_addr, target_port = NewConnection(address)
	if not sock then
		return nil
	end

	if timeout and timeout <= 0 then
		timeout = 60
	end

	local req = {
		SN = 0,
		CMD = "REQUEST_PROXY",
		headers = {
			SVC = serviceId,
			LOC = loc,
			KA = timeout
		}
	}

	local ack
	local try_again

	repeat
	--==============================================================
	--{ repeat until (to ... 'until here' below)
	--==============================================================
	try_again = false

	local protocol_data = table.concat( protocol.FormatContent(req) )
	
	if _M.debug_TX and debug_filt.MatchContent(req) then
		_dbg.Info1("CoServer", "(TX) >>>>>> RequestProxy - to:", address, loc, serviceId)
		_dbg.Print("CoServer", protocol_data)
	end

	local status, err = sock:send( protocol_data )
	if not status then
		sock:close()
		_dbg.Error("CoServer", "Fail to request proxy to CoServer (Send error). For proxy service:", loc, serviceId )
		return nil
	end

	local parser = protocol.NewParser()
	parser:EnableOriginal(_M.debug_RX)
	local data, partical
	--local ack
	while true do
		data, err, partical = sock:receive( parser.next_fetch )
		if not data then
			if err ~= "timeout" and err ~= "busy" then
				sock:close()
				_dbg.Error("CoServer", "Fail to request proxy to CoServer (I/O error). For proxy service:", loc, serviceId)
				return nil
			end
			if partial and #partial > 0 then
				data = partial
			elseif err == "timeout" then
				sock:close()
				_dbg.Error("CoServer", "Fail to request proxy to CoServer (timeout). For proxy service:", loc, serviceId)
				return nil
			else
			--NOTE: here! Very important: (BUG found)
			-- In coaf environment, 'busy' means the socket is using by others.
			-- In this case, if this task (coroutine, or main task) does not yield
			-- out, it will fall in a dead loop: Other task will never be waked up
			-- and this loop will always continue ('busy' is always get).
				coaf.Sleep(0.01)
			end
		end

		if data then
			status, err = parser:Push(data)
			if status and err == "complete" then
				local content = parser.content
				local orig_content = parser.orig_content
				parser:Clear()
				parser:EnableOriginal(_M.debug_RX)

				if _M.debug_RX and debug_filt.MatchContent(content) then
					protocol_data = table.concat( orig_content )
					_dbg.Info1("CoServer", "(RX) <<<<<< RequestProxy - from:", address, loc, serviceId)
					_dbg.Print("CoServer", protocol_data)
				end

				if content.CMD == "ACK/REQUEST_PROXY" and tonumber(content.SN) == req.SN then
					ack = content
					break
				end
			end
		end
	end

	if ack.headers.RESULT == "AUTH_REQ" and authKey and ack.headers.CHA then
		--_dbg.Info("CoServer", "[AUTH] Remote require authorization, challenge is:", ack.headers.CHA, ". For proxy service:", loc, serviceId)
		try_again = true
		local cocore = require "coserver.core"
		local unique_id = cocore.configs.UNIQUE_ID
		if type(unique_id) == "function" then
			unique_id = unique_id()
		end

		local tip = CoAF_CRYPTO.get_challenge()
		local dig = CoAF_CRYPTO.finger_print( ack.headers.CHA .. tip .. authKey )

		req = {
			SN = 0,
			CMD = "REQUEST_PROXY",
			headers = {
				SVC = serviceId,
				LOC = loc,
				KA = timeout,
				UNIQ = unique_id,
				CHA = ack.headers.CHA,
				TIP = tip,
				DIG = dig
			}
		}
	end
	--==============================================================
	-- } until here
	--==============================================================
	until not try_again;

	if ack.headers.RESULT == "OK" then
		local trunkCode
		if type(ack.data) == "table" then
			trunkCode = "return " .. table.concat( ack.data )
		elseif type(ack.data) == "string" then
			trunkCode = "return " .. ack.data
		end

		local validServices

		if trunkCode then
			local func, err = load(trunkCode)
			if not func then
				_dbg.Warn("CoServer", "Bad REQUEST_PROXY response, so only assume this service valid:", loc, serviceId)
			else
				local status, ret = pcall(func)
				if not status or type(ret) ~= "table" then
					_dbg.Warn("CoServer", "Bad REQUEST_PROXY response, so only assume this service valid:", loc, serviceId)
				else
					validServices = ret
				end
			end
		else
			_dbg.Warn("CoServer", "No allowed proxy services returned by REQUEST_PROXY, so only assume this service valid:", loc, serviceId)
		end

		local conn = connection.New(sock, _M, address)
		if not conn then
			sock:close()
			return nil
		end

		--Match notify.
		conn.MatchNotify = conn_PROXY_MatchNotify

		conn.debug_TAG = "CoServer"
		conn.debug_NAME = address
		conn.debug_TX = _M.debug_TX
		conn.debug_RX = _M.debug_RX

		conn.connected_addr = target_addr
		conn.connected_port = target_port

		conn.Type = "PROXY"

		if timeout then
			conn:SetKeepAlive( { keepAlive = true, keepAliveTimeout = timeout } )
		end

		if not validServices then
			conn.ValidServices = { {loc, serviceId} }
		else
			conn.ValidServices = {}
			local i,v
			for i,v in pairs(validServices) do
				if type(v) == "string" then
					loc, serviceId = v:match("(.*)/(.*)$")
					if not loc then
						loc = v
						serviceId = "ANY"
					end
					table.insert(conn.ValidServices, {loc, serviceId})
				elseif type(v) == "table" then
					table.insert( conn.ValidServices, v )
				end
			end
		end

		table.insert( _M.connections, conn )
		conn.via_name = tostring(conn):match(".*:%s*(.*)$")
		conn.WaitDie = connection_WaitDie	--WaitDie function.
		conn:Start()
		return conn
	else
		_dbg.Error("CoServer", "Fail to request proxy to CoServer, reason:", ack.headers.MSG, "For proxy service:", loc, serviceId)
		sock:close()
		return nil
	end
end

----------------------------------

function _M.NewProxyLink(address, kaTimeout, exposeTable, unique_id, authKey)
	if not exposeTable then
		local services = require "coserver.services"
		exposeTable = services.remote_expose
	end

--	if not unique_id then
--		local core = require "coserver.core"
--		unique_id = core.configs.UNIQUE_ID
--		if type(unique_id) == "function" then
--			unique_id = unique_id()
--		end
--	end

	local params = {}
	local k,v
	local loc, svc
	for k,v in pairs(exposeTable) do
		loc,svc = k:match("(.*)/(.*)$")
		if loc then
			if loc == "ANY" then
				loc = unique_id
			end
		end
		
		if loc then
			table.insert(params, loc .. "/" .. svc) -- Exposed remote services.
		end
	end

	if #params == 0 then
		_dbg.Error("CoServer", "No expose services configured or specified while new proxy link. Please check 'expose' configuration")
		return nil
	end

	local sock = NewConnection(address)
	if not sock then
		return nil
	end

	if type(kaTimeout) ~= "number" or kaTimeout <= 0 then
		kaTimeout = 60
	end

	local req = {
		SN = 0,
		CMD = "PROXY",
		ARGS = "multiple",
		headers = {
			SVC = "ANY",
			LOC = unique_id or "ANY",
			KA = kaTimeout
		},
		parameters = params
	}

	local ack
	local try_again

	repeat
	--=================================================
	--{ repeat until (to ... 'util here' below)
	--=================================================

	try_again = false

	local protocol_data = table.concat( protocol.FormatContent(req) )
	if _M.debug_TX and debug_filt.MatchContent(req) then
		_dbg.Info1("CoServer", "(TX) >>>>>> NewProxyLink - to:", address )
		_dbg.Print("CoServer", protocol_data)
	end

	local status, err = sock:send( protocol_data )
	if not status then
		sock:close()
		_dbg.Error("CoServer", "Fail to create proxy link to CoServer (Send error):", address )
		return nil
	end

	local parser = protocol.NewParser()
	parser:EnableOriginal(_M.debug_RX)
	local data, partical
	while true do
		data, err, partical = sock:receive( parser.next_fetch )
		if not data then
			if err ~= "timeout" and err ~= "busy" then
				sock:close()
				_dbg.Error("CoServer", "Fail to create proxy link to CoServer (I/O error):", address)
				return nil
			end
			if partial and #partial > 0 then
				data = partial
			elseif err == "timeout" then
				sock:close()
				_dbg.Error("CoServer", "Fail to create proxy link to CoServer (timeout):", address)
				return nil
			else
			--NOTE: here! Very important: (BUG found)
			-- In coaf environment, 'busy' means the socket is using by others.
			-- In this case, if this task (coroutine, or main task) does not yield
			-- out, it will fall in a dead loop: Other task will never be waked up
			-- and this loop will always continue ('busy' is always get).
				coaf.Sleep(0.01)
			end
		end

		if data then
			status, err = parser:Push(data)
			if status and err == "complete" then
				local content = parser.content
				local orig_content = parser.orig_content
				parser:Clear()
				parser:EnableOriginal(_M.debug_RX)

				if _M.debug_RX and debug_filt.MatchContent(content) then
					protocol_data = table.concat( orig_content )
					_dbg.Info1("CoServer", "(RX) <<<<<< NewProxyLink - from:", address )
					_dbg.Print("CoServer", protocol_data)
				end

				if content.CMD == "ACK/PROXY" and tonumber(content.SN) == req.SN then
					ack = content
					break
				end
			end
		end
	end

	if ack.headers.RESULT == "AUTH_REQ" and authKey and ack.headers.CHA then
		--_dbg.Info("CoServer", "[AUTH] Remote require authorization, challenge is:", ack.headers.CHA, ".For proxy link to CoServer:", address)
		try_again = true

		local tip = CoAF_CRYPTO.get_challenge()
		local dig = CoAF_CRYPTO.finger_print( ack.headers.CHA .. tip .. authKey )

		req = {
			SN = 0,
			CMD = "PROXY",
			ARGS = "multiple",
			headers = {
				SVC = "ANY",
				LOC = unique_id or "ANY",
				KA = kaTimeout,
				UNIQ = unique_id, --Add this field, for device ID check in remote.
				CHA = ack.headers.CHA,
				TIP = tip,
				DIG = dig
			},
			parameters = params
		}
	end

	--=================================================
	--} until here.
	--=================================================
	until not try_again;

	if ack.headers.RESULT == "OK" then
		local conn = connection.New(sock, _M, address)
		if not conn then
			sock:close()
			return nil
		end

		conn.debug_TAG = "CoServer"
		conn.debug_NAME = address
		conn.debug_TX = _M.debug_TX
		conn.debug_RX = _M.debug_RX

		conn:SetKeepAlive( { keepAlive = true, keepAliveTimeout = kaTimeout } )

		conn.Type = "SERVICE"
		conn.MatchNotify = conn_SERVICE_MatchNotify
		conn.ServiceTranslateTable = exposeTable
		table.insert( _M.connections, conn )
		conn.via_name = tostring(conn):match(".*:%s*(.*)$")
		conn.WaitDie = connection_WaitDie	--WaitDie function.
		conn:Start()
		return conn
	else
		_dbg.Error("CoServer", "Fail to create proxy link to CoServer, reason:", ack.headers.MSG, " to coserver:", address)
		sock:close()
		return nil
	end
end

-----------------------------------

local function process_PROXY(conn, content)
	if content.ARGS and content.ARGS == "multiple" then
		local trunkCode
		if type(content.data) == "table" then
			trunkCode = "return " .. table.concat( content.data )
		elseif type(content.data) == "string" then
			trunkCode = "return " .. content.data
		end

		local validServices

		local ack = {
			CMD = "ACK/PROXY",
			SN = content.SN,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "OK"
			}
		}
		local succed
		if trunkCode then
			local func, err = load(trunkCode)
			if not func then
				_dbg.Error("CoServer", "Bad multiple PROXY request from connection:", conn.name)
				ack.headers.RESULT = "ERROR"
				ack.headers.MSG = "Bad request content"
			else
				local status, ret = pcall(func)
				if not status or type(ret) ~= "table" then
					_dbg.Error("CoServer", "Bad multiple PROXY request from connection:", conn.name)
					ack.headers.RESULT = "ERROR"
					ack.headers.MSG = "Bad request content"
				else
					validServices = ret
					succed = true
				end
			end
		else
			_dbg.Error("CoServer", "No valid proxy services described by multiple PROXY from connection:", conn.name)
			ack.headers.RESULT = "ERROR"
			ack.headers.MSG = "No proxied service(s)"
		end

		if succed then
			conn.ValidServices = {}
			local i,v
			for i,v in pairs(validServices) do
				if type(v) == "string" then
					loc, serviceId = v:match("(.*)/(.*)$")
					if not loc then
						loc = v
						serviceId = "ANY"
					end
					if _M.IsProxyAllowed(loc, serviceId, conn.bind_from) then
						table.insert(conn.ValidServices, {loc, serviceId})
					end
				elseif type(v) == "table" then
					if _M.IsProxyAllowed(v[1], v[2], conn.bind_from) then
						table.insert(conn.ValidServices, v)
					end
				end
			end
			if #conn.ValidServices == 0 then
				ack.headers.RESULT = "ERROR"
				ack.headers.MSG = "Proxy disallowed"
			else
				conn.Type = "PROXY"
				local ka = tonumber(content.headers.KA)
				if ka then
					if ka <= 0 then
						ka = 60
					end
					conn:SetKeepAlive( {keepAlive=true, keepAliveTimeout = ka} )
				end
			end
		end

		conn:Send( table.concat( protocol.FormatContent(ack) ) )
	else
		local ack = {
			CMD = "ACK/PROXY",
			SN = content.SN,
			ARGS = content.ARGS,
			headers = {
				SVC = content.headers.SVC,
				LOC = content.headers.LOC,
				VIA = content.headers.VIA,
				RESULT = "OK"
			}
		}

		if not content.headers.LOC and not content.headers.SVC then
			_dbg.Error("CoServer", "No valid proxy service location/id provided by PROXY request from connection:", conn.name)
			ack.headers.RESULT = "ERROR"
			ack.headers.MSG = "Invalid location/service-id provided"
		else
			if _M.IsProxyAllowed(content.headers.LOC or "local", content.headers.SVC or "ANY", conn.bind_from) then
				conn.Type = "PROXY"
				conn.MatchNotify = conn_PROXY_MatchNotify
				conn.ValidServices = { {content.headers.LOC or "local", content.headers.SVC or "ANY"} }

				local ka = tonumber(content.headers.KA)
				if ka then
					if ka <= 0 then
						ka = 60
					end
					conn:SetKeepAlive( {keepAlive=true, keepAliveTimeout = ka} )
				end
			else
				ack.headers.RESULT = "ERROR"
				ack.headers.MSG = "Proxy disallowed"
			end
		end
		conn:Send( table.concat( protocol.FormatContent(ack) ) )
	end
end

----------

local function process_REQUEST_PROXY(conn, content)
	local services = require "coserver.services"
	local exposes = services.remote_expose
	local params = {}
	if exposes then
		local matched = services.MatchTranslatedServices( content.headers.LOC, content.headers.SVC, nil, conn.connected_addr)
		if matched then
			params = matched
		end
	end
	local ack = {
		CMD = "ACK/REQUEST_PROXY",
		SN = content.SN,
		ARGS = content.ARGS,
		headers = {
			SVC = content.headers.SVC,
			LOC = content.headers.LOC,
			VIA = content.headers.VIA
		}
	}
	if #params == 0 then
		_dbg.Error("CoServer", "No or no matched exposed remote services:", content.headers.LOC, content.headers.SVC, " while REQUEST_PROXY from:", conn.name )
		ack.headers.RESULT = "ERROR"
		ack.headers.MSG = "No matched service"
	else
		ack.headers.RESULT = "OK"
		ack.parameters = params
	
		conn.Type = "SERVICE"
		conn.MatchNotify = conn_SERVICE_MatchNotify
		conn.ServiceTranslateTable = exposes

		local ka = tonumber(content.headers.KA)
		if ka then
			if ka <= 0 then
				ka = 60
			end
			conn:SetKeepAlive( {keepAlive=true, keepAliveTimeout = ka} )
		end
	end

	conn:Send( table.concat( protocol.FormatContent(ack) ) )
end

----------

local function process_ACKs(conn, content)
	if not content.headers.VIA or content.headers.VIA == "" then
		return
	end
	local services = require "coserver.services"

	local via_saved, via_cur = content.headers.VIA:match("(.+)%s+([^%s]+)")
	if not via_saved then
		via_cur = content.headers.VIA
	end

	content.headers.VIA = via_saved
	local backwdConn
	local i,c
	for i,c in ipairs(_M.connections) do
		if c and c.via_name == via_cur then
			backwdConn = c
			break
		end
	end

	if not backwdConn then
		_dbg.Warn("CoServer", "Got " .. content.CMD .. " but not found backward router, VIA:", via_cur)
		return
	end

	if backwdConn.Type == "SERVICE" and backwdConn.ServiceTranslateTable then
		local _t, t_loc, t_svc
		_t, t_loc, t_svc = services.TranslateLocalService(content.headers.LOC or "#INVALID#", content.headers.SVC or "#INVALID#", backwdConn.ServiceTranslateTable)
		if _t then
			content.headers.LOC = t_loc
			content.headers.SVC = t_svc
		end
	end

	if content.CMD == "ACK/SUBS" then
		local pending = backwdConn.pendingSubs
		if pending and pending.SN == content.SN and (pending.FROM == content.headers.FROM or 
													(pending.LOC == content.headers.LOC and pending.SVC == content.headers.SVC and pending.OBJ == content.headers.OBJ) 
		) then
			if content.headers.RESULT == "OK" then
				if not backwdConn.subscribes then
					backwdConn.subscribes = {}
				end

				local subs = backwdConn.subscribes
				local locSubs = subs[pending.LOC]
				if not locSubs then
					locSubs = {}
					subs[pending.LOC] = locSubs
				end

				local svcSubs = locSubs[pending.SVC]
				if not svcSubs then
					svcSubs = {}
					locSubs[pending.SVC] = svcSubs
				end

				local objSubs = svcSubs[pending.OBJ]
				if not objSubs then
					objSubs = {}
					svcSubs[pending.OBJ] = objSubs
				end

				if pending.TOPICS then
					local i, t
					for i,t in ipairs(pending.TOPICS) do
						if not objSubs[t] then
							objSubs[t] = {[pending.FROM or "unknown"] = true}
						else
							objSubs[t][pending.FROM or "unknown"] = true
						end
					end
				end
			end
			backwdConn.pendingSubs = nil
		end
	elseif content.CMD == "ACK/UNSUBS" then
		local pending = backwdConn.pendingUnsubs
		if pending and pending.SN == content.SN and (pending.FROM == content.headers.FROM or 
													(pending.LOC == content.headers.LOC and pending.SVC == content.headers.SVC and pending.OBJ == content.headers.OBJ) 
		) then
			if content.headers.RESULT == "OK" then
				local objSubs
				if 	backwdConn.subscribes and 
					backwdConn.subscribes[pending.LOC] and 
					backwdConn.subscribes[pending.LOC][pending.SVC] and
					backwdConn.subscribes[pending.LOC][pending.SVC][pending.OBJ] then
					objSubs = backwdConn.subscribes[pending.LOC][pending.SVC][pending.OBJ]
				end

				if objSubs and pending.TOPICS then
					local i, t
					for i,t in ipairs(pending.TOPICS) do
						if objSubs[t] then
							objSubs[t][pending.FROM or "unknown"] = nil
							local kk, vv, found
							for kk, vv in pairs(objSubs[t]) do
								found = true
								break
							end
							if not found then
								objSubs[t] = nil
							end
						end
					end
				end
			end
			backwdConn.pendingUnsubs = nil
		end
	-- Processing ALT-ADDR (CONNECT, or returned ALT_ADDR)
	elseif content.CMD == "ACK/CONNECT" or content.headers.ALT_ADDR then
		local key, svcObj = services.GetExactService(content.headers.LOC, content.headers.SVC)
		if svcObj and svcObj.ALT_SOCKETS then
			content.headers.ALT_ADDR = "yes"
			if type(svcObj.ALT_SOCKETS) == "string" then
				content.parameters = { svcObj.ALT_SOCKETS }
			else
				content.parameters = svcObj.ALT_SOCKETS
			end
		else
			content.headers.ALT_ADDR = nil	--Drop ALT_ADDRs
			content.data = nil
		end
	end

	backwdConn:Send( table.concat( protocol.FormatContent(content) ) )
end

----------

local function process_REQUESTs(threadObj, conn, content)
	local loc = content.headers.LOC or "local"
	local svc = content.headers.SVC
	local obj = content.headers.OBJ
	local services = require "coserver.services"

	if content.CMD == "NOTIFY" and conn.Type == "PROXY" then --Service notify
		local topic = content.ARGS
		if not svc or not obj or not topic then
			_dbg.Warn("CoServer", "Received an invalid NOTIFY message (with wrong SVC/OBJ or topic)")
			return
		end

		local basicNotifyIdr = loc .. "|" .. svc .. "|" .. obj .. "|" .. topic .. "|" .. content.SN
		local i, c
		local shouldFwd

		for i, c in ipairs(_M.connections) do
			if c ~= conn then -- Don't pass message to self.
				shouldFwd = false
				if c.Type == "SERVICE" and c.ServiceTranslateTable then -- So need translate service location
					local _t, t_loc, t_svc
					_t, t_loc, t_svc = services.TranslateLocalService(loc, svc, c.ServiceTranslateTable)
					if _t and t_loc and c.MatchNotify then
						shouldFwd = c:MatchNotify(t_loc, t_svc, obj, topic, t_loc .. "|" .. t_svc .. "|" .. obj .. "|" .. topic .. "|" .. content.SN)
						if shouldFwd then
							content.headers.LOC = t_loc --Change the LOC
							content.headers.SVC = t_svc
						end
					end
				else
					shouldFwd = c:MatchNotify(loc, svc, obj, topic, basicNotifyIdr)
					if shouldFwd then
						content.headers.LOC = loc --MUST assign, since it may be changed in previous.
						content.headers.SVC = svc
					end
				end

				if shouldFwd then
					c:Send(table.concat(protocol.FormatContent(content)) )
				end
			end --Not connection self.
		end -- For each connections.
		return
	elseif (content.CMD == "SUBS" or content.CMD == "UNSUBS") and conn.Type ~= "PROXY" and conn.Type ~= "SERVICE" then
		local from = content.headers.FROM or "unknown"
		if svc and obj then
			local trunkCode
			if type(content.data) == "table" then
				trunkCode = "return " .. table.concat( content.data )
			elseif type(content.data) == "string" then
				trunkCode = "return " .. content.data
			end
			if trunkCode then
				local func, err = load(trunkCode)
				if func then
					local st, topics = pcall(func)
					if st and topics then
						if content.CMD == "SUBS" then
							conn.pendingSubs = {
								SN = content.SN,
								LOC = loc,
								SVC = svc,
								OBJ = obj,
								FROM = from,
								TOPICS = topics
							}
						else
							conn.pendingUnsubs = {
								SN = content.SN,
								LOC = loc,
								SVC = svc,
								OBJ = obj,
								FROM = from,
								TOPICS = topics
							}
						end
					end -- With valid topics (Subs/Unsubs)
				end -- content valid.
			end -- With valid protocol body.
		end -- With valid protocol headers.
	end

	if conn.Type == "SERVICE" then -- Translate service location and ID
		local targetSvc
		targetSvc, loc, svc = services.TranslateRemoteService(loc, svc or "#INVALID#", conn.ServiceTranslateTable, content.headers.OBJ, conn.connected_addr )
		if not targetSvc then
			local ack = {
				CMD = "ACK/" .. content.CMD,
				SN = content.SN,
				ARGS = content.ARGS,
				headers = {
					RESULT = "ERROR",
					MSG = "Invalid location or permission denied",
					VIA = content.headers.VIA,
					LOC = content.headers.LOC,
					SVC = content.headers.SVC,
					OBJ = content.headers.OBJ,
					FROM = content.headers.FROM
				}
			}
			conn:Send(table.concat( protocol.FormatContent(ack) ))
			return
		else
			content.headers.LOC = loc
			content.headers.SVC = svc
		end
	end

	local fwdConn
	local startTime
	local svcObj

	--New feature - route to daemon services.
	local daemon_svc = svc and svc:match("^_CO_DAEMON%.(.+)")
	if loc == "local" and daemon_svc then
		fwdConn = daemon_svc
	end

	while not fwdConn do
		--Looking, and forward.
		local matchs = {}

		local i, c
		local ii, ss
		for i,c in ipairs(_M.connections) do
			if c ~= conn and c.Type == "PROXY" and c.ValidServices then
				for ii, ss in ipairs(c.ValidServices) do
					if loc == ss[1] and svc == ss[2] then
						matchs[1] = c
						break
					elseif loc == ss[1] and (not svc or svc == "ANY" or not ss[2] or ss[2] == "ANY") then
						matchs[2] = c
					elseif svc == ss[2] and (not loc or loc == "ANY" or not ss[1] or ss[1] == "ANY") then
						matchs[3] = c
					elseif (not loc or loc == "ANY" or not ss[1] or ss[1] == "ANY") and (not svc or svc == "ANY" or not ss[2] or ss[2] == "ANY") then
						matchs[4] = c
					end
				end

				if matchs[1] then break end  -- Best matched, so don't check next.
			end
		end

		fwdConn = matchs[1] or matchs[2] or matchs[3] or matchs[4]
		if fwdConn then 
			break
		end

		if not startTime then
			local key
			key, svcObj = services.GetExactService(loc, svc)
			if not svcObj then break end
			if not services.StartService(svcObj.LOC, svcObj.ID) then
				svcObj = nil 
				break
			end
			if svcObj.PROXY == "none" or svcObj.PROXY == "NONE" then
				break
			end
			startTime = os.time()
		else
			local waitTime = os.time() - startTime
			if waitTime < 0 or waitTime >= (svcObj.TIMEOUT or 30) then
				break
			else
				coaf.Sleep(0.1) --Sleep 0.1 second, wait and try check again.
			end
		end
	end

	local fwdLocalDaemon
	if type(fwdConn) == "string" then --This is the case of route to daemon services.
		local local_svc_obj = coaf_service.GetService( fwdConn )
		if local_svc_obj and local_svc_obj.daemons and local_svc_obj.daemons[1] then
			fwdLocalDaemon = local_svc_obj.daemons[1]
		else
			fwdConn = nil --Clear so returns failure.
		end
	end

	if not fwdConn then
		local ack = {
			CMD = "ACK/" .. content.CMD,
			SN = content.SN,
			ARGS = content.ARGS,
			headers = {
				RESULT = "ERROR",
				MSG = "Invalid target/service",
				LOC = loc,
				SVC = svc,
				VIA = content.headers.VIA,
				OBJ = content.headers.OBJ,
				FROM = content.headers.FROM
			}
		}

		if svcObj and svcObj.ALT_SOCKETS then
			ack.headers.RESULT = "ALT-ADDR"
			ack.headers.ALT_ADDR = "yes"
			ack.headers.MSG = "Should choose an alternative address"
			if type(svcObj.ALT_SOCKETS) == "string" then
				ack.parameters = { svcObj.ALT_SOCKETS }
			else
				ack.parameters = svcObj.ALT_SOCKETS
			end
		end

		conn:Send(table.concat( protocol.FormatContent(ack) ))
		return
	end

	--Now, forward request to connection
	if fwdLocalDaemon then
		--local daemon services. (_CO_DAEMON.???)
		fwdLocalDaemon:SetServiceAlias( "_CO_DAEMON." .. fwdConn, fwdConn )
		fwdLocalDaemon:Process(conn, content)
	else --Connection forward.
		if not content.headers.VIA then
			content.headers.VIA = conn.via_name
		else
			content.headers.VIA = content.headers.VIA .. " " .. conn.via_name
		end
	
		fwdConn:Send( table.concat( protocol.FormatContent(content) ))
	end
end

-----------------------------------
-- Connection handlers
-----------------------------------

function _M:Process(conn, content)
	if not conn.__authorized then
		local cocore = require "coserver.core"
		local auth_key = cocore.configs.AUTHOR_KEY
		
		--Ensured type, ignore authorization
		if conn.Type == "PROXY" or conn.Type == "SERVICE" then
			conn.__authorized = true
		elseif not auth_key then
		--Global not configured to require authorize.
			conn.__authorized = true
		elseif conn.connected_addr and (conn.connected_addr:match("^unix") or conn.connected_addr:match("^127%.[0-9]*%.[0-9]*%.[0-9]") or conn.connected_addr == "localhost") then
		--Skip authorization for local link
			conn.__authorized = true 
		else
			local result = "ERROR"
			local msg = "Authorize failed"

			if not conn.__auth_challenge then
				conn.__auth_challenge = CoAF_CRYPTO.get_challenge()
				result = "AUTH_REQ"
				msg = "Authorization required"
			else
				local challenge = content.headers.CHA
				local tip = content.headers.TIP
				local digest = content.headers.DIG

				if challenge == conn.__auth_challenge and tip and digest then
					if type(auth_key) == "function" then
						auth_key = auth_key(content, conn.connected_addr)
					end

					if type(auth_key) == "string" then
						local sum = CoAF_CRYPTO.finger_print( challenge .. tip .. auth_key )
						if sum == digest then
							conn.__authorized = true
						end
					elseif type(auth_key) == "table" then
						local i
						for i=1, #auth_key do
							local sum = CoAF_CRYPTO.finger_print( challenge .. tip .. auth_key[i] )
							if sum == digest then
								conn.__authorized = true
								break
							end
						end
					elseif type(auth_key) == "nil" then --No authorization required.
						conn.__authorized = true
					end

					if not conn.__authorized then
						conn.__auth_challenge = nil
					else
						--TODO: For next check.
						conn.__auth_for_location = content.headers.LOC
						conn.__auth_for_service = content.headers.SVC
					end
				else
					conn.__auth_challenge = CoAF_CRYPTO.get_challenge()
					result = "AUTH_REQ"
					msg = "Authorization required"
				end
			end

			if not conn.__authorized then
				local ack = {
					CMD = "ACK/" .. content.CMD,
					SN = content.SN,
					ARGS = content.ARGS,
					headers = {
						RESULT = result,
						MSG = msg,
						LOC = content.headers.LOC,
						SVC = content.headers.SVC,
						VIA = content.headers.VIA,
						OBJ = content.headers.OBJ,
						FROM = content.headers.FROM,
						CHA = conn.__auth_challenge
					}
				}
				conn:Send(table.concat( protocol.FormatContent(ack) ))
				return
			end
		end
	end

	if content.CMD == "PROXY" then
		process_PROXY(conn, content)
	elseif content.CMD == "REQUEST_PROXY" then
		process_REQUEST_PROXY(conn, content)
	elseif content.CMD:match("^ACK/") then
		--[[
		local thr = thread.New(process_ACKs)
		thr:StartNow(conn, content)
		]]
		process_ACKs(conn, content)
	else
		local thr = thread.New(process_REQUESTs)
		thr:StartNow(conn, content)
	end
end

function _M:KeepAliveRequest(conn)
	local task = thread.New(function()
		conn:Send("--#KEEP-ALIVE#--\r\n") 
	end)
	task:StartNow()
end

function _M:DestroyConnection(conn)
	if conn.daemon and conn.daemon.peers then
		conn.daemon.peers[conn.name] = nil
	end
	local i, v
	for i,v in ipairs(_M.connections) do
		if v == conn then
			table.remove( _M.connections, i )
			break
		end
	end

	if conn.DieWaitters then
		for i,v in ipairs(conn.DieWaitters) do
			local status = coroutine.status( v )
			if status == "suspended" then
				coroutine.resume( v, true, "dead" )
			end
		end
	end

	conn.DieWaitters = nil
end

------------------------------------
return _M
