-- This is the coserver core functions.
-- It's a stand-alone module, you can or not use it, expose it to CoService, ...

local co_daemon = require "coserver.conn_daemon"
local co_services = require "coserver.services"
local coaf = require "coaf.core"
local thread = require "coaf.thread"
local _dbg = require "coaf.debug"
local debug_filt = require "coaf.rpc.debug_filter"

local update_host_firstrun = true

local _M = {
	config_location = "/etc/coserver",
	config_file = "coserver.conf",
	configs = {
		UNIQUE_ID = coaf.GetHostname,
		HOSTNAME = nil,
		BIND = { "unix:/var/run/coaf.coserver" },
		CONTROL_BIND = { "unix:/var/run/coaf.coserver.ctrl" },
		DEBUG_TX = false,
		DEBUG_RX = false,
		DEBUG_LEVEL = 3,
		LOGFILE = nil,
		DISABLE_DEBUG_CONSOLE = false,
		DISCOVERY_HOST_DOMAIN = ".coaf",
		RUN = nil
	},
	proxy_links = {}
}

local PATH_DELIM = "/"

if coaf.OSType() == "Windows" then
	_M.config_location = ".\\config"
	PATH_DELIM = "\\"
end

------------------------------------------------------------
-- Internal use: parse and load configure (string).
-- The configure should like:
-- UNIQUE_ID = "unique-location-id",
-- or
-- UNIQUE_ID = function_to_get_location_id,
--
-- HOSTNAME = "hostname",
-- or
-- HOSTNAME = function() --To set hostname manually.
-- or (default)
-- HOSTNAME = nil --Don't change the hostname.
--
-- BIND = { "server-bind-address1", "server-bind-address2", ... }
-- CONTROL_BIND = { "server-coService-bind-address", "server-coService-bind-address2", ... }
-- DEBUG_TX = true|false  --Debug protocol TX details.
-- DEBUG_RX = true|false  --Debug protocol RX details.
-- DEBUG_LEVEL = <0..3>   --Debug message output level
-- LOGFILE = "filename"   --Log debug message to file, default is output to console, if DISABLE_DEBUG_CONSOLE is not set.
-- DISABLE_DEBUG_CONSOLE = true|false   --If set, disable debug message output to console.
-- RUN = {"run-lua-script-file1", "file2", ... }
-- AUTHOR_KEY = "Remote-author-key",
-- or
-- AUTHOR_KEY = function_to_get_author_key( request_content, remote_addr ),
--
-- DEFAULT_REMOTE_KEY = "Default-remote-link-key",
-- or
-- DEFAULT_REMOTE_KEY = function_to_get_remote_link_key
--
-- ENABLE_DISCOVERY = true|false --Enable/disable discovery service.
--
-- DISCOVERY_HOST_DOMAIN = ".coaf"  --The discovered host domain (default is .coaf)
--
-- DISCOVERY_HOSTFILE = "/path/to/hosts"  --Write discovered host into hosts file.
-- or:
-- DISCOVERY_HOSTFILE = function(address-map, domain) --You can write hosts (in map, [host]=address style) by your self.

function _M._LoadConfigure(cfgStr)
	local funcGen = {
[[
	local args={...}
	local arg_cfgObj = args[1]

	local UNIQUE_ID
	local BIND
	local CONTROL_BIND
	local DEBUG_TX
	local DEBUG_RX
	local DEBUG_LEVEL
	local LOGFILE
	local DISABLE_DEBUG_CONSOLE
	local RUN
	local AUTHOR_KEY
	local DEFAULT_REMOTE_KEY
	local ENABLE_DISCOVERY
	local DISCOVERY_HOSTFILE
	local DISCOVERY_HOST_DOMAIN
	local HOSTNAME
]]
	}

	table.insert( funcGen, cfgStr )
	table.insert( funcGen, [[
	
	arg_cfgObj.UNIQUE_ID = UNIQUE_ID or arg_cfgObj.UNIQUE_ID
	arg_cfgObj.HOSTNAME = HOSTNAME or arg_cfgObj.HOSTNAME
	arg_cfgObj.BIND = BIND or arg_cfgObj.BIND
	arg_cfgObj.CONTROL_BIND = CONTROL_BIND or arg_cfgObj.CONTROL_BIND
	arg_cfgObj.ENABLE_DISCOVERY = ENABLE_DISCOVERY or arg_cfgObj.ENABLE_DISCOVERY
	arg_cfgObj.DEBUG_TX = DEBUG_TX or arg_cfgObj.DEBUG_TX
	arg_cfgObj.DEBUG_RX = DEBUG_RX or arg_cfgObj.DEBUG_RX
	arg_cfgObj.DEBUG_LEVEL = DEBUG_LEVEL or arg_cfgObj.DEBUG_LEVEL
	arg_cfgObj.LOGFILE = LOGFILE or arg_cfgObj.LOGFILE
	arg_cfgObj.DISABLE_DEBUG_CONSOLE = DISABLE_DEBUG_CONSOLE or arg_cfgObj.DISABLE_DEBUG_CONSOLE
	arg_cfgObj.RUN = RUN or arg_cfgObj.RUN
	arg_cfgObj.AUTHOR_KEY = AUTHOR_KEY or arg_cfgObj.AUTHOR_KEY
	arg_cfgObj.DEFAULT_REMOTE_KEY = DEFAULT_REMOTE_KEY or arg_cfgObj.DEFAULT_REMOTE_KEY
	arg_cfgObj.DISCOVERY_HOSTFILE = DISCOVERY_HOSTFILE or arg_cfgObj.DISCOVERY_HOSTFILE
	arg_cfgObj.DISCOVERY_HOST_DOMAIN = DISCOVERY_HOST_DOMAIN or arg_cfgObj.DISCOVERY_HOST_DOMAIN

	return arg_cfgObj
]])
	local func, err = load(table.concat( funcGen ))
	if not func then
		_dbg.Error("CoServer", "Fail to parse server configure:", err )
		return nil
	end
	local flag, ret = pcall(func, _M.configs)
	if not flag then
		_dbg.Error("CoServer", "Fail to load server configure:", ret)
		return nil
	end
	return ret
end

-----------------------------------------------------------
-- Internal use: Run scripts.
-----------------------------------------------------------
local function RunScript(file)
	local trunk, err = loadfile(file)
	if not trunk then
		_dbg.Error("CoServer", "Fail to load run script:", file)
		return false, "Fail to load script"
	end
	
	local st, err = pcall(trunk, _M)
	if not st then
		_dbg.Error("CoServer", "Fail to run script:", file, " With error:", err )
		return false, err
	end
	return true
end

-----------------------------------------------------------
-- Fix unix socket privilleges.
-----------------------------------------------------------
local function FixUnixSocket(path)
	os.execute( "/bin/chmod 777 " .. path )
end

-----------------------------------------------------------
-- Initialize, to parse configuration, load services and
-- auto start specified service(s), also run startup scripts
-- if RUN configure item is specified.
-----------------------------------------------------------
function _M.Init( versionTag )
	local cfgFile = _M.config_location .. PATH_DELIM .. _M.config_file
	local cfgF = io.open(cfgFile, "r")
	local log_cfgOpenErr, log_cfgParseErr

	if not cfgF then
		log_cfgOpenErr = true
	else
		local cfgStr = cfgF:read("*a")
		cfgF:close()
		local cfg = _M._LoadConfigure(cfgStr)
		if not cfg then
			log_cfgParseErr = true
		else

		end
	end

	--Change hostname first!
	if type(_M.configs.HOSTNAME) == "string" then
		coaf.SetHostname(_M.configs.HOSTNAME)
	elseif type(_M.configs.HOSTNAME) == "function" then
		pcall(_M.configs.HOSTNAME)
	end

	if type(_M.configs.LOGFILE) == "string" then
		local roll = require "coaf.log.rollfile"
		local logF = roll.Create(_M.configs.LOGFILE)
		if not logF or not logF:init() then
			_dbg.Error("CoServer", "Fail to open log file:", _M.configs.LOGFILE, " so the log messages should be output to console.")
		else
			_dbg.AddLogger("logfile", logF, nil, nil)
		end
	end 

	if _M.configs.DISABLE_DEBUG_CONSOLE then
		_dbg.DisableConsole()
	end

	if _M.configs.DEBUG_LEVEL then
		_dbg.DebugLevel( tonumber(_M.configs.DEBUG_LEVEL) )
	end

	_dbg.Info("CoServer", "*** CoServer (" .. (versionTag or "0.01") .. ") is running *** ")

	if log_cfgOpenErr then
		_dbg.Error("CoServer", "Fail to open configure file:", cfgFile )
		_dbg.Warn("CoServer", "Since no configure file, the CoServer runs with default configurations.")
	end

	if log_cfgParseErr then
		_dbg.Error("CoServer", "Fail to parse configure file:", cfgFile )
		_dbg.Warn("CoServer", "Since with wrong configure file, the CoServer runs with default configurations.")
	end

	co_daemon.debug_TX = _M.configs.DEBUG_TX
	co_daemon.debug_RX = _M.configs.DEBUG_RX

	co_services.config_location = _M.config_location

	co_services.LoadServices()
	co_services.LoadBlacklist()
	co_services.LoadRemoteExposeList()

	if type(_M.configs.BIND) == "string" then
		co_daemon.Bind(_M.configs.BIND)
		local unix, path = _M.configs.BIND:match("^(unix:)(.+)$")
		if unix then
			FixUnixSocket(path)
		end
	elseif type(_M.configs.BIND) == "table" then
		local i, addr
		for i, addr in ipairs(_M.configs.BIND) do
			co_daemon.Bind(addr)
			local unix, path = addr:match("^(unix:)(.+)$")
			if unix then
				FixUnixSocket(path)
			end
		end
	else
		_dbg.Error("CoServer", "Invalid CoServer bind address(s)")
	end

	if _M.configs.ENABLE_DISCOVERY then
		local discovery = require "coserver.discovery"
		discovery.Init()
	end

	if _M.configs.ENABLE_DISCOVERY and _M.configs.DISCOVERY_HOSTFILE and 
		( type(_M.configs.DISCOVERY_HOSTFILE) == "string" or type(_M.configs.DISCOVERY_HOSTFILE) == "function" ) then
		local dyn_service = require "coserver.dynamic_service"
		coaf.NewTimer( function()
			if update_host_firstrun or dyn_service.host_items_changed then
				update_host_firstrun = false
				dyn_service.host_items_changed = false
				local domain = _M.configs.DISCOVERY_HOST_DOMAIN or ""
				if domain ~= "" and domain:sub(1,1) ~= "." then
					domain = "." .. domain
				end

				if type(_M.configs.DISCOVERY_HOSTFILE) == "function" then
					pcall( _M.configs.DISCOVERY_HOSTFILE, dyn_service.host_items, domain )
				else
					local file = io.open( _M.configs.DISCOVERY_HOSTFILE, "w" )
					if file then
						file:write( "127.0.0.1\tlocalhost\n" )
						local k,v
						for k,v in pairs(dyn_service.host_items) do
							file:write( v .. "\t" .. k .. domain .. "\n" )
						end
						file:close()
					end
				end
			end
		end,0.001,0.5):start()
	end
	--Note: old actions are removed into 'GoStart()'
end

function _M.GoStart()
	if type(co_services.auto_starts) == "table" then
		local i, svc
		for i, svc in pairs(co_services.auto_starts) do
			_dbg.Info("CoServer", "* Auto start:", svc.service_id )
			co_services.StartService( svc.service_id:match("(.*)/(.*)$"))
		end
	end

	if type(_M.configs.RUN) == "string" then
		RunScript(_M.configs.RUN)
	elseif type(_M.configs.RUN) == "table" then
		local i, file
		for i, file in ipairs(_M.configs.RUN) do
			RunScript(file)
		end
	elseif _M.configs.RUN then
		_dbg.Warn("CoServer", "Invalid 'RUN' option configured, ignore.")
	end
end

-----------------------------------------------------------
-- Run script
-----------------------------------------------------------
function _M.Run(file)
	return RunScript(file)
end
-----------------------------------------------------------
-- Enable debug Tx detail.
-----------------------------------------------------------
function _M.EnableDebugTxDetail(en)
	_M.configs.DEBUG_TX = en
	co_daemon.EnableDebugTx(en)
end

-----------------------------------------------------------
-- Enable debug Rx detail.
-----------------------------------------------------------
function _M.EnableDebugRxDetail(en)
	_M.configs.DEBUG_RX = en
	co_daemon.EnableDebugRx(en)
end

-----------------------------------------------------------
-- Add debug filter, should be:
-- CMD=...,LOC=...,SVC=...,OBJ=...
-- NOTE:
-- 1. Separated by , or ;
-- 2. Each filter (CMD/LOC/SVC/OBJ) can specify multiple times.
-- 3. filter can be Lua matching pattern.
--
-- e.g.:
-- CMD=CALL,CMD=STATIC_CALL      #Filter 'CALL' or 'STATIC_CALL'
-- LOC=local,LOC=other,SVC=net.* 
--   #Filter location be 'local' or 'other', and any services
--   #start with 'net'.
-----------------------------------------------------------
function _M.AddDebugFilter(filter)
	return debug_filt.AddFilter(filter)
end

function _M.ClearDebugFilter()
	debug_filt.ClearFilter()
end

-----------------------------------------------------------

function _M.GetLocations( name_filter, type_filter )
	local rslt = {}
	local addr = {}
	if not type_filter or type_filter == "*" or type_filter == "configured" then
		local k, svc
		for k, svc in pairs(co_services.services) do
			if not name_filter or name_filter == "*" or svc.LOC:match(name_filter) then
				rslt[svc.LOC] = "configured"
			end
		end
	end

	if not type_filter or type_filter == "*" or type_filter == "online" then
		local k, conn
		for k, conn in pairs(co_daemon.connections) do
			if conn and conn.Type == "PROXY" and conn.ValidServices then
				local i, item
				for i, item in pairs(conn.ValidServices) do
					if not name_filter or name_filter == "*" or item[1]:match(name_filter) then
						rslt[ item[1] ] = "online"
						if conn.connected_addr then
							addr[ item[1] ] = conn.connected_addr
						end
					end
				end
			end
		end
	end

	return rslt, addr
end

-----------------------------------------------------------
function _M.GetServices( name_filter, type_filter )
	local rslt = {}
	if not type_filter or type_filter == "*" or type_filter == "configured" then
		local k, svc
		for k, svc in pairs(co_services.services) do
			if not name_filter then
				rslt[k] = "configured"
			elseif name_filter == "*" or k:match(name_filter) then
				rslt[k] = "configured"
			end
		end
	end

	if not type_filter or type_filter == "*" or type_filter == "disabled" then
		local k, svc
		for k, svc in pairs(co_services.blacklist) do
			if not name_filter or name_filter == "*" or svc:match(name_filter) then
				rslt[svc] = "disabled"
			end
		end
	end

	if not type_filter or type_filter == "*" or type_filter == "running" then
		local svc, _nouse
		for svc, _nouse in pairs(co_services.running) do
			if not name_filter or name_filter == "*" or svc:match(name_filter) then
				rslt[svc] = "running"
			end
		end
	end

	if not type_filter or type_filter == "*" or type_filter == "online" then
		local k, conn
		for k, conn in pairs(co_daemon.connections) do
			if conn and conn.Type == "PROXY" and conn.ValidServices then
				local i, item, key
				for i, item in pairs(conn.ValidServices) do
					key = (item[1] or "unknown") .. "/" .. (item[2] or "unknown")
					if not name_filter or name_filter == "*" or key:match(name_filter) then
						rslt[ key ] = "online"
					end
				end
			end
		end
	end

	return rslt
end

-----------------------------------------------------------
function _M.GetServiceDetail(loc, svcId)
	local key, svc = co_services.GetExactService(loc, svcId)
	if not key then
		return nil
	else
		local rslt = {
			exact_service = key,
			exact_LOC = svc.LOC,
			exact_ID = svc.ID,
			address = nil,
			port = nil,
			status = co_services.ServiceIsDisabled(loc, svcId) and "disabled" or "configured"
		}

		if not loc or loc == "ANY" or not svcId or svcId == "ANY" then
			return rslt
		elseif rslt.status == "disabled" then
			return rslt
		else
			if co_services.running[ key ] then
				rslt.status = "running"
			end

			local k, conn
			for k, conn in pairs(co_daemon.connections) do
				if conn and conn.Type == "PROXY" and conn.ValidServices then
					local i, item
					for i, item in pairs(conn.ValidServices) do
						if item[1] == loc and item[2] == svcId then
							rslt.status = "online"
							rslt.address = conn.connected_addr
							rslt.port = conn.connected_port
							return rslt
						end
					end
				end
			end
			return rslt
		end
	end
end

-----------------------------------------------------------
function _M.GetBlacklist()
	return co_services.blacklist
end

-----------------------------------------------------------
function _M.AddService(loc, svcId, cfg, save)
	if not co_services.AddService(loc, svcId, cfg, save) then
		return false
	else
		return true
	end
end

-----------------------------------------------------------
function _M.RemoveService(loc, svcId, save)
	return co_services.RemoveService(loc, svcId, save)
end

-----------------------------------------------------------
function _M.EnableService(loc, svcId, save)
	return co_services.EnableService(loc, svcId, save)
end

-----------------------------------------------------------
function _M.DisableService(loc, svcId, save)
	return co_services.DisableService(loc, svcId, save)
end

-----------------------------------------------------------
function _M.StartService(loc, svcId)
	return co_services.StartService(loc, svcId)
end

-----------------------------------------------------------
function _M.StopService(loc, svcId)
	return co_services.StopService(loc, svcId)
end

-----------------------------------------------------------
function _M.SignalService(loc, svcId, sig)
	return co_services.SignalService(loc, svcId, sig)
end
-----------------------------------------------------------
function _M.StopAllServices()
	return co_services.StopAllServices()
end
-----------------------------------------------------------
function _M.SignalAll(sig)
	return co_services.SignalAll(sig)
end

-----------------------------------------------------------
local function proxyLinkThread(thrObj, name, address, autoReLink, kaTimeout, exposeTable, locationId, authKey)
	local conn
	local unique_id, remote_auth_key
	while not thrObj.terminated do
		if not locationId then
			unique_id = _M.configs.UNIQUE_ID
		else
			unique_id = locationId
		end

		if type(unique_id) == "function" then
			unique_id = unique_id()
		end

		if not authKey then
			remote_auth_key = _M.configs.DEFAULT_REMOTE_KEY
		else
			remote_auth_key = authKey
		end

		if type(remote_auth_key) == "function" then
			remote_auth_key = remote_auth_key()
		end

		_dbg.Info("CoServer", "Proxy link '" .. name .. "' is connecting to:", address, "with specified location ID:", unique_id or "<NIL>" )
		conn = co_daemon.NewProxyLink(address, kaTimeout, exposeTable, unique_id, remote_auth_key)
		if conn then
			_dbg.Info("CoServer", "Proxy link '" .. name .. "' is established to:", address)
			if _M.on_TriggerProxyLinkEvent then
				_M.on_TriggerProxyLinkEvent(name, "established")
			end
			local err = conn:WaitDie()
			if _M.on_TriggerProxyLinkEvent then
				_M.on_TriggerProxyLinkEvent(name, "disconnected")
			end
			if err == "#CoAF-KILL#" and type(conn.Destroy) == "function" then
				_dbg.Info("CoServer", "Proxy link '" .. name .. "' is force aborted, destroy the link." )
				conn:Destroy()
			end
		else
			_dbg.Info("CoServer", "Proxy link '" .. name .. "' connects fail!")
			if _M.on_TriggerProxyLinkEvent then
				_M.on_TriggerProxyLinkEvent(name, "fail")
			end
		end

		if autoReLink then
			if type(autoReLink) == "number" and autoReLink > 0 then
				coaf.Sleep(autoReLink)
			end
		else
			break
		end
	end

	_dbg.Info("CoServer", "Proxy link '" .. name .. "' is destroied.")
	if _M.on_TriggerProxyLinkEvent then
		_M.on_TriggerProxyLinkEvent(name, "destroied")
	end
	_M.proxy_links[name] = nil
end

-----------------------------------------------------------
--name: The proxy link name (any string, for identifier)
--address: Target co-server address
--autoRelink: true or <number> [optional, default is false]
--		if true, relink immediately;
--		if <number>, sleep <number> seconds, then relink.
--kaTimeout: keep-alive timeout number [optional]
--exposeTable: The service expose table [optional]
--locationId: This service's location ID [optional, nil means use UNIQUE_ID configured]
--authKey: Remote authorization key.

function _M.CreateProxyLink(name, address, autoReLink, kaTimeout, exposeTable, locationId, authKey)
	if type(name) ~= "string" then
		return false, "Invalid proxy link name"
	end
	if _M.proxy_links[name] then
		return false, "Proxy link is already exist"
	end

	local linkObj = thread.New( proxyLinkThread )
	if not linkObj then
		return false, "Fail to create proxy link"
	else
		_M.proxy_links[name] = linkObj
		linkObj:StartNow(name, address, autoReLink, kaTimeout, exposeTable, locationId, authKey)
		return true
	end
end

-----------------------------------------------------------
function _M.DestroyProxyLink(name)
	if type(name) ~= "string" or not _M.proxy_links[name] then
		return false, "Invalid proxy link name specified"
	else
		_M.proxy_links[name]:Terminate()
		_M.proxy_links[name] = nil
	end
end
-----------------------------------------------------------
function _M.UpdateProxyLink(name)
	if not name then
		local link
		for name, link in pairs(_M.proxy_links) do
			link:Kill("reconnect")
		end
	else
		if _M.proxy_links[name] then
			_M.proxy_links[name]:Kill("reconnect")
		end
	end
end
-----------------------------------------------------------
--NOTE: Nothing did by default, but you can overwrite this function
--      in your implementation (by thisM.on_TriggerProxyLinkEvent = <your function>)
--      so you can do your actions, e.g. Notify event in CoService.
function _M.on_TriggerProxyLinkEvent(name, eventId)
end

-----------------------------------------------------------
return _M

