--[==[

The TCP Co-Socket for CoAF (Co-Application Framework)

The cosocket module is similar to lua-socket, but, it works
asynchronously. If you are calling cosocket APIs in a 
coroutine context and called API may be blocked, it 
should yield current coroutine, util unblocked or
timeouted, it returns.

The cosocket module depends 'lua-ev' module.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local ev = require "ev"
local socket = require "socket"

local _M = {}

local tcp_meta = {}

local function TRACE_CORO_RESUME(co,...)
	local r, st = coroutine.resume(co, ...)
	if not r then
		print("*** Thread error:", st)
	end
end

function tcp_meta.getfd(tcp)
	return tcp.sock:getfd()
end

function tcp_meta.setfd(tcp, fd)
	return tcp.sock:setfd(fd)
end

function tcp_meta.dirty(tcp)
	return tcp.sock:dirty()
end

function tcp_meta.accept(tcp)
	if tcp.accept_task then
		return nil, "busy"
	end

	local coro, ismain = coroutine.running()
	tcp.sock:settimeout(0)
	if tcp.timeout == 0 then  -- Timeout is 0, so return immediately.
		local sock, err = tcp.sock:accept()
		if not sock then
			return sock, err
		else
			return _M.build_socket(sock)
		end
	else
		local mainResult = { nil, "timeout" }

		-- Create IO task.
		local task = ev.IO.new( function(loop,wa)
			wa:stop(loop)
			--Clear timeout timer.
			if tcp.accept_task and tcp.accept_task.tmo_task then
				if tcp.accept_task.tmo_task:is_pending() then
					tcp.accept_task.tmo_task:clear_pending(loop)
				end
				tcp.accept_task.tmo_task:stop(loop)
			end

			tcp.accept_task = nil

			local sock, err = tcp.sock:accept()
			if sock then
				sock = _M.build_socket(sock, tcp.metatype)
				if tcp.metatype == "unix" then
					sock.unixaddr = tcp.unixaddr
				end
			end
			if ismain then
				mainResult = {sock,err}
				--print("**** accept()/1: unloop")
				loop:unloop()
			else
				TRACE_CORO_RESUME(coro, sock, err)
			end
		end, tcp.sock:getfd(), ev.READ)

		if task then
			local tmoTask
			-- That, if timeout < 0 or nil, wait infinity.
			if tcp.timeout and tcp.timeout > 0 then
				tmoTask = ev.Timer.new( function(loop,tmr)
					tmr:stop(loop)

					if task:is_pending() then
						task:clear_pending(loop)
					end
					task:stop(loop)
					tcp.accept_task = nil
					if ismain then
						mainResult = {nil, "timeout"} --TODO: Why can't set correct value to upvalue?
						--print("**** accept()/2: unloop")
						loop:unloop()
					else
						TRACE_CORO_RESUME(coro, nil, "timeout")
					end
				end, tcp.timeout, 0)
			end

			local loop = _G.CoAF_LOOP or ev.Loop.default
			tcp.accept_task = { 
				io_task = task,
				tmo_task = tmoTask,
				-- Cancel (for closing)
				cancel = function()
					if task:is_pending() then
						task:clear_pending(loop)
					end
					task:stop(loop)

					if tcp.accept_task.tmo_task then
						if tcp.accept_task.tmo_task:is_pending() then
							tcp.accept_task.tmo_task:clear_pending(loop)
						end
						tcp.accept_task.tmo_task:stop(loop)
					end
					tcp.accept_task = nil
					if ismain then
						mainResult = {nil,"closed"} --TODO: Why can't set correct upvalue?
						--print("**** accept()/3: unloop")
						loop:unloop()
					else
						TRACE_CORO_RESUME(coro, nil, "closed")
					end
				end
			}

			task:start(loop)
			if tmoTask then
				tmoTask:start(loop)
			end
			if ismain then
				loop:loop()
				return table.unpack(mainResult)
			else
				local r,err = coroutine.yield(coro)
				if not r and err == "#CoAF-KILL#" then
					if task:is_pending() then task:clear_pending(loop) end
					task:stop(loop)
					if tmoTask then
						if tmoTask:is_pending() then tmoTask:clear_pending(loop) end
						tmoTask:stop(loop)
					end
					tcp.accept_task = nil
				end
				return r,err
			end
		else
			return nil, "fail to start async accept"
		end
	end
end

function tcp_meta.bind(tcp, addr, port)
	if addr and addr:match("^unix:") then
		addr = addr:match("^unix:(.+)")
		if tcp.metatype ~= "unix" then
			local unix_sock = require "socket.unix"
			if not unix_sock then
				return nil, "fail to initialize unix socket"
			end
			local sock = unix_sock()
			if not sock then
				return nil, "fail to create unix socket"
			end
			if tcp.sock then
				tcp:close()
				collectgarbage()
			end
			local socket_core = require "cosocket.core"
			socket_core.flag_close_on_exec( sock:getfd(), socket_core.default_close_on_exec )
			tcp.sock = sock
			tcp.metatype = "unix"
		end
		tcp.unixaddr = addr
	end
	tcp.sock:setoption("reuseaddr", true)
	return tcp.sock:bind(addr,port)
end

function tcp_meta.cleanup(tcp, mode)
	local clear_tx, clear_rx
	if not mode or mode == "both" then
		clear_tx = true
		clear_rx = true
	elseif mode == "send" then
		clear_tx = true
		clear_rx = false
	elseif mode == "receive" then
		clear_tx = false
		clear_rx = true
	end

	if clear_rx and tcp.accept_task then
		tcp.accept_task.cancel()
	end

	if clear_tx and tcp.connect_task then
		tcp.connect_task.cancel()
	end

	if clear_tx and tcp.send_task then
		tcp.send_task.cancel()
	end

	if clear_rx and tcp.recv_task then
		tcp.recv_task.cancel()
	end
end

function tcp_meta.close(tcp)
	tcp:cleanup("both")
	tcp.sock:close()

end

function tcp_meta.connect(tcp,addr,port)
	if tcp.connect_task then
		return nil, "busy"
	end

	if addr and addr:match("^unix:") then
		addr = addr:match("^unix:(.+)")
		if tcp.metatype ~= "unix" then
			local unix_sock = require "socket.unix"
			if not unix_sock then
				return nil, "fail to initialize unix socket"
			end
			local sock = unix_sock()
			if not sock then
				return nil, "fail to create unix socket"
			end
			if tcp.sock then
				tcp:close()
				collectgarbage()
			end
			local socket_core = require "cosocket.core"
			socket_core.flag_close_on_exec( sock:getfd(), socket_core.default_close_on_exec )
			tcp.sock = sock
			tcp.metatype = "unix"
		end
		tcp.unixaddr = addr
	end

	tcp.sock:settimeout(0)
	local stat, err = tcp.sock:connect(addr,port)
	if not stat then
		if err ~= "timeout" or tcp.timeout == 0 then
			return stat, err
		end
	else
		return stat
	end
	
	-- If connect timeout, wait unitl it connected.
	local coro, ismain = coroutine.running()
	local mainResult = { nil, "timeout" } --TODO: Why can't set correct upvalue?

	local task = ev.IO.new( function(loop, wa)
		wa:stop(loop)
		--Clear timeout timer.
		if tcp.connect_task and tcp.connect_task.tmo_task then
			if tcp.connect_task.tmo_task:is_pending() then
				tcp.connect_task.tmo_task:clear_pending(loop)
			end
			tcp.connect_task.tmo_task:stop(loop)
		end

		tcp.connect_task = nil
		if ismain then
			mainResult = {1}
			--print("**** connect()/1: unloop")
			loop:unloop()
		else	
			TRACE_CORO_RESUME(coro, 1)
		end
	end, tcp.sock:getfd(), ev.WRITE)

	if task then
		local tmoTask
		-- That, if timeout < 0 or nil, wait infinity.
		if tcp.timeout and tcp.timeout > 0 then
			tmoTask = ev.Timer.new( function(loop,tmr)
				tmr:stop(loop)

				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)
				tcp.connect_task = nil
				if ismain then
					mainResult = {nil, "timeout"} --TODO: Why can't set correct upvalue?
					--print("**** connect()/2: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "timeout")
				end
			end, tcp.timeout, 0)
		end

		local loop = _G.CoAF_LOOP or ev.Loop.default
		tcp.connect_task = { 
			io_task = task,
			tmo_task = tmoTask,
			-- Cancel (for closing)
			cancel = function()
				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)

				if tcp.connect_task.tmo_task then
					if tcp.connect_task.tmo_task:is_pending() then
						tcp.connect_task.tmo_task:clear_pending(loop)
					end
					tcp.connect_task.tmo_task:stop(loop)
				end
				tcp.connect_task = nil
				if ismain then
					mainResult = { nil, "closed" } --TODO: Why can't set correct upvalue?
					--print("**** connect()/3: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "closed")
				end
			end
		}

		task:start(loop)
		if tmoTask then
			tmoTask:start(loop)
		end
		if ismain then
			loop:loop()
			return table.unpack(mainResult)
		else
			local r, err = coroutine.yield(coro)
			if not r and err == "#CoAF-KILL#" then
				if task:is_pending() then task:clear_pending(loop) end
				task:stop(loop)
				if tmoTask then
					if tmoTask:is_pending() then tmoTask:clear_pending(loop) end
					tmoTask:stop(loop)
				end
				tcp.connect_task = nil
			end
			return r,err
		end
	else
		return nil, "fail to start async connect"
	end
end

function tcp_meta.getpeername(tcp)
	if tcp.metatype == "unix" then
		return tcp.unixaddr, tcp.sock:getfd(), "unix"
	else
		return tcp.sock:getpeername()
	end
end

function tcp_meta.getsockname(tcp)
	if tcp.metatype == "unix" then
		return tcp.unixaddr, tcp.sock:getfd(), "unix"
	else
		return tcp.sock:getsockname()
	end
end

function tcp_meta.getstats(tcp)
	return tcp.sock:getstats()
end

function tcp_meta.listen(tcp,backlogs)
	return tcp.sock:listen(backlogs)
end

function tcp_meta.receive(tcp, pattern, prefix)
	if tcp.recv_task then
		return nil, "busy"
	end

	tcp.sock:settimeout(0)
	local coro, ismain = coroutine.running()

	if tcp.timeout == 0 then
		return tcp.sock:receive(pattern, prefix)
	else
		--Since the socket is internal buffered, so must check receive first!
		local data, err, partial = tcp.sock:receive(pattern,prefix)
		if data or err ~= "timeout" or (partial and #partial > 0) then
			return data, err, partial
		end
	end

	local mainResult = { nil, "timeout" }  --TODO: Why can't set correct upvalue?

	-- Wait read.
	local task = ev.IO.new( function(loop, wa)
		wa:stop(loop)
		--Clear timeout timer.
		if tcp.recv_task and tcp.recv_task.tmo_task then
			if tcp.recv_task.tmo_task:is_pending() then
				tcp.recv_task.tmo_task:clear_pending(loop)
			end
			tcp.recv_task.tmo_task:stop(loop)
		end

		tcp.recv_task = nil

		local data, err, partial = tcp.sock:receive(pattern, prefix)
		if ismain then
			mainResult = {data,err,partial}
			--print("**** receive()/1: unloop")
			loop:unloop()
		else
			TRACE_CORO_RESUME(coro, data, err, partial)
		end
	end, tcp.sock:getfd(), ev.READ)

	if task then
		local tmoTask
		-- That, if timeout < 0 or nil, wait infinity.
		if tcp.timeout and tcp.timeout > 0 then
			tmoTask = ev.Timer.new( function(loop,tmr)
				tmr:stop(loop)

				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)
				tcp.recv_task = nil
				if ismain then
					mainResult = {nil,"timeout"} --TODO: Why can't set correct upvalue?
					--print("**** receive()/2: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "timeout")
				end
			end, tcp.timeout, 0)
		end

		local loop = _G.CoAF_LOOP or ev.Loop.default
		tcp.recv_task = { 
			io_task = task,
			tmo_task = tmoTask,
			-- Cancel (for closing)
			cancel = function()
				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)

				if tcp.recv_task.tmo_task then
					if tcp.recv_task.tmo_task:is_pending() then
						tcp.recv_task.tmo_task:clear_pending(loop)
					end
					tcp.recv_task.tmo_task:stop(loop)
				end
				tcp.recv_task = nil
				if ismain then
					mainResult = {nil,"closed"} --TODO: Why can't set correct upvalue?
					--print("**** receive()/3: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "closed")
				end
			end
		}

		task:start(loop)
		if tmoTask then
			tmoTask:start(loop)
		end
		if ismain then
			loop:loop()
			return table.unpack(mainResult)
		else
			local r, err, part = coroutine.yield(coro)
			if not r and err == "#CoAF-KILL#" then
				if task:is_pending() then task:clear_pending(loop) end
				task:stop(loop)
				if tmoTask then
					if tmoTask:is_pending() then tmoTask:clear_pending(loop) end
					tmoTask:stop(loop)
				end
				tcp.recv_task = nil
			end
			return r,err,part
		end
	else
		return nil, "fail to start async recv"
	end
end

function tcp_meta.send(tcp, data, i, j)
	if not data then
		return nil, "invalid data"
	end
	
	if tcp.send_task then
		return nil, "busy"
	end

	local coro, ismain = coroutine.running()

	if not i then
		i=1
		j=#data
	elseif not j then
		j=#data
	end

	tcp.sock:settimeout(0)
	local ii, ee = tcp.sock:send(data, i, (j-i>65535) and (i+65535) or j )
	if tcp.timeout == 0 then
		return ii, ee
	elseif ii and  ii >= j then
		return ii, ee
	elseif not ii and ee ~= "timeout" then
		return ii, ee
	end

	if ii then
		i = ii + 1
	end
	
	-----

	local mainResult = { nil, "timeout" }	--TODO: Why can't set correct upvalue?

	local task = ev.IO.new(function(loop,wa)
		local x, e = tcp.sock:send( data, i, (j-i>65535) and (i+65535) or j )
		if (x and x >= j) or (not x and e ~= "timeout") then --all sent or error
			wa:stop(loop)
			if tcp.send_task.tmo_task then
				if tcp.send_task.tmo_task:is_pending() then
					tcp.send_task.tmo_task:clear_pending(loop)
				end
				tcp.send_task.tmo_task:stop(loop)
			end
			tcp.send_task = nil
			if ismain then
				--print("**** send()/1: unloop")
				loop:unloop()
				mainResult = {x, e}
			else
				TRACE_CORO_RESUME(coro, x, e)
			end
		elseif x then
			i = x + 1
		end
	end, tcp.sock:getfd(), ev.WRITE)

	if task then
		local tmoTask
		-- That, if timeout < 0 or nil, wait infinity.
		if tcp.timeout and tcp.timeout > 0 then
			tmoTask = ev.Timer.new( function(loop,tmr)
				tmr:stop(loop)

				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)
				tcp.send_task = nil
				if ismain then
					mainResult = {nil,"timeout"}  --TODO: Why can't set correct upvalue?
					--print("**** send()/2: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "timeout")
				end
			end, tcp.timeout, 0)
		end

		local loop = _G.CoAF_LOOP or ev.Loop.default
		tcp.send_task = { 
			io_task = task,
			tmo_task = tmoTask,
			-- Cancel (for closing)
			cancel = function()
				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)

				if tcp.send_task.tmo_task then
					if tcp.send_task.tmo_task:is_pending() then
						tcp.send_task.tmo_task:clear_pending(loop)
					end
					tcp.send_task.tmo_task:stop(loop)
				end
				tcp.send_task = nil
				if ismain then
					mainResult = {nil,"closed"}  --TODO: Why can't set correct upvalue?
					--print("**** send()/3: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "closed")
				end
			end
		}

		task:start(loop)
		if tmoTask then
			tmoTask:start(loop)
		end
		if ismain then
			loop:loop()
			return table.unpack(mainResult)
		else
			local r,err = coroutine.yield(coro)
			if not r and err == "#CoAF-KILL#" then
				if task:is_pending() then task:clear_pending(loop) end
				task:stop(loop)
				if tmoTask then
					if tmoTask:is_pending() then tmoTask:clear_pending(loop) end
					tmoTask:stop(loop)
				end
				tcp.send_task = nil
			end
			return r,err
		end
	else
		return nil, "fail to start async send"
	end
end

function tcp_meta.setoption(tcp,option,val)
	return tcp.sock:setoption(option,val)
end

function tcp_meta.setstats(tcp,received,sent,age)
	return tcp.sock:setstats(received,sent,age)
end

function tcp_meta.settimeout(tcp,value,mode)
	tcp.timeout = value or tcp.timeout
end

function tcp_meta.shutdown(tcp,mode)
	--TODO: close tasks.
	tcp:cleanup(mode)
	return tcp.sock:shutdown(mode)
end

function _M.build_socket(tcp,metatype)
	local socket_core = require "cosocket.core"

	tcp:settimeout(0)	--non-block socket.
	socket_core.flag_close_on_exec( tcp:getfd(), socket_core.default_close_on_exec )

	local tcp_sock = {
		sock = tcp,
		metatype = metatype or "tcp",
		unixaddr = '',
		timeout = 30,	-- The default action timeout!
		accept_task = nil,
		recv_task = nil,
		send_task = nil
	}
	return setmetatable(tcp_sock, {__index = tcp_meta})
end

return _M
